/*
 * Copyright (c) 2018, Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <unistd.h>
#include <malloc.h>
#include <stdlib.h>
#include "mlx5dv_dr.h"
#include "mlx5.h"

#define MAX_SEND_CQE	64
#define MAX_NUM_CORES	1
#define DR_MR_SIZE	4096
#define QUEUE_SIZE	128
#define SG_LIST_SIZE	1
#define WRITE_STE_ID	0
#define READ_STE_ID	1
#define MIN_READ_SYNC	64
#define DR_NS_MAX_POSTSEND_REQ	60
#define SIGNAL_PER_DIV_QUEUE	16
#define TH_NUMS_TO_DRAIN	2

#define NS_LOG_MAX_TABLE_SEND	10
#define NS_LOG_MIN_TABLE_SIZE	15

struct dr_send_ring {
	pthread_mutex_t		mutex;
	struct dr_cq		*cq;
	struct dr_qp		*qp;
	struct ibv_mr		*mr;
	/* How much wqe are waiting for completion */
	uint32_t		pending_wqe;
	/* Signal request per this trash hold value  */
	uint16_t		signal_th;
	/* Each post_send_size less than max_post_send_size*/
	uint32_t		max_post_send_size;
	/* manage the send queue */
	uint32_t		tx_head;
	char			*buf;
	struct ibv_send_wr	*wr_post;
	struct ibv_sge		*sg_read;
	uint32_t		buf_size;
	struct ibv_send_wr	wr;
	struct ibv_wc		wc[MAX_SEND_CQE];
	uint8_t			sync_buff[MIN_READ_SYNC];
	struct ibv_mr		*sync_mr;
	uint32_t		index;
};

/*
 * dr_fill_and_append_ste_send_info: Add data to update-list
 * parameters:
 * ste - The data that attached to this specific ste
 * size - of data to write
 * offset - of the data from start of the hw_ste entry
 * data - data
 * send_list - to append into it
 * copy_data - if true indicates that the data should be kept because it not
 *	       backuped any where (like in re-hash).
 *	       if false, it lets the data to be updated after it was added to
 *	       the list (only if the memory is backuped somewhere)
 */
void dr_fill_and_append_ste_send_info(struct dr_ste *ste, uint16_t size,
				      uint16_t offset, uint8_t *data,
				      struct dr_ste_send_info *ste_info,
				      struct list_head *send_list,
				      bool copy_data)
{
	ste_info->size		= size;
	ste_info->ste		= ste;
	ste_info->offset	= offset;

	if (copy_data) {
		memcpy(ste_info->data_cont, data, size);
		ste_info->data = ste_info->data_cont;
	} else {
		ste_info->data = data;
	}

	list_add_tail(send_list, &ste_info->send_list);
}

/* TBD: The function assumes few hard/coded parameters */
static int dr_prepare_qp_to_rts(struct mlx5dv_dr_ns *ns)
{
	struct dr_devx_qp_rts_attr rts_attr = {};
	struct dr_devx_qp_rtr_attr rtr_attr = {};
	struct dr_qp *dr_qp = ns->send_ring->qp;
	struct ibv_port_attr port_attr = {};
	enum ibv_mtu mtu = IBV_MTU_1024;
	uint16_t gid_index = 0;
	uint32_t rq_psn = 0x7; /* TBD */
	int port = 1; /* CX-4 */
	int ret;

	ret = ibv_query_port(ns->ctx, port, &port_attr);
	if (ret) {
		fprintf(stderr, "Failed to query port\n");
		return ret;
	}

	/* init */
	ret = dr_devx_modify_qp_rst2init(ns->ctx, dr_qp->obj, port);
	if (ret) {
		fprintf(stderr, "Failed to modify QP to INIT, ret: %d\n", ret);
		return ret;
	}

	/* RTR */
	ret = dr_devx_query_gid(ns->ctx, port, gid_index, &rtr_attr.dgid_attr);
	if (ret) {
		fprintf(stderr, "can't read sgid of index 0\n");
		return ret;
	}

	rtr_attr.mtu		= mtu;
	rtr_attr.qp_num		= dr_qp->obj->object_id;
	rtr_attr.rq_psn		= rq_psn;
	rtr_attr.min_rnr_timer	= 12;
	rtr_attr.port_num	= port;
	rtr_attr.sgid_index	= gid_index;

	ret = dr_devx_modify_qp_init2rtr(ns->ctx, dr_qp->obj,  &rtr_attr);
	if (ret) {
		fprintf(stderr, "Failed to modify QP to RTR, ret: %d\n", ret);
		return ret;
	}

	/* RTS */
	rts_attr.timeout	= 14;
	rts_attr.retry_cnt	= 7;
	rts_attr.rnr_retry	= 7;
	rts_attr.sq_psn		= rq_psn; /*TBD*/

	ret = dr_devx_modify_qp_rtr2rts(ns->ctx, dr_qp->obj, &rts_attr);
	if (ret) {
		fprintf(stderr, "Failed to modify QP to RTS, ret: %d\n", ret);
		return ret;
	}

	return 0;
}

/* each send-ring has its own ib resources */
static int dr_alloc_post_send_ring(struct mlx5dv_dr_ns *ns)
{
	struct dr_qp_init_attr init_attr = {};
	struct mlx5dv_pd mlx5_pd = {};
	int cq_size, page_size;
	struct mlx5dv_obj obj;
	int size;
	int access_flags = IBV_ACCESS_LOCAL_WRITE | IBV_ACCESS_REMOTE_WRITE |
			   IBV_ACCESS_REMOTE_READ;
	int ret;

	ns->send_ring = calloc(1, sizeof(*ns->send_ring));
	if (!ns->send_ring) {
		fprintf(stderr, "Couldn't allocate send-ring\n");
		return ENOMEM;
	}

	cq_size = QUEUE_SIZE + 1;
	ns->send_ring->cq = dr_create_cq(ns->ctx, ns->uar, cq_size);
	if (!ns->send_ring->cq) {
		fprintf(stderr, "failed to create CQ with %u entries\n",
			cq_size);
		ret = ENODEV;
		goto free_send_ring;
	}

	obj.pd.in = ns->pd;
	obj.pd.out = &mlx5_pd;

	ret = mlx5dv_init_obj(&obj, MLX5DV_OBJ_PD);
	if (ret)
		goto clean_cq;

	init_attr.cq			= ns->send_ring->cq;
	init_attr.pdn			= mlx5_pd.pdn;
	init_attr.uar			= ns->uar;
	init_attr.cap.max_send_wr	= QUEUE_SIZE;
	init_attr.cap.max_recv_wr	= 1;
	init_attr.cap.max_send_sge	= SG_LIST_SIZE;
	init_attr.cap.max_recv_sge	= SG_LIST_SIZE;
	init_attr.cap.max_inline_data	= DR_STE_SIZE;
	init_attr.qp_type		= IBV_QPT_RC;

	ns->send_ring->qp = dr_create_qp(ns->ctx, &init_attr);
	if (!ns->send_ring->qp)  {
		fprintf(stderr, "Couldn't create QP\n");
		ret = ENODEV;
		goto clean_cq;
	}

	ns->info.max_ste_inline = (ns->send_ring->qp->max_inline_data  / DR_STE_SIZE);
	ns->info.max_inline_size = (ns->send_ring->qp->max_inline_data > DR_STE_SIZE) ?
				    DR_STE_SIZE : init_attr.cap.max_inline_data;
	ns->info.max_send_wr = QUEUE_SIZE;

	ns->send_ring->signal_th = ns->info.max_send_wr / SIGNAL_PER_DIV_QUEUE;

	/* prepare qp to be used */
	if (dr_prepare_qp_to_rts(ns)) {
		fprintf(stderr, "Couldn't prepare QP\n");
		ret = ENODEV;
		goto clean_qp;
	}

	ns->send_ring->max_post_send_size =
		dr_icm_pool_chunk_size_to_byte(NS_LOG_MAX_TABLE_SEND,
					       DR_ICM_TYPE_STE);

	/* Allocating the max size as a buffer for writing */
	size = ns->send_ring->signal_th * ns->send_ring->max_post_send_size;
	page_size = sysconf(_SC_PAGESIZE);
	ns->send_ring->buf = memalign(page_size, size);
	if (!ns->send_ring->buf) {
		fprintf(stderr, "Couldn't allocate send-ring buf.\n");
		ret = ENOMEM;
		goto clean_qp;
	}

	memset(ns->send_ring->buf, 0, size);
	ns->send_ring->buf_size = size;

	ns->send_ring->mr = ibv_reg_mr(ns->pd, ns->send_ring->buf, size,
				       access_flags);
	if (!ns->send_ring->mr) {
		fprintf(stderr, "Couldn't register MR\n");
		ret = EACCES;
		goto free_mem;
	}

	ns->send_ring->sync_mr = ibv_reg_mr(ns->pd, ns->send_ring->sync_buff,
					    MIN_READ_SYNC,
					    IBV_ACCESS_LOCAL_WRITE |
					    IBV_ACCESS_REMOTE_READ |
					    IBV_ACCESS_REMOTE_WRITE);
	if (!ns->send_ring->sync_mr) {
		fprintf(stderr, "Couldn't allocate sync mr!\n");
		ret = EACCES;
		goto clean_mr;
	}

	ns->send_ring->wr_post = malloc(sizeof(struct ibv_send_wr) * (ns->info.max_send_wr + 1));
	if (!ns->send_ring->wr_post)
		goto clean_sync_mr;

	ns->send_ring->sg_read = malloc(sizeof(struct ibv_sge) * (ns->info.max_send_wr / 2));
	if (!ns->send_ring->sg_read)
		goto clean_wr_post;

	return 0;

clean_wr_post:
	free(ns->send_ring->wr_post);
clean_sync_mr:
	ibv_dereg_mr(ns->send_ring->sync_mr);
clean_mr:
	ibv_dereg_mr(ns->send_ring->mr);
free_mem:
	free(ns->send_ring->buf);
clean_qp:
	dr_destroy_qp(ns->send_ring->qp);
clean_cq:
	dr_destroy_cq(ns->send_ring->cq);
free_send_ring:
	free(ns->send_ring);

	return ret;
}

static void dr_free_post_send_ring(struct dr_send_ring *send_ring)
{
	free(send_ring->sg_read);

	free(send_ring->wr_post);

	dr_destroy_qp(send_ring->qp);

	dr_destroy_cq(send_ring->cq);

	ibv_dereg_mr(send_ring->sync_mr);

	ibv_dereg_mr(send_ring->mr);

	free(send_ring->buf);

	free(send_ring);
}

/*
 * namespace resources:
 * 1. ib resources: pd
 * 2. send ring resources: qp, cq, WR's, CQ's, MR's
 * 3. icm memory pool
 */
static int dr_init_resources(struct mlx5dv_dr_ns *ns)
{
	int ret = -1;

	ns->pd = ibv_alloc_pd(ns->ctx);
	if (!ns->pd) {
		fprintf(stderr, "Couldn't allocate PD\n");
		return EACCES;
	}

	ns->uar = mlx5dv_devx_alloc_uar(ns->ctx, 0);
	if (!ns->uar) {
		fprintf(stderr, "Can't allocate UAR\n");
		ret = EINVAL;
		goto clean_pd;
	}

	ns->ste_icm_pool = dr_icm_pool_create(ns, DR_ICM_TYPE_STE);
	if (!ns->ste_icm_pool) {
		fprintf(stderr, "Couldn't get icm memory for %s\n",
			ibv_get_device_name(ns->ctx->device));
		goto clean_uar;
	}

	ns->action_icm_pool = dr_icm_pool_create(ns, DR_ICM_TYPE_MODIFY_ACTION);
	if (!ns->action_icm_pool) {
		fprintf(stderr, "Couldn't get action icm memory for %s\n",
			ibv_get_device_name(ns->ctx->device));
		goto free_ste_icm_pool;
	}

	ret = dr_alloc_post_send_ring(ns);
	if (ret) {
		fprintf(stderr, "Couldn't create send-ring for %s\n",
			ibv_get_device_name(ns->ctx->device));
		goto free_action_icm_pool;
	}

	return 0;

free_action_icm_pool:
	dr_icm_pool_destroy(ns->action_icm_pool);
free_ste_icm_pool:
	dr_icm_pool_destroy(ns->ste_icm_pool);
clean_uar:
	mlx5dv_devx_free_uar(ns->uar);
clean_pd:
	ibv_dealloc_pd(ns->pd);

	return ret;
}

static void dr_free_resources(struct mlx5dv_dr_ns *ns)
{
	dr_free_post_send_ring(ns->send_ring);

	dr_icm_pool_destroy(ns->action_icm_pool);
	dr_icm_pool_destroy(ns->ste_icm_pool);

	mlx5dv_devx_free_uar(ns->uar);
	ibv_dealloc_pd(ns->pd);
}


static int dr_query_vport_cap(struct ibv_context *ctx, uint16_t vport_number,
			      struct dr_devx_vport_cap *cap)
{
	bool other_vport = vport_number ? true : false;
	int ret;

	ret = dr_devx_query_esw_vport_context(ctx, other_vport, vport_number,
					      &cap->icm_address_rx,
					      &cap->icm_address_tx);
	if (ret)
		return ret;


	ret = dr_devx_query_gvmi(ctx, other_vport, vport_number, &cap->gvmi);
	if (ret)
		return ret;

	return 0;
}

static int dr_query_fdb_caps(struct ibv_context *ctx,
			     struct mlx5dv_dr_ns *ns)
{
	struct dr_esw_caps esw_caps = {};
	int num_vports;
	int ret;
	int i;

	num_vports = ns->info.attr.phys_port_cnt - 1;
	ns->info.caps.vports_caps = calloc(num_vports + 1,
					   sizeof(struct dr_devx_vport_cap));
	if (!ns->info.caps.vports_caps)
		return ENOMEM;

	/* Query vports */
	for (i = 0; i < num_vports; i++) {
		ret = dr_query_vport_cap(ctx, i, &ns->info.caps.vports_caps[i]);
		if (ret)
			goto err;
	}

	/* Query uplink */
	ret = dr_devx_query_esw_caps(ctx, &esw_caps);
	if (ret)
		goto err;

	ns->info.caps.fdb_sw_owner = esw_caps.sw_owner;
	ns->info.caps.vports_caps[i].icm_address_rx = esw_caps.uplink_icm_address_rx;
	ns->info.caps.vports_caps[i].icm_address_tx = esw_caps.uplink_icm_address_tx;
	ns->info.caps.esw_rx_drop_address = esw_caps.drop_icm_address_rx;
	ns->info.caps.esw_tx_drop_address = esw_caps.drop_icm_address_tx;
	ns->info.caps.num_vports = num_vports;

	return 0;
err:
	free(ns->info.caps.vports_caps);
	ns->info.caps.vports_caps = NULL;
	return ret;
}

static int dr_ns_caps_init(struct ibv_context *ctx, struct mlx5dv_dr_ns *ns)
{
	struct dr_devx_vport_cap *vport_cap;
	int ret;

	ret = ibv_query_device(ctx, &ns->info.attr);
	if (ret)
		return ret;

	ret = dr_devx_query_device(ctx, &ns->info.caps);
	if (ret) {
		/* Ignore devx query fail, to allow steering on root level
		 * tables in case devx is not supported over mlx5dv_dr API
		 */
		return 0;
	}

	ret = dr_query_fdb_caps(ctx, ns);
	if (ret)
		/* Ignore fail since query might not be supported */
		ret = 0;

	switch (ns->domain) {
	case MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS:
		if (!ns->info.caps.rx_sw_owner)
			return 0;

		ns->info.supp_sw_steering = true;
		ns->info.rx.ste_type = DR_STE_TYPE_RX;
		ns->info.rx.default_icm_addr = ns->info.caps.nic_rx_drop_address;
		ns->info.rx.drop_icm_addr = ns->info.caps.nic_rx_drop_address;
		break;
	case MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS:
		if (!ns->info.caps.tx_sw_owner)
			return 0;

		ns->info.supp_sw_steering = true;
		ns->info.tx.ste_type = DR_STE_TYPE_TX;
		ns->info.tx.default_icm_addr = ns->info.caps.nic_tx_allow_address;
		ns->info.tx.drop_icm_addr = ns->info.caps.nic_tx_drop_address;
		break;
	case MLX5DV_DR_NS_DOMAIN_FDB_BYPASS:
		if (!ns->info.caps.eswitch_manager)
			return 0;

		if (!ns->info.caps.fdb_sw_owner)
			return 0;

		ns->info.rx.ste_type = DR_STE_TYPE_RX;
		ns->info.tx.ste_type = DR_STE_TYPE_TX;
		vport_cap = dr_get_vport_cap(&ns->info.caps ,0);
		if (!vport_cap)
			return EINVAL;

		ns->info.rx.default_icm_addr = vport_cap->icm_address_rx;
		vport_cap = dr_get_vport_cap(&ns->info.caps, WIRE_PORT);
		if (!vport_cap)
			return EINVAL;

		ns->info.supp_sw_steering = true;
		ns->info.tx.default_icm_addr = vport_cap->icm_address_tx;
		ns->info.rx.drop_icm_addr = ns->info.caps.esw_rx_drop_address;
		ns->info.tx.drop_icm_addr = ns->info.caps.esw_tx_drop_address;
		break;
	default:
		fprintf(stderr, "Invalid domain\n");
		ret = EINVAL;
	}

	return ret;
}

static struct dr_send_ring *dr_get_send_ring(struct mlx5dv_dr_ns *ns)
{
	return ns->send_ring;
}

static bool dr_is_device_fatal(struct mlx5dv_dr_ns *ns)
{
	struct mlx5_context *mlx5_ctx = to_mctx(ns->ctx);

	if (mlx5_ctx->async_event_on & (1 << IBV_EVENT_DEVICE_FATAL))
		return true;
	return false;
}

/*
 * The function tries to consume one wc each time, unless the queue is full, in
 * that case, which means that the hw is behind the sw in a full queue len
 * the function will drain the cq till it empty.
 */
static int dr_handle_panding_wc(struct mlx5dv_dr_ns *ns,
				struct dr_send_ring *send_ring)
{
	bool is_drain = false;
	int ne;

	if (send_ring->pending_wqe >= send_ring->signal_th) {
		/* queue is full, its size signal_th, but we have READ+WRITE */
		if (send_ring->pending_wqe >= ns->send_ring->signal_th * TH_NUMS_TO_DRAIN)
			is_drain = true;

		do {
			/* on IBV_EVENT_DEVICE_FATAL go out of draining */
			if (is_drain && dr_is_device_fatal(ns))
				return ENODEV;

			// TODO Get CQE status
			ne = dr_poll_cq(send_ring->cq, 1);
			if (ne < 0) {
				fprintf(stderr, "poll CQ failed after IBV_WR_SEND\n");
				return ne;
			} else if (ne == 1) {
				send_ring->pending_wqe =
					send_ring->pending_wqe - send_ring->signal_th;
			}
		} while (is_drain && send_ring->pending_wqe);
	}

	return 0;
}

/*
 * The function prepares the list of wqe's, read after each write,
 * This in order to sync HW cache via read request after each write.
 * TBD: Change the api to understand when we don't need read.
 */
static void dr_ns_build_wqe_list(struct dr_send_ring *send_ring,
				 struct postsend_info *send_info,
				 struct ibv_send_wr *wr_post,
				 struct ibv_sge *sg_read,
				 uint16_t num_of_reqs)
{
	int i, j;

	/*
	 * i for the wqe list, which is double in its size,
	 * j for the rest (send_info, sg_read)
	 */
	for (i = 0, j = 0; j < num_of_reqs; i = i + 2, j++) {
		unsigned int inline_flag;

		/* WRITE */
		send_ring->pending_wqe++;
		wr_post[i].sg_list		= &send_info[j].sg;
		wr_post[i].num_sge		= 1;
		wr_post[i].opcode		= IBV_WR_RDMA_WRITE;

		if (!send_info[j].sg.lkey)
			inline_flag = IBV_SEND_INLINE;
		else
			inline_flag = 0;
		wr_post[i].send_flags		= inline_flag;

		if (send_ring->pending_wqe % send_ring->signal_th == 0)
			wr_post[i].send_flags |= IBV_SEND_SIGNALED;

		wr_post[i].wr.rdma.remote_addr	= send_info[j].mr_addr;
		wr_post[i].wr.rdma.rkey		= send_info[j].rkey;
		wr_post[i].next			= &wr_post[i + 1];

		/* READ */
		send_ring->pending_wqe++;
		sg_read[j].length = send_info[j].sg.length;
		if (inline_flag) {
			/* read into dedicated buffer */
			sg_read[j].addr		= (uintptr_t)send_ring->sync_buff;
			sg_read[j].lkey		= send_ring->sync_mr->lkey;
		} else {
			/* read into the same write area */
			sg_read[j].addr		= (uintptr_t)send_info[j].sg.addr;
			sg_read[j].lkey		= send_ring->mr->lkey;
		}

		wr_post[i + 1].opcode			= IBV_WR_RDMA_READ;
		wr_post[i + 1].wr.rdma.remote_addr	= send_info[j].mr_addr;
		wr_post[i + 1].wr.rdma.rkey		= send_info[j].rkey;
		wr_post[i + 1].sg_list			= &sg_read[j];
		wr_post[i + 1].num_sge			= 1;
		wr_post[i + 1].next			= &wr_post[i + 2];
		if (send_ring->pending_wqe % send_ring->signal_th == 0)
			wr_post[i + 1].send_flags = IBV_SEND_SIGNALED;
	}

	/* last entry */
	wr_post[i - 1].next = NULL;
}

/* TBD: Add lock over send */
static int dr_postsend_icm_data(struct mlx5dv_dr_ns *ns,
				struct postsend_info *send_info,
				uint16_t num_of_reqs)
{
	struct ibv_send_wr *wr_post;
	struct dr_send_ring *send_ring;
	struct ibv_sge *sg_read;
	uint32_t mem_offset = 0;
	uint32_t actual_reqs = 2 * num_of_reqs; /* READ + WRITE per request */
	uint32_t max_req_size;
	int i, ret;

	/* on IBV_EVENT_DEVICE_FATAL no operations over the HW */
	if (dr_is_device_fatal(ns))
		return 0;

	if (!num_of_reqs || actual_reqs > ns->info.max_send_wr) {
		fprintf(stderr, "%s (num-of-requests * 2) %d is bigger than %d!\n",
			__func__, actual_reqs, ns->info.max_send_wr);
		errno = EINVAL;
		return errno;
	}

	send_ring = dr_get_send_ring(ns);

	/* try get available completion */
	ret = dr_handle_panding_wc(ns, send_ring);
	if (ret)
		return ret;

	max_req_size = send_ring->max_post_send_size;

	wr_post = &send_ring->wr_post[send_ring->tx_head & (ns->info.max_send_wr - 1)];
	memset(wr_post, 0, actual_reqs * sizeof(struct ibv_send_wr));

	sg_read = &send_ring->sg_read[send_ring->tx_head & (ns->info.max_send_wr / 2 - 1)];
	memset(sg_read, 0, num_of_reqs * sizeof(struct ibv_sge));

	/* Fill all relevant fields in send_info */
	for (i = 0; i < num_of_reqs; i++) {
		uint32_t buff_offset;
		/* check if can inline */
		if (send_info[i].sg.length > ns->info.max_inline_size) {
			/* TBD: when we will have more than 1 request,
			 * the mem/offset calculation should be changed
			 */
			if (max_req_size < mem_offset + send_info[i].sg.length) {
				fprintf(stderr, "%s data size %d is bigger than mr_size %d!\n",
					__func__, mem_offset, send_ring->buf_size);
				errno = EINVAL;
				return errno;
			}

			/*
			 * The queue size is signal_th, each WRITE req has its
			 * place at the wq array.
			 */
			buff_offset = mem_offset +
				(send_ring->tx_head & (ns->send_ring->signal_th - 1)) *
				max_req_size;
			/* copy to ring mr */
			memcpy(send_ring->buf + buff_offset,
			       (void *)send_info[i].sg.addr,
			       send_info[i].sg.length);
			send_info[i].sg.addr	= (uintptr_t)send_ring->buf + buff_offset;
			send_info[i].sg.lkey	= send_ring->mr->lkey;
			mem_offset		+= send_info[i].sg.length;
		}
		send_ring->tx_head++;
	}

	dr_ns_build_wqe_list(send_ring, send_info, wr_post, sg_read,
			     num_of_reqs);

	ret = dr_post_send(send_ring->qp, wr_post);
	if (ret) {
		fprintf(stderr, "Failed post_send for IBV_WR_SEND ret: %d\n",
			ret);
		send_ring->pending_wqe = send_ring->pending_wqe - actual_reqs;
		send_ring->tx_head = send_ring->tx_head - num_of_reqs;
		return ret;
	}

	return 0;
}

/*
 * dr_postsend_ste: write size bytes into offset from the hw icm.
 *
 * Input:
 *	ns	- Name space
 *	ste	- The ste struct that contains the data (at least part of it)
 *	data	- The real data to send
 *	size	- data size for writing.
 *	offset	- The offset from the icm mapped data to start write to.
 *		  this for write only part of the buffer.
 *
 * Return: 0 on success.
 */
int dr_postsend_ste(struct mlx5dv_dr_ns *ns, struct dr_ste *ste,
		    uint8_t *data, uint16_t size, uint16_t offset)
{
	struct postsend_info send_info;

	send_info.sg.addr	= (uintptr_t) data;
	send_info.sg.length	= size;
	send_info.sg.lkey	= 0;
	send_info.mr_addr	= dr_ste_get_mr_addr(ste) + offset;
	send_info.rkey		= ste->htbl->chunk->rkey;

	return dr_postsend_icm_data(ns, &send_info, 1);
}


static int dr_ns_get_tbl_copy_details(struct mlx5dv_dr_ns *ns,
				      struct dr_ste_htbl *htbl,
				      uint8_t **data,
				      uint32_t *byte_size,
				      int *iterations,
				      int *num_stes)
{
	int alloc_size;

	if (htbl->chunk->byte_size > ns->send_ring->max_post_send_size) {
		*iterations = htbl->chunk->byte_size / ns->send_ring->max_post_send_size;
		*byte_size = ns->send_ring->max_post_send_size;
		alloc_size = *byte_size;
		*num_stes = *byte_size / DR_STE_SIZE;
	} else {
		*iterations = 1;
		*num_stes = htbl->chunk->num_of_entries;
		alloc_size = *num_stes * DR_STE_SIZE;
	}

	*data = calloc(1, alloc_size);
	if (!*data)
		return ENOMEM;

	return 0;
}

int dr_postsend_htbl(struct mlx5dv_dr_ns *ns, struct dr_ste_htbl *htbl,
		     uint8_t *formated_ste, uint8_t *mask)
{
	uint32_t byte_size = htbl->chunk->byte_size;
	struct postsend_info send_info;
	int i, j, num_stes_per_iter, iterations;
	uint8_t *data;
	int ret;

	ret = dr_ns_get_tbl_copy_details(ns, htbl, &data, &byte_size,
					 &iterations, &num_stes_per_iter);
	if (ret)
		return ret;

	/* Send the data iteration times */
	for (i = 0; i < iterations; i++) {
		uint32_t ste_index = (i > 0) ? (i * (byte_size / DR_STE_SIZE)) - 1 : 0;

		/* copy all ste's on the data buffer, need to add the bit_mask */
		for (j = 0; j < num_stes_per_iter; j++) {
			if (dr_ste_is_not_valid_entry(htbl->ste_arr[ste_index + j].hw_ste)) {
				memcpy(data + (j * DR_STE_SIZE),
				       formated_ste, DR_STE_SIZE);
			} else {
				/* copy data */
				memcpy(data + (j * DR_STE_SIZE), htbl->ste_arr[ste_index + j].hw_ste,
				       DR_STE_SIZE_REDUCED);
				/* copy bit_mask */
				memcpy(data + (j * DR_STE_SIZE) + DR_STE_SIZE_REDUCED,
				       mask, DR_STE_SIZE_MASK);
			}
		}

		send_info.sg.addr	= (uintptr_t) data;
		send_info.sg.length	= byte_size;
		send_info.sg.lkey	= 0;
		send_info.mr_addr	= dr_ste_get_mr_addr(htbl->ste_arr + ste_index);
		send_info.rkey		= htbl->chunk->rkey;

		ret = dr_postsend_icm_data(ns, &send_info, 1);
		if (ret)
			goto out_free;
	}

out_free:
	free(data);
	return ret;
}

/*
 * The function writing htbl with the same ste_init_data in all its STE's.
 * It used to init htbl on its creation.
 */
int dr_postsend_formated_htbl(struct mlx5dv_dr_ns *ns,
			      struct dr_ste_htbl *htbl,
			      uint8_t *ste_init_data,
			      bool update_hw_ste)
{
	struct postsend_info send_info;
	uint32_t byte_size = htbl->chunk->byte_size;
	uint8_t *data;
	int i, num_stes, iterations;
	int ret;

	ret = dr_ns_get_tbl_copy_details(ns, htbl, &data, &byte_size,
					 &iterations, &num_stes);
	if (ret)
		return ret;

	for (i = 0; i < num_stes; i++) {
		uint8_t *copy_dst;

		/* copy the same ste on the data buffer */
		copy_dst = data + i * DR_STE_SIZE;
		memcpy(copy_dst, ste_init_data, DR_STE_SIZE);

		if (update_hw_ste) {
			/* copy the reduced ste to hash table ste_arr */
			copy_dst = htbl->hw_ste_arr + i * DR_STE_SIZE_REDUCED;
			memcpy(copy_dst, ste_init_data, DR_STE_SIZE_REDUCED);
		}
	}

	/* Send the data iteration times */
	for (i = 0; i < iterations; i++) {
		uint32_t ste_index = (i > 0) ? (i * (byte_size / DR_STE_SIZE)) - 1 : 0;

		send_info.sg.addr	= (uintptr_t) data;
		send_info.sg.length	= byte_size;
		send_info.sg.lkey	= 0;
		send_info.mr_addr	= dr_ste_get_mr_addr(htbl->ste_arr + ste_index);
		send_info.rkey		= htbl->chunk->rkey;

		ret = dr_postsend_icm_data(ns, &send_info, 1);
		if (ret)
			goto out_free;
	}

out_free:
	free(data);
	return ret;
}

int dr_postsend_action(struct mlx5dv_dr_ns *ns, struct mlx5dv_dr_action *action)
{
	struct postsend_info send_info;

	send_info.sg.addr	= (uintptr_t) action->re_write.data;
	send_info.sg.length	= action->re_write.chunk->byte_size;
	send_info.sg.lkey	= 0;
	send_info.mr_addr	= action->re_write.chunk->mr_addr;
	send_info.rkey		= action->re_write.chunk->rkey;

	return dr_postsend_icm_data(ns, &send_info, 1);
}

static uint32_t dr_ns_calculate_max_table_allocation(struct dr_ns_rx_tx *nic_ns)
{
	uint32_t max_log_table_size;
	int i;

	/* minimum size of a table, in order to save allocation of mem and mr */
	max_log_table_size = NS_LOG_MIN_TABLE_SIZE;

	for (i = 0; i < DR_RULE_MAX_STES; i++)
		if (max_log_table_size < nic_ns->table_sizes[i])
			max_log_table_size = nic_ns->table_sizes[i];

	return max_log_table_size;
}

static void dr_ns_calculate_max_tables_allocation(struct mlx5dv_dr_ns *ns)
{
	uint32_t max_rx = DR_CHUNK_SIZE_1024K;
	uint32_t max_tx = DR_CHUNK_SIZE_1024K;

	if (ns->info.rx.table_size_was_set)
		max_rx = dr_ns_calculate_max_table_allocation(&ns->info.rx);

	if (ns->info.tx.table_size_was_set)
		max_tx = dr_ns_calculate_max_table_allocation(&ns->info.tx);

	ns->info.max_log_table_size = max(max_rx, max_tx);
}

static void dr_ns_get_log_table_size(const char *env_name,
				     struct dr_ns_rx_tx *nic_ns)
{
	char *p_init_tbl_size, *env;

	p_init_tbl_size = getenv("MLX5_INIT_LOG_TABLE_SIZE");
	if (p_init_tbl_size) {
		int i, tbl_size = 0;

		tbl_size = atoi(p_init_tbl_size);
		printf("NS: table start log size is %d\n", tbl_size);

		for (i = 0; i < DR_RULE_MAX_STES; i++)
			nic_ns->table_sizes[i] = tbl_size;

		nic_ns->table_size_was_set = true;
		return;
	}

	env = getenv(env_name);
	if (env) {
		sscanf(env, "%d %d %d %d %d %d %d %d %d %d",
		       &nic_ns->table_sizes[0],
		       &nic_ns->table_sizes[1],
		       &nic_ns->table_sizes[2],
		       &nic_ns->table_sizes[3],
		       &nic_ns->table_sizes[4],
		       &nic_ns->table_sizes[5],
		       &nic_ns->table_sizes[6],
		       &nic_ns->table_sizes[7],
		       &nic_ns->table_sizes[8],
		       &nic_ns->table_sizes[9]);

		printf("%s: ns from type: %s was created with log table sizes: %d %d %d %d %d %d %d %d %d %d\n",
		       __func__, env_name,
		       nic_ns->table_sizes[0],
		       nic_ns->table_sizes[1],
		       nic_ns->table_sizes[2],
		       nic_ns->table_sizes[3],
		       nic_ns->table_sizes[4],
		       nic_ns->table_sizes[5],
		       nic_ns->table_sizes[6],
		       nic_ns->table_sizes[7],
		       nic_ns->table_sizes[8],
		       nic_ns->table_sizes[9]);

		nic_ns->table_size_was_set = true;
	}
}

static void dr_ns_get_log_tables_size(struct mlx5dv_dr_ns *ns)
{
	if (ns->domain == MLX5DV_DR_NS_DOMAIN_FDB_BYPASS ||
	    ns->domain == MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS)
		dr_ns_get_log_table_size("MLX5_RX_LOG_TABLE_SIZE", &ns->info.rx);

	if (ns->domain == MLX5DV_DR_NS_DOMAIN_FDB_BYPASS ||
	    ns->domain == MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS)
		dr_ns_get_log_table_size("MLX5_TX_LOG_TABLE_SIZE", &ns->info.tx);
}

static void dr_ns_caps_uninit(struct mlx5dv_dr_ns *ns)
{
	if (ns->info.caps.vports_caps)
		free(ns->info.caps.vports_caps);
}

struct mlx5dv_dr_ns *mlx5dv_dr_create_ns(struct ibv_context *ctx,
					 enum mlx5dv_dr_ns_domain domain)
{
	struct mlx5dv_dr_ns *ns;
	int ret;

	if (domain >= MLX5DV_DR_NS_DOMAIN_RESERVED) {
		fprintf(stderr, "Failed allocate NS bad domain given\n");
		errno = EINVAL;
		return NULL;
	}

	ns = calloc(1, sizeof(*ns));
	if (!ns) {
		fprintf(stderr, "Failed allocate NS mem.\n");
		return NULL;
	}

	ns->ctx = ctx;
	ns->domain = domain;
	atomic_init(&ns->refcount, 1);

	if (dr_ns_caps_init(ctx, ns)) {
		fprintf(stderr, "Failed init NS, no Caps\n");
		goto free_ns;
	}

	if (ns->info.supp_sw_steering) {
		dr_ns_get_log_tables_size(ns);
		dr_ns_calculate_max_tables_allocation(ns);

		/* Allocate resources */
		ret = dr_init_resources(ns);
		if (ret) {
			fprintf(stderr, "Failed init NS resources for %s\n",
				ibv_get_device_name(ctx->device));
			goto uninit_caps;
		}
	}

	return ns;

uninit_caps:
	dr_ns_caps_uninit(ns);
free_ns:
	free(ns);
	return NULL;
}

int mlx5dv_dr_sync_domain(struct mlx5dv_dr_ns *ns, uint32_t flags)
{
	int ret = 0;

	if (!ns->info.supp_sw_steering) {
		errno = EOPNOTSUPP;
		return errno;
	}

	/* TBD: Change the signaled mechanism in order to make it
	 * simplier than that, like signal each req and create
	 * special request for drain.
	 */
	if (flags & MLX5DV_DR_FLAGS_SYNC_SW) {
		struct dr_send_ring *send_ring;
		struct postsend_info send_info;
		uint8_t data[DR_STE_SIZE];
		int i, num_of_sends_req;

		send_ring = dr_get_send_ring(ns);

		/* sending this amount of requests makes sure we will get drain */
		num_of_sends_req = send_ring->signal_th * TH_NUMS_TO_DRAIN / 2;

		/* send fake requests forcing the last to be signaled */
		send_info.sg.addr	= (uintptr_t) data;
		send_info.sg.length	= DR_STE_SIZE;
		send_info.sg.lkey	= 0;
		/* using the sync_mr in order to write/read */
		send_info.mr_addr	= (uintptr_t) send_ring->sync_mr->addr;
		send_info.rkey	= send_ring->sync_mr->rkey;

		for (i = 0; i < num_of_sends_req; i++) {
			ret = dr_postsend_icm_data(ns, &send_info, 1);
			if (ret)
				return ret;
		}

		ret = dr_handle_panding_wc(ns, send_ring);

		if (ret)
			return ret;
	}

	if (flags & MLX5DV_DR_FLAGS_SYNC_HW)
		ret = dr_devx_sync_steering(ns->ctx);

	return ret;
}

int mlx5dv_dr_destroy_ns(struct mlx5dv_dr_ns *ns)
{
	if (atomic_load(&ns->refcount) > 1)
		return EBUSY;

	if (ns->info.supp_sw_steering) {
		/* sync_ste to make sure resources are not used by the hardware */
		dr_devx_sync_steering(ns->ctx);
		dr_free_resources(ns);
	}

	dr_ns_caps_uninit(ns);

	free(ns);
	return 0;
}
