/*
 * Copyright (c) 2018, Mellanox Technologies. All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdlib.h>
#include <ccan/minmax.h>
#include "mlx5dv_dr.h"

#define DR_RULE_MAX_STE_CHAIN (DR_RULE_MAX_STES + DR_ACTION_MAX_STES)

struct dr_rule_action {
	struct mlx5dv_dr_action *action;
	/* atached to dr_rule via this */
	struct list_node	list;
};

static int dr_rule_connect_last_to_miss_list(struct dr_ste *new_last_ste,
					     struct list_head *miss_list,
					     struct list_head *send_list)
{
	struct dr_ste_send_info *ste_info_last;
	struct dr_ste *last_ste;

	/*
	 * the new entry will be inserted after the last:
	 * before:
	 * |_ste_| -->|__| ->|_last_ste_| -->/
	 * After:
	 * |_ste_| -->|__| ->|_last_ste_| --> |_new_last_ste_| -->/
	 */
	last_ste = list_tail(miss_list, struct dr_ste, miss_list_node);
	if (!last_ste) {
		fprintf(stderr, "%s No last STE !!\n", __func__);
		errno = EINVAL;
		return errno;
	}

	ste_info_last = calloc(1, sizeof(*ste_info_last));
	if (!ste_info_last) {
		fprintf(stderr, "%s failed allocating ste_info_last\n", __func__);
		return ENOMEM;
	}

	dr_ste_set_miss_addr(last_ste->hw_ste, dr_ste_get_icm_addr(new_last_ste));
	list_add_tail(miss_list, &new_last_ste->miss_list_node);

	dr_fill_and_append_ste_send_info(last_ste, DR_STE_SIZE_REDUCED,
					 0, last_ste->hw_ste,
					 ste_info_last, send_list, true);

	return 0;
}

static struct dr_ste
*dr_create_one_ste_and_set_hw_ste(struct mlx5dv_dr_matcher *matcher,
				  struct dr_matcher_rx_tx *nic_matcher,
				  uint8_t *hw_ste)
{
	struct mlx5dv_dr_ns *ns = matcher->tbl->ns;
	struct dr_ste_htbl *new_htbl;
	struct dr_ste *ste;

	/* TBD: do we need ste entry or htble ?*/
	/* create new table for miss  entry */
	new_htbl = dr_ste_htbl_alloc(ns->ste_icm_pool,
				     DR_CHUNK_SIZE_1,
				     DR_STE_LU_TYPE_DONT_CARE,
				     0);
	if (!new_htbl) {
		printf("%s failed allocating table\n", __func__);
		return NULL;
	}

	/* one entry only */
	ste = new_htbl->ste_arr;

	dr_ste_set_miss_addr(hw_ste, nic_matcher->e_anchor->chunk->icm_addr);

	dr_htbl_get(new_htbl);

	return ste;
}

/* TBD: do we need ste entry or htble ?*/
static struct dr_ste *dr_rule_create_collision_entry(struct mlx5dv_dr_matcher *matcher,
						     struct dr_matcher_rx_tx *nic_matcher,
						     uint8_t *hw_ste,
						     struct dr_ste *orig_ste)
{
	struct dr_ste *ste;

	ste = dr_create_one_ste_and_set_hw_ste(matcher, nic_matcher, hw_ste);
	if (!ste) {
		printf("%s failed allocating ste\n", __func__);
		return NULL;
	}

	ste->ste_chain_location = orig_ste->ste_chain_location;

	/* in collision entry, all members share the same miss_list_head */
	ste->htbl->miss_list = dr_ste_get_miss_list(orig_ste);

	/* next table */
	if (dr_ste_create_next_htbl(matcher, nic_matcher, ste, hw_ste,
				    DR_CHUNK_SIZE_16)) {
		printf("%s failed allocating table\n", __func__);
		goto free_tbl;
	}

	return ste;

free_tbl:
	dr_htbl_put(ste->htbl);
	return NULL;
}

static struct dr_ste *dr_rule_find_ste_in_miss_list(struct list_head *miss_list,
						    uint8_t *hw_ste)
{
	struct dr_ste *ste;

	/* Check if hw_ste is present in the list */
	list_for_each(miss_list, ste, miss_list_node)
		if (dr_ste_equal_tag(ste->hw_ste, hw_ste))
			return ste;

	return NULL;
}

static struct dr_ste *
dr_rule_rehash_handle_collision(struct mlx5dv_dr_matcher *matcher,
				struct dr_matcher_rx_tx *nic_matcher,
				struct list_head *update_list,
				struct dr_ste *col_ste,
				uint8_t *hw_ste)
{
	struct dr_ste *new_ste;
	int ret;

	new_ste = dr_create_one_ste_and_set_hw_ste(matcher, nic_matcher, hw_ste);
	if (!new_ste)
		return NULL;

	/* in collision entry, all members share the same miss_list_head */
	new_ste->htbl->miss_list = dr_ste_get_miss_list(col_ste);

	/* update the previous from the list */
	ret = dr_rule_connect_last_to_miss_list(new_ste,
						dr_ste_get_miss_list(col_ste),
						update_list);
	if (ret) {
		fprintf(stderr, "%s failed update dup entry\n", __func__);
		goto err_exit;
	}

	return new_ste;

err_exit:
	dr_ste_free(new_ste, matcher, nic_matcher);
	return NULL;

}

static void dr_rule_rehash_copy_ste_ctrl(struct mlx5dv_dr_matcher *matcher,
					 struct dr_matcher_rx_tx *nic_matcher,
					 struct dr_ste *cur_ste,
					 struct dr_ste *new_ste)
{
	new_ste->next_htbl = cur_ste->next_htbl;
	new_ste->ste_chain_location = cur_ste->ste_chain_location;

	if (!dr_ste_is_last_in_rule(nic_matcher, new_ste->ste_chain_location))
		new_ste->next_htbl->pointing_ste = new_ste;

	/*
	 * we need to copy the refcount since this ste
	 * may have been traversed several times
	 */
	atomic_init(&new_ste->refcount, atomic_load(&cur_ste->refcount));

	/* link old ste's rule_mem list to the new ste */
	dr_rule_update_rule_member(cur_ste, new_ste);
	list_head_init(&new_ste->rule_list);
	list_append_list(&new_ste->rule_list, &cur_ste->rule_list);
}

static struct dr_ste *dr_rule_rehash_copy_ste(struct mlx5dv_dr_matcher *matcher,
					      struct dr_matcher_rx_tx *nic_matcher,
					      struct dr_ste *cur_ste,
					      struct dr_ste_htbl *new_htbl,
					      struct list_head *update_list)
{
	uint8_t hw_ste[DR_STE_SIZE] = {0};
	struct dr_ste_send_info *ste_info;
	bool use_update_list = false;
	struct dr_ste *new_ste;
	uint8_t sb_idx;
	int new_idx;

	/* Copy STE mask from the matcher */
	sb_idx = cur_ste->ste_chain_location - 1;
	dr_ste_set_bit_mask(hw_ste, nic_matcher->ste_builder[sb_idx].bit_mask);

	/* Copy STE control and tag */
	memcpy(hw_ste, cur_ste->hw_ste, DR_STE_SIZE_REDUCED);

	/* TBD: perhaps keep the hash somewhere in the origin ste instead of re-calculate */
	new_idx = dr_ste_calc_hash_index(hw_ste, new_htbl);
	new_ste = &new_htbl->ste_arr[new_idx];

	if (dr_ste_not_used_ste(new_ste)) {
		dr_htbl_get(new_htbl);
		list_add_tail(dr_ste_get_miss_list(new_ste), &new_ste->miss_list_node);
	} else {
		new_ste = dr_rule_rehash_handle_collision(matcher,
							  nic_matcher,
							  update_list,
							  new_ste,
							  hw_ste);
		if (!new_ste) {
			fprintf(stderr, "%s failed adding collision entry, index: %d\n",
				__func__, new_idx);
			return NULL;
		}
		new_htbl->ctrl.num_of_collisions++;
		use_update_list = true;
	}

	memcpy(new_ste->hw_ste, hw_ste, DR_STE_SIZE_REDUCED);

	new_htbl->ctrl.num_of_valid_entries++;

	if (use_update_list) {
		ste_info = calloc(1, sizeof(*ste_info));
		if (!ste_info) {
			fprintf(stderr, "%s failed allocating ste_info\n",
				__func__);
			goto err_exit;
		}
		dr_fill_and_append_ste_send_info(new_ste, DR_STE_SIZE, 0, hw_ste,
						 ste_info, update_list, true);
	}

	dr_rule_rehash_copy_ste_ctrl(matcher, nic_matcher, cur_ste, new_ste);

	return new_ste;

err_exit:
	dr_ste_free(new_ste, matcher, nic_matcher);
	return NULL;
}

static int dr_rule_rehash_copy_miss_list(struct mlx5dv_dr_matcher *matcher,
					 struct dr_matcher_rx_tx *nic_matcher,
					 struct list_head *cur_miss_list,
					 struct dr_ste_htbl *new_htbl,
					 struct list_head *update_list)
{
	struct dr_ste *tmp_ste, *cur_ste, *new_ste;

	list_for_each_safe(cur_miss_list, cur_ste, tmp_ste, miss_list_node) {
		new_ste = dr_rule_rehash_copy_ste(matcher,
						  nic_matcher,
						  cur_ste,
						  new_htbl,
						  update_list);
		if (!new_ste)
			goto err_insert;

		list_del(&cur_ste->miss_list_node);
		dr_htbl_put(cur_ste->htbl);
	}
	return 0;

err_insert:
	/* we don't try to undo a resize hash. consider abort */
	fprintf(stderr, "Fatal error during resize\n");
	return 0;
}

static int dr_rule_rehash_copy_htbl(struct mlx5dv_dr_matcher *matcher,
				    struct dr_matcher_rx_tx *nic_matcher,
				    struct dr_ste_htbl *cur_htbl,
				    struct dr_ste_htbl *new_htbl,
				    struct list_head *update_list)
{
	struct dr_ste *cur_ste;
	int cur_entries;
	int err = 0;
	int i;

	cur_entries = dr_icm_pool_chunk_size_to_entries(cur_htbl->chunk_size);

	if (cur_entries < 1) {
		printf("%s invalid number of entries\n", __func__);
		errno = EINVAL;
		return errno;
	}

	for (i = 0; i < cur_entries; i++) {
		cur_ste = &cur_htbl->ste_arr[i];
		if (dr_ste_not_used_ste(cur_ste)) /* empty, nothing to copy */
			continue;

		err = dr_rule_rehash_copy_miss_list(matcher,
						    nic_matcher,
						    dr_ste_get_miss_list(cur_ste),
						    new_htbl,
						    update_list);
		if (err)
			goto clean_copy;
	}

clean_copy:
	/* If we got here through goto then this is fatal. not handling */
	return err;
}

static struct dr_ste_htbl *dr_rule_rehash_htbl(struct mlx5dv_dr_rule *rule,
					       struct dr_rule_rx_tx *nic_rule,
					       struct dr_ste_htbl *cur_htbl,
					       uint8_t ste_location,
					       struct list_head *update_list,
					       enum dr_icm_chunk_size new_size)
{
	struct dr_matcher_rx_tx *nic_matcher = nic_rule->nic_matcher;
	struct dr_ns_rx_tx *nic_ns = nic_matcher->nic_tbl->nic_ns;
	struct mlx5dv_dr_matcher *matcher = rule->matcher;
	struct mlx5dv_dr_ns *ns = matcher->tbl->ns;
	uint8_t formated_ste[DR_STE_SIZE] = {0};
	struct dr_ste_send_info *ste_info;
	struct dr_htbl_connect_info info;
	struct dr_ste_htbl *new_htbl;
	struct dr_ste *ste_to_update;
	int err;

	ste_info = calloc(1, sizeof(*ste_info));
	if (!ste_info) {
		fprintf(stderr, "%s failed allocating ste_info\n", __func__);
		return NULL;
	}

	new_htbl = dr_ste_htbl_alloc(ns->ste_icm_pool,
				     new_size,
				     cur_htbl->lu_type,
				     cur_htbl->byte_mask);
	if (!new_htbl) {
		fprintf(stderr, "failed to allocate new hash table\n");
		goto free_ste_info;
	}

	/* write new table to HW */
	info.type = CONNECT_MISS;
	info.miss_icm_addr = nic_matcher->e_anchor->chunk->icm_addr;
	dr_ste_set_formated_ste(ns, nic_ns, new_htbl, formated_ste, &info);

	new_htbl->pointing_ste = cur_htbl->pointing_ste;
	new_htbl->pointing_ste->next_htbl = new_htbl;
	err = dr_rule_rehash_copy_htbl(matcher,
				       nic_matcher,
				       cur_htbl,
				       new_htbl,
				       update_list);
	if (err)
		goto free_new_htbl; // TODO currently this is fatal, it should not be

	if (dr_postsend_htbl(ns, new_htbl, formated_ste,
			     nic_matcher->ste_builder[ste_location - 1].bit_mask)) {
		fprintf(stderr, "%s failed writing table to HW\n", __func__);
		goto free_new_htbl;
	}

	/*
	 * Connect previous hash table to matcher start hash table.
	 * Note: We assume each table here is with size 1.
	 */
	if (ste_location == 1) {
		struct dr_ste_htbl *prev_htbl = cur_htbl->pointing_ste->htbl;

		/* On matcher s_anchor we keep an extra refcount */
		dr_htbl_get(new_htbl);
		dr_htbl_put(cur_htbl);

		nic_matcher->s_htbl = new_htbl;
		/*
		 * it is safe to operate dr_ste_set_hit_addr on the hw_ste here
		 * (48B len) which works only on first 32B
		 */
		dr_ste_set_hit_addr(prev_htbl->ste_arr[0].hw_ste,
				    new_htbl->chunk->icm_addr,
				    new_htbl->chunk->num_of_entries);

		ste_to_update = &prev_htbl->ste_arr[0];
	} else {
		dr_ste_set_hit_addr_by_next_htbl(cur_htbl->pointing_ste->hw_ste,
						 new_htbl);
		ste_to_update = cur_htbl->pointing_ste;
	}

	dr_fill_and_append_ste_send_info(ste_to_update, DR_STE_SIZE_REDUCED, 0,
					 ste_to_update->hw_ste, ste_info,
					 update_list, false);

	return new_htbl;

free_new_htbl:
	dr_ste_htbl_free(new_htbl);
free_ste_info:
	free(ste_info);
	return NULL;
}

static struct dr_ste_htbl *dr_rule_rehash(struct mlx5dv_dr_rule *rule,
					  struct dr_rule_rx_tx *nic_rule,
					  struct dr_ste_htbl *cur_htbl,
					  uint8_t ste_location,
					  struct list_head *update_list,
					  bool grow)
{
	enum dr_icm_chunk_size new_size;

	new_size = grow ? dr_icm_next_higher_chunk(cur_htbl->chunk_size) :
			  dr_icm_next_lower_chunk(cur_htbl->chunk_size);

	return dr_rule_rehash_htbl(rule, nic_rule, cur_htbl, ste_location,
				   update_list, new_size);
}

static struct dr_ste *dr_rule_handle_collision(struct mlx5dv_dr_matcher *matcher,
					       struct dr_matcher_rx_tx *nic_matcher,
					       struct dr_ste *ste,
					       uint8_t *hw_ste,
					       struct list_head *miss_list,
					       struct list_head *send_list)
{
	struct dr_ste_send_info *ste_info;
	struct dr_ste *new_ste;

	ste_info = calloc(1, sizeof(*ste_info));
	if (!ste_info) {
		fprintf(stderr, "%s failed allocating ste_info\n", __func__);
		return NULL;
	}

	/* Add new friend */
	new_ste = dr_rule_create_collision_entry(matcher, nic_matcher, hw_ste, ste);
	if (!new_ste) {
		fprintf(stderr, "failed creating collision entry\n");
		goto free_send_info;
	}

	dr_fill_and_append_ste_send_info(new_ste, DR_STE_SIZE, 0, hw_ste, ste_info,
					 send_list, false);

	if (dr_rule_connect_last_to_miss_list(new_ste, miss_list, send_list)) {
		fprintf(stderr, "%s failed to update prev miss_list\n",
			__func__);
		goto err_exit;
	}

	ste->htbl->ctrl.num_of_collisions++;
	ste->htbl->ctrl.num_of_valid_entries++;

	return new_ste;

err_exit:
	dr_ste_free(new_ste, matcher, nic_matcher);
free_send_info:
	free(ste_info);
	return NULL;
}

/*
 * Each rule keeps a list of actions that attached to it.
 * Each action can be used in more than one rule, so each rule while is
 * beening destruct remove this action from the list and try do destruct
 * the action, the action will be destruct according to its refcount.
 */
static int dr_rule_add_action(struct mlx5dv_dr_rule *rule,
			      struct mlx5dv_dr_action *action)
{
	struct dr_rule_action *rule_mem;

	rule_mem = calloc(1, sizeof(*rule_mem));
	if (!rule_mem)
		return ENOMEM;

	rule_mem->action = action;
	list_add_tail(&rule->rule_actions_list, &rule_mem->list);

	return 0;
}

static void dr_rule_clean_rule_actions(struct mlx5dv_dr_rule *rule)
{
	struct dr_rule_action *rule_mem;
	struct dr_rule_action *tmp_mem;

	list_for_each_safe(&rule->rule_actions_list, rule_mem, tmp_mem, list) {
		list_del(&rule_mem->list);
		mlx5dv_dr_destroy_action(rule_mem->action);
		free(rule_mem);
	}
}

/*
 * while the pointer of ste is no longer valid, like while moving ste to be
 * the first in the miss_list, and to be in the origin table,
 * all rule-mem that are attached to this ste should update their ste member
 * to the new pointer
 */
void dr_rule_update_rule_member(struct dr_ste *ste, struct dr_ste *new_ste)
{
	struct dr_rule_member *rule_mem;

	list_for_each(&ste->rule_list, rule_mem, use_ste_list)
		rule_mem->ste = new_ste;
}

static void dr_rule_clean_rule_members(struct mlx5dv_dr_rule *rule,
				       struct dr_rule_rx_tx *nic_rule)
{
	struct dr_rule_member *rule_mem;
	struct dr_rule_member *tmp_mem;

	list_for_each_safe(&nic_rule->rule_members_list, rule_mem, tmp_mem, list) {
		list_del(&rule_mem->list);
		list_del(&rule_mem->use_ste_list);
		dr_ste_put(rule_mem->ste, rule->matcher, nic_rule->nic_matcher);
		free(rule_mem);
	}
}

static bool dr_rule_need_enlarge_hash(struct dr_ste_htbl *htbl,
				      struct mlx5dv_dr_ns *ns,
				      struct dr_ns_rx_tx *nic_ns)
{
	struct dr_ste_htbl_ctrl *ctrl = &htbl->ctrl;

	if (nic_ns->table_size_was_set)
		return false;

	if (ns->info.max_log_table_size <= htbl->chunk_size)
		return false;

	if (!ctrl->may_grow)
		return false;

	if (ctrl->num_of_collisions >= ctrl->increase_threashold &&
	    (ctrl->num_of_valid_entries - ctrl->num_of_collisions) >= ctrl->increase_threashold)
		return true;

	return false;
}

static int dr_rule_add_member(struct dr_rule_rx_tx *nic_rule,
			      struct dr_ste *ste)
{
	struct dr_rule_member *rule_mem;

	rule_mem = calloc(1, sizeof(*rule_mem));
	if (!rule_mem)
		return ENOMEM;

	rule_mem->ste = ste;
	list_add_tail(&nic_rule->rule_members_list, &rule_mem->list);

	list_add_tail(&ste->rule_list, &rule_mem->use_ste_list);

	return 0;
}

static int dr_rule_handle_rule_action(struct mlx5dv_dr_rule *rule,
				      struct dr_rule_rx_tx *nic_rule,
				      struct list_head *send_ste_list,
				      struct dr_ste *last_ste,
				      uint8_t *hw_ste_arr,
				      uint32_t new_hw_ste_arr_sz)
{
	struct dr_matcher_rx_tx *nic_matcher = nic_rule->nic_matcher;
	struct dr_ste_send_info *ste_info_arr[DR_ACTION_MAX_STES];
	uint8_t num_of_builders = nic_matcher->num_of_builders;
	struct mlx5dv_dr_matcher *matcher = rule->matcher;
	struct dr_ste *action_ste;
	uint8_t *curr_hw_ste, *prev_hw_ste;
	int i, k, ret;

	/* Two cases:
	 * 1. num_of_builders is eq to new_hw_ste_arr_sz, the action in the ste
	 * 2. num_of_builders is less then new_hw_ste_arr_sz, new ste was added
	 *    to support the action.
	 */
	if (num_of_builders == new_hw_ste_arr_sz)
		return 0;

	for (i = num_of_builders, k = 0; i < new_hw_ste_arr_sz; i++, k++) {

		curr_hw_ste = hw_ste_arr + i * DR_STE_SIZE;
		prev_hw_ste = (i == 0) ? curr_hw_ste : hw_ste_arr + ((i - 1) * DR_STE_SIZE);
		action_ste = dr_create_one_ste_and_set_hw_ste(matcher,
							      nic_matcher,
							      curr_hw_ste);
		if (!action_ste)
			return ENOMEM;

		dr_ste_get(action_ste);
		/*
		 * Add the action-ste to the list that contains only itself,
		 * This will let it to be free at the dr_ste_free function as
		 * the firast (and only) member in the list.
		*/
		list_add_tail(dr_ste_get_miss_list(action_ste),
			      &action_ste->miss_list_node);

		ste_info_arr[k] = calloc(1, sizeof(struct dr_ste_send_info));
		if (!ste_info_arr[k]) {
			fprintf(stderr, " %s failed alloce ste_info, k: %d\n",
				__func__, k);
			goto err_exit;
		}

		/* point current ste to the new action */
		dr_ste_set_hit_addr_by_next_htbl(prev_hw_ste, action_ste->htbl);
		ret = dr_rule_add_member(nic_rule, action_ste);
		if (ret) {
			fprintf(stderr, "%s failed adding rule member\n", __func__);
			goto free_ste_info;
		}
		dr_fill_and_append_ste_send_info(action_ste, DR_STE_SIZE, 0,
						 curr_hw_ste, ste_info_arr[k],
						 send_ste_list, false);
	}

	return 0;

free_ste_info:
	free(ste_info_arr[k]);
err_exit:
	dr_ste_put(action_ste, matcher, nic_matcher);
	return ENOMEM;
}

static struct dr_ste *dr_rule_handle_ste_branch(struct mlx5dv_dr_rule *rule,
						struct dr_rule_rx_tx *nic_rule,
						struct list_head *send_ste_list,
						struct dr_ste_htbl *cur_htbl,
						uint8_t *hw_ste,
						uint8_t ste_location,
						struct dr_ste_htbl **put_htbl)
{
	struct dr_matcher_rx_tx *nic_matcher = nic_rule->nic_matcher;
	struct dr_ns_rx_tx *nic_ns = nic_matcher->nic_tbl->nic_ns;
	struct mlx5dv_dr_matcher *matcher = rule->matcher;
	struct mlx5dv_dr_ns *ns = matcher->tbl->ns;
	struct dr_ste_send_info *ste_info;
	struct dr_ste_htbl *new_htbl;
	struct list_head *miss_list;
	struct dr_ste *matched_ste;
	bool skip_rehash = false;
	struct dr_ste *ste;
	int index;

again:
	index = dr_ste_calc_hash_index(hw_ste, cur_htbl);
	miss_list = &cur_htbl->chunk->miss_list[index];
	ste = &cur_htbl->ste_arr[index];

	if (dr_ste_not_used_ste(ste)) {
		/* Hash table index is free */
		uint32_t table_size = nic_ns->table_sizes[ste_location];

		/* take ref on table, only on first time this ste is used */
		dr_htbl_get(cur_htbl);

		/* new entry -> new branch */
		list_add_tail(miss_list, &ste->miss_list_node);

		dr_ste_set_miss_addr(hw_ste, nic_matcher->e_anchor->chunk->icm_addr);

		ste->ste_chain_location = ste_location;

		ste_info = calloc(1, sizeof(*ste_info));
		if (!ste_info) {
			fprintf(stderr, " %s failed allocating ste_info, index: %d\n",
				__func__, index);
			return NULL;
		}

		if (dr_ste_create_next_htbl(matcher,
					    nic_matcher,
					    ste,
					    hw_ste,
					    table_size)) {
			fprintf(stderr, "failed allocating table, index: %d\n",
				index);
			free(ste_info);
			return NULL;
		}

		cur_htbl->ctrl.num_of_valid_entries++;

		dr_fill_and_append_ste_send_info(ste, DR_STE_SIZE, 0, hw_ste,
						 ste_info, send_ste_list, false);
	} else {
		/* Hash table index in use. Is the hw_ste in the miss list ? */
		matched_ste = dr_rule_find_ste_in_miss_list(miss_list, hw_ste);
		if (matched_ste) {
			/*
			 * if it is last STE in the chain, and has the same tag
			 * it means that all the previous ste's are the same,
			 * so this rule was already inserted
			 */
			if (dr_ste_is_last_in_rule(nic_matcher, matched_ste->ste_chain_location)) {
				fprintf(stderr, "Duplicate rule inserted, aborting!!\n");
				return NULL;
			}
			return matched_ste;
		}

		if (!skip_rehash && dr_rule_need_enlarge_hash(cur_htbl, ns, nic_ns)) {
			/* Hash table index in use. Try to resize of the hash */
			skip_rehash = true;

			/*
			 * Hold the table till we update.
			 * Release in dr_rule_create_rule_nr()
			 */
			*put_htbl = cur_htbl;
			dr_htbl_get(cur_htbl);

			new_htbl = dr_rule_rehash(rule, nic_rule, cur_htbl,
						  ste_location, send_ste_list,
						  true);
			if (!new_htbl) {
				dr_htbl_put(cur_htbl);
				fprintf(stderr, "failed creating rehash table, htbl-log_size: %d\n",
					cur_htbl->chunk_size);
			} else {
				cur_htbl = new_htbl;
			}
			goto again;
		} else {
			/* Hash table index in use, add another collision (miss) */
			ste = dr_rule_handle_collision(matcher,
						       nic_matcher,
						       ste,
						       hw_ste,
						       miss_list,
						       send_ste_list);
			if (!ste) {
				fprintf(stderr, "failed adding collision entry, index: %d\n",
					index);
				return NULL;
			}
		}
	}
	return ste;
}

static bool dr_rule_to_h32_and_cmp(uint8_t *mask, uint8_t *value,
				   uint32_t s_idx, uint32_t e_idx)
{
	uint32_t i;

	for (i = s_idx; i < e_idx; i++) {
		if (value[i] & ~mask[i]) {
			fprintf(stderr, "Rule parameters contains a value not specified by mask\n");
			errno = EINVAL;
			return false;
		}
	}
	return true;
}

static bool dr_rule_verify(struct mlx5dv_dr_matcher *matcher,
			   struct mlx5dv_flow_match_parameters *value,
			   struct dr_match_param *param)
{
	uint8_t match_criteria = matcher->match_criteria;
	uint8_t *value_p = (uint8_t *)value->match_buf;
	uint8_t *param_p = (uint8_t *)param;
	uint8_t *mask_p = (uint8_t *)&matcher->mask;
	size_t value_size = value->match_sz;
	uint32_t s_idx, e_idx;

	if (!value_size ||
	    (value_size > sizeof(struct dr_match_param) ||
	     (value_size % sizeof(uint32_t)))) {
		fprintf(stderr, "Rule parameters length is incorrect\n");
		errno = EINVAL;
		return NULL;
	}

	if (match_criteria & DR_MATCHER_CRITERIA_OUTER) {
		s_idx = offsetof(struct dr_match_param, outer);
		e_idx = min(s_idx + sizeof(param->outer), value_size);

		memcpy(&param_p[s_idx], &value_p[s_idx], e_idx - s_idx);
		dr_be32_buffer_to_h32((uint64_t *)&param_p[s_idx], e_idx - s_idx);

		if (!dr_rule_to_h32_and_cmp(mask_p, param_p, s_idx, e_idx))
			return false;
	}

	if (match_criteria & DR_MATCHER_CRITERIA_MISC) {
		s_idx = offsetof(struct dr_match_param, misc);
		e_idx = min(s_idx + sizeof(param->misc), value_size);

		memcpy(&param_p[s_idx], &value_p[s_idx], e_idx - s_idx);
		dr_be32_buffer_to_h32((uint64_t *)&param_p[s_idx], e_idx - s_idx);

		if (!dr_rule_to_h32_and_cmp(mask_p, param_p, s_idx, e_idx))
			return false;
	}

	if (match_criteria & DR_MATCHER_CRITERIA_INNER) {
		s_idx = offsetof(struct dr_match_param, inner);
		e_idx = min(s_idx + sizeof(param->inner), value_size);

		memcpy(&param_p[s_idx], &value_p[s_idx], e_idx - s_idx);
		dr_be32_buffer_to_h32((uint64_t *)&param_p[s_idx], e_idx - s_idx);

		if (!dr_rule_to_h32_and_cmp(mask_p, param_p, s_idx, e_idx))
			return false;
	}

	if (match_criteria & DR_MATCHER_CRITERIA_MISC2) {
		s_idx = offsetof(struct dr_match_param, misc2);
		e_idx = min(s_idx + sizeof(param->misc2), value_size);

		memcpy(&param_p[s_idx], &value_p[s_idx], e_idx - s_idx);
		dr_be32_buffer_to_h32((uint64_t *)&param_p[s_idx], e_idx - s_idx);

		if (!dr_rule_to_h32_and_cmp(mask_p, param_p, s_idx, e_idx))
			return false;
	}

	if (match_criteria & DR_MATCHER_CRITERIA_MISC3) {
		s_idx = offsetof(struct dr_match_param, misc3);
		e_idx = min(s_idx + sizeof(param->misc3), value_size);

		memcpy(&param_p[s_idx], &value_p[s_idx], e_idx - s_idx);
		dr_be32_buffer_to_h32((uint64_t *)&param_p[s_idx], e_idx - s_idx);

		if (!dr_rule_to_h32_and_cmp(mask_p, param_p, s_idx, e_idx))
			return false;
	}
	return true;
}

static int dr_rule_destroy_rule_nic(struct mlx5dv_dr_rule *rule,
				    struct dr_rule_rx_tx *nic_rule)
{
	pthread_mutex_lock(&rule->matcher->mutex);
	dr_rule_clean_rule_members(rule, nic_rule);
	pthread_mutex_unlock(&rule->matcher->mutex);
	return 0;
}

static int dr_rule_destroy_rule_fdb(struct mlx5dv_dr_rule *rule)
{
	dr_rule_destroy_rule_nic(rule, &rule->rx);
	dr_rule_destroy_rule_nic(rule, &rule->tx);
	return 0;
}

static int dr_rule_destroy_rule(struct mlx5dv_dr_rule *rule)
{
	struct mlx5dv_dr_ns *ns = rule->matcher->tbl->ns;

	switch (ns->domain) {
	case MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS:
		dr_rule_destroy_rule_nic(rule, &rule->rx);
		break;
	case MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS:
		dr_rule_destroy_rule_nic(rule, &rule->tx);
		break;
	case MLX5DV_DR_NS_DOMAIN_FDB_BYPASS:
		dr_rule_destroy_rule_fdb(rule);
		break;
	default:
		errno = EINVAL;
		return errno;
	}

	atomic_fetch_sub(&rule->matcher->refcount, 1);
	free(rule);
	return 0;
}

static int dr_rule_skip(enum mlx5dv_dr_ns_domain domain,
			enum dr_ste_entry_type ste_type,
			struct dr_match_param *mask,
			struct dr_match_param *value)
{
	if (domain == MLX5DV_DR_NS_DOMAIN_FDB_BYPASS) {
		if (mask->misc.source_port) {
			if (ste_type == DR_STE_TYPE_RX)
				if (value->misc.source_port != WIRE_PORT)
					return 1;

			if (ste_type == DR_STE_TYPE_TX)
				if (value->misc.source_port == WIRE_PORT)
					return 1;
		}
	}

	return 0;
}

static int
dr_rule_create_rule_nic(struct mlx5dv_dr_rule *rule,
			struct dr_rule_rx_tx *nic_rule,
			struct dr_match_param *param,
			size_t num_actions,
			struct mlx5dv_dr_action *actions[])
{
	uint8_t hw_ste_arr[DR_RULE_MAX_STE_CHAIN * DR_STE_SIZE] = {0};
	struct dr_matcher_rx_tx *nic_matcher = nic_rule->nic_matcher;
	struct dr_ns_rx_tx *nic_ns = nic_matcher->nic_tbl->nic_ns;
	struct mlx5dv_dr_matcher *matcher = rule->matcher;
	struct mlx5dv_dr_ns *ns = matcher->tbl->ns;
	struct dr_ste_send_info *ste_info, *tmp_ste_info;
	struct dr_ste_htbl *htbl = NULL;
	struct dr_ste_htbl *cur_htbl;
	uint32_t new_hw_ste_arr_sz;
	LIST_HEAD(send_ste_list);
	struct dr_ste *ste = NULL; /* fix compilation warning*/
	int ret, i;

	list_head_init(&nic_rule->rule_members_list);

	if (dr_rule_skip(ns->domain, nic_ns->ste_type, &matcher->mask, param))
		return 0;

	ret = dr_ste_build_ste_arr(matcher, nic_matcher, param, hw_ste_arr);
	if (ret)
		goto out_err;

	ret = dr_actions_build_ste_arr(matcher, nic_matcher, actions,
				       num_actions, hw_ste_arr,
				       &new_hw_ste_arr_sz);
	if (ret)
		goto out_err;

	/* TBD: lock only the specific entry in the -table- */
	pthread_mutex_lock(&matcher->mutex);

	cur_htbl = nic_matcher->s_htbl;

	/* go over all the list of ste's */
	for (i = 0; i < nic_matcher->num_of_builders; i++) {
		/* calc crc and keep new ste entry */
		uint8_t *cur_hw_ste_ent = hw_ste_arr + (i * DR_STE_SIZE);
		ste = dr_rule_handle_ste_branch(rule,
						nic_rule,
						&send_ste_list,
						cur_htbl,
						cur_hw_ste_ent,
						i + 1,
						&htbl);
		if (!ste) {
			fprintf(stderr, "failed creating next branch\n");
			ret = ENOENT;
			goto free_rule_unlock;
		}

		cur_htbl = ste->next_htbl;

		/* keep all ste's in the rule struct */
		ret = dr_rule_add_member(nic_rule, ste);
		if (ret) {
			fprintf(stderr, "failed adding rule member index %d\n", i);
			goto free_ste;
		}

		dr_ste_get(ste);
	}

	/* connect actions */
	ret = dr_rule_handle_rule_action(rule, nic_rule, &send_ste_list,
					 ste, hw_ste_arr, new_hw_ste_arr_sz);
	if (ret) {
		fprintf(stderr, "%s: failed apply actions\n", __func__);
		goto free_rule_unlock;
	}

	list_for_each_rev_safe(&send_ste_list, ste_info, tmp_ste_info, send_list) {
		list_del(&ste_info->send_list);
		ret = dr_postsend_ste(matcher->tbl->ns, ste_info->ste,
				      ste_info->data, ste_info->size,
				      ste_info->offset);
		if (ret)
			printf("%s Failed sending ste!!!\n", __func__);
		/* copy data to ste, only reduced size */
		memcpy(ste_info->ste->hw_ste, ste_info->data,
		       DR_STE_SIZE_REDUCED);

		free(ste_info);
	}

	if (htbl)
		dr_htbl_put(htbl);

	pthread_mutex_unlock(&matcher->mutex);
	return 0;

free_ste:
	dr_ste_put(ste, matcher, nic_matcher);
free_rule_unlock:
	pthread_mutex_unlock(&matcher->mutex);
	dr_rule_clean_rule_members(rule, nic_rule);
	/* clean all ste_info's */
	list_for_each_safe(&send_ste_list, ste_info, tmp_ste_info, send_list) {
		list_del(&ste_info->send_list);
		free(ste_info);
	}
out_err:
	errno = ret;
	return errno;
}

static int
dr_rule_create_rule_fdb(struct mlx5dv_dr_rule *rule,
			struct dr_match_param *param,
			size_t num_actions,
			struct mlx5dv_dr_action *actions[])
{
	struct dr_match_param copy_param = {};
	int ret;

	/* Copy match_param since they will be consumed during the first
	 * nic_rule insertion.
	 *
	 * TODO: Might be possible not to copy or do a smart copy
	 * in case this will have a performance impact
	 */
	memcpy(&copy_param, param, sizeof(struct dr_match_param));

	ret = dr_rule_create_rule_nic(rule, &rule->rx, param,
				      num_actions, actions);
	if (ret)
		return ret;

	ret = dr_rule_create_rule_nic(rule, &rule->tx, &copy_param,
				      num_actions, actions);
	if (ret)
		goto destroy_rule_nic_rx;

	return 0;

destroy_rule_nic_rx:
	dr_rule_destroy_rule_nic(rule, &rule->rx);
	return ret;
}

static struct mlx5dv_dr_rule *
dr_rule_create_rule(struct mlx5dv_dr_matcher *matcher,
		    struct mlx5dv_flow_match_parameters *value,
		    size_t num_actions,
		    struct mlx5dv_dr_action *actions[])
{
	struct mlx5dv_dr_ns *ns = matcher->tbl->ns;
	struct dr_match_param param = {};
	struct mlx5dv_dr_rule *rule;
	int ret;

	if (!dr_rule_verify(matcher, value, &param))
		return NULL;

	rule = calloc(1, sizeof(*rule));
	if (!rule)
		return NULL;

	rule->matcher = matcher;

	switch (ns->domain) {
	case MLX5DV_DR_NS_DOMAIN_INGRESS_BYPASS:
		rule->rx.nic_matcher = &matcher->rx;
		ret = dr_rule_create_rule_nic(rule, &rule->rx, &param,
					      num_actions, actions);
		break;
	case MLX5DV_DR_NS_DOMAIN_EGRESS_BYPASS:
		rule->tx.nic_matcher = &matcher->tx;
		ret = dr_rule_create_rule_nic(rule, &rule->tx, &param,
					      num_actions, actions);
		break;
	case MLX5DV_DR_NS_DOMAIN_FDB_BYPASS:
		rule->rx.nic_matcher = &matcher->rx;
		rule->tx.nic_matcher = &matcher->tx;
		ret = dr_rule_create_rule_fdb(rule, &param,
					      num_actions, actions);
		break;
	default:
		ret = EINVAL;
		errno = ret;
		break;
	}

	if (ret)
		goto free_rule;

	atomic_fetch_add(&matcher->refcount, 1);
	return rule;

free_rule:
	free(rule);
	return NULL;
}

static struct mlx5dv_dr_rule *
dr_rule_create_rule_root(struct mlx5dv_dr_matcher *matcher,
			 struct mlx5dv_flow_match_parameters *value,
			 size_t num_actions,
			 struct mlx5dv_dr_action *actions[])
{
	struct mlx5dv_flow_action_attr *attr;
	struct mlx5dv_dr_rule *rule;
	uint32_t action_type;
	int i;

	rule = calloc(1, sizeof(*rule));
	if (!rule)
		return NULL;

	list_head_init(&rule->rule_actions_list);

	attr = calloc(num_actions, sizeof(*attr));
	if (!attr)
		goto free_rule;

	// TODO below should be part of dr_action.c
	for (i = 0; i < num_actions; i++) {
		action_type = actions[i]->action_type;

		switch (action_type) {
		case DR_ACTION_TYP_FT:
			if (actions[i]->dest_ft->ns != matcher->tbl->ns) {
				fprintf(stderr, "Destination table belongs to a different NS\n");
				errno = EINVAL;
				goto free_attr;
			}
			attr[i].type = MLX5DV_FLOW_ACTION_DEST_DEVX;
			attr[i].obj = actions[i]->dest_ft->devx_obj;
			break;
		case DR_ACTION_TYP_TNL_L2_TO_L2:
		case DR_ACTION_TYP_L2_TO_TNL_L2:
		case DR_ACTION_TYP_TNL_L3_TO_L2:
		case DR_ACTION_TYP_L2_TO_TNL_L3:
			if (!actions[i]->reformat.is_root_level) {
				fprintf(stderr, "Reformat action cannot be used on current table\n");
				goto free_attr;
			}
			attr[i].type = MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION;
			attr[i].action = actions[i]->reformat.flow_action;
			break;
		case DR_ACTION_TYP_MODIFY_HDR:
			if (!actions[i]->re_write.is_root_level) {
				fprintf(stderr, "Modify header action cannot be used on current table\n");
				goto free_attr;
			}
			attr[i].type = MLX5DV_FLOW_ACTION_IBV_FLOW_ACTION;
			attr[i].action = actions[i]->re_write.flow_action;
			break;
		case DR_ACTION_TYP_QP:
			attr[i].type = MLX5DV_FLOW_ACTION_DEST_IBV_QP;
			attr[i].qp = actions[i]->qp;
			break;
		case DR_ACTION_TYP_CTR:
			if (actions[i]->ctr.offeset != 0) {
				fprintf(stderr, "Bulk counter action cannot be used on current table\n");
				goto free_attr;
			}
			attr[i].type = MLX5DV_FLOW_ACTION_COUNTERS_DEVX;
			attr[i].obj = actions[i]->ctr.devx_obj;
			break;
		case DR_ACTION_TYP_TAG:
			attr[i].type = MLX5DV_FLOW_ACTION_TAG;
			attr[i].tag_value = actions[i]->flow_tag;
			break;
		default: /* TBD the rest of actions */
			printf("%s Found unsupported action type: %d\n",
			       __func__, action_type);
			goto free_attr;
		}

		atomic_fetch_add(&actions[i]->refcount, 1);

		/* collect all actions on the rule */
		if (dr_rule_add_action(rule, actions[i])) {
			mlx5dv_dr_destroy_action(actions[i]); /* include this */
			goto free_attr;
		}
	}

	rule->flow = mlx5dv_create_flow(matcher->dv_matcher, value, num_actions, attr);
	if (!rule->flow)
		goto free_attr;

	free(attr);
	rule->matcher = matcher;

	atomic_fetch_add(&matcher->refcount, 1);
	return rule;

free_attr:
	free(attr);
free_rule:
	dr_rule_clean_rule_actions(rule);
	free(rule);

	return NULL;
}

struct mlx5dv_dr_rule *mlx5dv_dr_create_rule(struct mlx5dv_dr_matcher *matcher,
					     struct mlx5dv_flow_match_parameters *value,
					     size_t num_actions,
					     struct mlx5dv_dr_action *actions[])
{
	if (dr_is_root_table(matcher->tbl))
		return dr_rule_create_rule_root(matcher, value, num_actions, actions);

	return dr_rule_create_rule(matcher, value, num_actions, actions);
}

static int dr_rule_destroy_rule_root(struct mlx5dv_dr_rule *rule)
{
	int err;

	dr_rule_clean_rule_actions(rule);

	if (rule->flow) {
		err = ibv_destroy_flow(rule->flow);
		if (err)
			return err;
	}

	atomic_fetch_sub(&rule->matcher->refcount, 1);
	free(rule);

	return 0;
}

int mlx5dv_dr_destroy_rule(struct mlx5dv_dr_rule *rule)
{

	if (dr_is_root_table(rule->matcher->tbl))
		return dr_rule_destroy_rule_root(rule);

	return dr_rule_destroy_rule(rule);
}
