/*
 * Copyright (c) 2018, Mellanox Technologies. All rights reserved.
 * Modified work Copyright 2018 Stephan Brumme
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *	Redistribution and use in source and binary forms, with or
 *	without modification, are permitted provided that the following
 *	conditions are met:
 *
 *	- Redistributions of source code must retain the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer.
 *
 *	- Redistributions in binary form must reproduce the above
 *	  copyright notice, this list of conditions and the following
 *	  disclaimer in the documentation and/or other materials
 *	  provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdlib.h>
#include <string.h>
#include <arpa/inet.h>
#include "mlx5dv_dr.h"

#define DR_STE_CRC_POLY		0xEDB88320L
#define IPV4_ETHERTYPE		0x0800
#define IPV6_ETHERTYPE		0x86DD
#define STE_IPV4		0x1
#define STE_IPV6		0x2
#define STE_TCP			0x1
#define STE_UDP			0x2
#define STE_SPI			0x3
#define IP_VERSION_IPV4		0x4
#define IP_VERSION_IPV6		0x6
#define IP_PROTOCOL_UDP		0x11
#define IP_PROTOCOL_TCP		0x06
#define IP_PROTOCOL_IPSEC	0x33
#define STE_SVLAN		0x1
#define STE_CVLAN		0x2
#define TCP_PROTOCOL		0x6
#define UDP_PROTOCOL 		0x11
#define IPSEC_PROTOCOL		0x33

#define DR_STE_ENABLE_FLOW_TAG (1 << 31)

/* helper macros */
#define __dr_ste_nullp(typ) ((struct dr_ste_struct_##typ##_bits *)0)
#define __dr_ste_bit_sz(typ, fld) sizeof(__dr_ste_nullp(typ)->fld)
#define __dr_ste_bit_off(typ, fld) (offsetof(struct dr_ste_struct_##typ##_bits, fld))
#define __dr_ste_dw_off(typ, fld) (__dr_ste_bit_off(typ, fld) / 32)
#define __dr_ste_dw_bit_off(typ, fld) (32 - __dr_ste_bit_sz(typ, fld) - (__dr_ste_bit_off(typ, fld) & 0x1f))
#define __dr_ste_mask(typ, fld) ((uint32_t)((1ull << __dr_ste_bit_sz(typ, fld)) - 1))
#define __dr_ste_dw_mask(typ, fld) (__dr_ste_mask(typ, fld) << __dr_ste_dw_bit_off(typ, fld))
#define __dr_ste_st_sz_bits(typ) sizeof(struct dr_ste_struct_##typ##_bits)

#define DR_STE_ST_SZ_BYTES(typ) (sizeof(struct dr_ste_struct_##typ##_bits) / 8)
#define DR_STE_ST_SZ_DW(typ) (sizeof(struct dr_ste_struct_##typ##_bits) / 32)
#define DR_STE_ST_SZ_QW(typ) (sizeof(struct dr_ste_struct_##typ##_bits) / 64)

/* Write to layout a value */
#define DR_STE_SET(typ, p, fld, v) do { \
	uint32_t _v = v; \
	static_assert(!(__dr_ste_st_sz_bits(typ) % 32), \
		      "struct size must be a multiple of 32 bit dwords"); \
	*((__be32 *)(p) + __dr_ste_dw_off(typ, fld)) = \
	htonl((ntohl(*((__be32 *)(p) + __dr_ste_dw_off(typ, fld))) & \
		     (~__dr_ste_dw_mask(typ, fld))) | (((_v) & __dr_ste_mask(typ, fld)) \
		     << __dr_ste_dw_bit_off(typ, fld))); \
} while (0)

/* Write to layout an array */
#define DR_STE_ARRAY_SET(typ, p, fld, idx, v) do { \
	static_assert(!(__dr_ste_bit_off(typ, fld) % 32), \
		      "field size must be a multiple of 32 bit dwords"); \
	DR_STE_SET(typ, p, fld[idx], v); \
} while (0)

/* Read from layout struct */
#define DR_STE_GET(typ, p, fld) \
	((ntohl(*((__be32 *)(p) + \
	__dr_ste_dw_off(typ, fld))) >> __dr_ste_dw_bit_off(typ, fld)) & \
	__dr_ste_mask(typ, fld))

/* Set to STE a specific value using DR_STE_SET */
#define DR_STE_SET_VAL(lookup_type, tag, t_fname, spec, s_fname, value) do { \
	if (spec->s_fname) { \
		DR_STE_SET(lookup_type, tag, t_fname, value); \
		spec->s_fname = 0; \
	} \
} while (0)

/* Set to STE spec->s_fname to tag->t_fname */
#define DR_STE_SET_TAG(lookup_type, tag, t_fname, spec, s_fname) do { \
	DR_STE_SET_VAL(lookup_type, tag, t_fname, spec, s_fname, spec->s_fname); \
} while (0)

/* Set to STE -1 to bit_mask->bm_fname and set spec->s_fname as used */
#define DR_STE_SET_MASK(lookup_type, bit_mask, bm_fname, spec, s_fname) do { \
	DR_STE_SET_VAL(lookup_type, bit_mask, bm_fname, spec, s_fname, -1); \
} while (0)

/* Set to STE spec->s_fname to bit_mask->bm_fname and set spec->s_fname as used */
#define DR_STE_SET_MASK_V(lookup_type, bit_mask, bm_fname, spec, s_fname) do { \
	DR_STE_SET_VAL(lookup_type, bit_mask, bm_fname, spec, s_fname, spec->s_fname); \
} while (0)

#define DR_STE_SET_TCP_FLAGS(lookup_type, tag, spec) do { \
	DR_STE_SET(lookup_type, tag, tcp_ns, !!(spec->tcp_flags & (1 << 8))); \
	DR_STE_SET(lookup_type, tag, tcp_cwr, !!(spec->tcp_flags & (1 << 7))); \
	DR_STE_SET(lookup_type, tag, tcp_ece, !!(spec->tcp_flags & (1 << 6))); \
	DR_STE_SET(lookup_type, tag, tcp_urg, !!(spec->tcp_flags & (1 << 5))); \
	DR_STE_SET(lookup_type, tag, tcp_ack, !!(spec->tcp_flags & (1 << 4))); \
	DR_STE_SET(lookup_type, tag, tcp_psh, !!(spec->tcp_flags & (1 << 3))); \
	DR_STE_SET(lookup_type, tag, tcp_rst, !!(spec->tcp_flags & (1 << 2))); \
	DR_STE_SET(lookup_type, tag, tcp_syn, !!(spec->tcp_flags & (1 << 1))); \
	DR_STE_SET(lookup_type, tag, tcp_fin, !!(spec->tcp_flags & (1 << 0))); \
} while (0)

#define DR_STE_SET_MPLS_MASK(lookup_type, mask, in_out, bit_mask) do { \
	DR_STE_SET_MASK_V(lookup_type, mask, mpls0_label, mask, \
		       in_out##_first_mpls_label);\
	DR_STE_SET_MASK_V(lookup_type, mask, mpls0_s_bos, mask, \
		       in_out##_first_mpls_s_bos); \
	DR_STE_SET_MASK_V(lookup_type, mask, mpls0_exp, mask, \
		       in_out##_first_mpls_exp); \
	DR_STE_SET_MASK_V(lookup_type, mask, mpls0_ttl, mask, \
		       in_out##_first_mpls_ttl); \
} while (0)

#define DR_STE_SET_MPLS_TAG(lookup_type, mask, in_out, tag) do { \
	DR_STE_SET_TAG(lookup_type, tag, mpls0_label, mask, \
		       in_out##_first_mpls_label);\
	DR_STE_SET_TAG(lookup_type, tag, mpls0_s_bos, mask, \
		       in_out##_first_mpls_s_bos); \
	DR_STE_SET_TAG(lookup_type, tag, mpls0_exp, mask, \
		       in_out##_first_mpls_exp); \
	DR_STE_SET_TAG(lookup_type, tag, mpls0_ttl, mask, \
		       in_out##_first_mpls_ttl); \
} while (0)

#define DR_STE_IS_OUTER_MPLS_OVER_GRE_SET(_misc) (\
	!!_misc->outer_first_mpls_over_gre_label \
	|| !!_misc->outer_first_mpls_over_gre_exp \
	|| !!_misc->outer_first_mpls_over_gre_s_bos \
	|| !!_misc->outer_first_mpls_over_gre_ttl)
#define DR_STE_IS_OUTER_MPLS_OVER_UDP_SET(_misc) (\
	!!_misc->outer_first_mpls_over_udp_label \
	|| !!_misc->outer_first_mpls_over_udp_exp \
	|| !!_misc->outer_first_mpls_over_udp_s_bos \
	|| !!_misc->outer_first_mpls_over_udp_ttl)

#define DR_STE_CALC_LU_TYPE(lookup_type, rx, inner) \
	((inner) ? DR_STE_LU_TYPE_##lookup_type##_I : \
		   (rx) ? DR_STE_LU_TYPE_##lookup_type##_D : \
			  DR_STE_LU_TYPE_##lookup_type##_O)

enum dr_ste_tunl_action {
	DR_STE_TUNL_ACTION_NONE		= 0,
	DR_STE_TUNL_ACTION_ENABLE	= 1,
	DR_STE_TUNL_ACTION_DECAP	= 2,
	DR_STE_TUNL_ACTION_L3_DECAP	= 3,
	DR_STE_TUNL_ACTION_POP_VLAN	= 4,
};

enum dr_ste_action_type {
	DR_STE_ACTION_TYPE_PUSH_VLAN	= 1,
	DR_STE_ACTION_TYPE_ENCAP_L3	= 3,
	DR_STE_ACTION_TYPE_ENCAP	= 4,
};

static bool dr_ste_crc32_init;
static uint32_t dr_ste_crc_tab32[8][256];

/* STE shared fields */
struct dr_ste_struct_general_bits {
	uint8_t		entry_type[0x4];
	uint8_t		reserved_at_4[0x4];
	uint8_t		entry_sub_type[0x8];
	uint8_t		byte_mask[0x10];
	uint8_t		next_table_base_63_48[0x10];
	uint8_t		next_lu_type[0x8];
	uint8_t		next_table_base_39_32_size[0x8];
	uint8_t		next_table_base_31_5_size[0x1b];
	uint8_t		linear_hash_enable[0x1];
	uint8_t		reserved_at_5c[0x2];
	uint8_t		next_table_rank[0x2];
	uint8_t		reserved_at_60[0xa0];
	uint8_t		tag_value[0x60];
	uint8_t		bit_mask[0x60];
};

struct dr_ste_struct_sx_transmit_bits {
	uint8_t		entry_type[0x4];
	uint8_t		reserved_at_4[0x4];
	uint8_t		entry_sub_type[0x8];
	uint8_t		byte_mask[0x10];
	uint8_t		next_table_base_63_48[0x10];
	uint8_t		next_lu_type[0x8];
	uint8_t		next_table_base_39_32_size[0x8];
	uint8_t		next_table_base_31_5_size[0x1b];
	uint8_t		linear_hash_enable[0x1];
	uint8_t		reserved_at_5c[0x2];
	uint8_t		next_table_rank[0x2];
	uint8_t		sx_wire;
	uint8_t		sx_func_lb;
	uint8_t		sx_sniffer;
	uint8_t		sx_wire_enable;
	uint8_t		sx_func_lb_enable;
	uint8_t		sx_sniffer_enable;
	uint8_t		action_type[3];
	uint8_t		reserved_at_69;
	uint8_t		action_description[6];
	uint8_t		gvmi[16];
	uint8_t		encap_pointer_vlan_data[32];
	uint8_t		loopback_syndome_en[8];
	uint8_t		loopback_syndome[8];
	uint8_t		counter_trigger[16];
	uint8_t		miss_address_63_48[16];
	uint8_t		counter_trigger_23_16[8];
	uint8_t		miss_address_39_32[8];
	uint8_t		miss_address_31_6[26];
	uint8_t		learning_point;
	uint8_t		go_back;
	uint8_t		match_polarity;
	uint8_t		mask_mode;
	uint8_t		miss_rank[2];
};

/* STE type rx_steering_multiple */
struct dr_ste_struct_rx_steering_mult_bits {
	uint8_t		entry_type[0x4];
	uint8_t		reserved_at_4[0x4];
	uint8_t		entry_sub_type[0x8];
	uint8_t		byte_mask[0x10];
	uint8_t		next_table_base_63_48[0x10];
	uint8_t		next_lu_type[0x8];
	uint8_t		next_table_base_39_32_size[0x8];
	uint8_t		next_table_base_31_5_size[0x1b];
	uint8_t		linear_hash_enable[0x1];
	uint8_t		reserved_at_[0x2];
	uint8_t		next_table_rank[0x2];
	uint8_t		member_count[0x10];
	uint8_t		gvmi[0x10];
	uint8_t		qp_list_pointer[0x20];
	uint8_t		reserved_at_a0[0x1];
	uint8_t		tunneling_action[0x3];
	uint8_t		action_description[0x4];
	uint8_t		reserved_at_a8[0x8];
	uint8_t		counter_trigger_15_0[0x10];
	uint8_t		miss_address_63_48[0x10];
	uint8_t		counter_trigger_23_16[0x08];
	uint8_t		miss_address_39_32[0x8];
	uint8_t		miss_address_31_6[0x1a];
	uint8_t		learning_point[0x1];
	uint8_t		fail_on_error[0x1];
	uint8_t		match_polarity[0x1];
	uint8_t		mask_mode[0x1];
	uint8_t		miss_rank[0x2];
};

/* STE type modify packet */
struct dr_ste_struct_modify_packet_bits {
	uint8_t		entry_type[0x4];
	uint8_t		reserved_at_4[0x4];
	uint8_t		entry_sub_type[0x8];
	uint8_t		byte_mask[0x10];
	uint8_t		next_table_base_63_48[0x10];
	uint8_t		next_lu_type[0x8];
	uint8_t		next_table_base_39_32_size[0x8];
	uint8_t		next_table_base_31_5_size[0x1b];
	uint8_t		linear_hash_enable[0x1];
	uint8_t		reserved_at_[0x2];
	uint8_t		next_table_rank[0x2];
	uint8_t		number_of_re_write_actions[0x10];
	uint8_t		gvmi[0x10];
	uint8_t		header_re_write_actions_pointer[0x20];
	uint8_t		reserved_at_a0[0x1];
	uint8_t		tunneling_action[0x3];
	uint8_t		action_description[0x4];
	uint8_t		reserved_at_a8[0x8];
	uint8_t		counter_trigger_15_0[0x10];
	uint8_t		miss_address_63_48[0x10];
	uint8_t		counter_trigger_23_16[0x08];
	uint8_t		miss_address_39_32[0x8];
	uint8_t		miss_address_31_6[0x1a];
	uint8_t		learning_point[0x1];
	uint8_t		fail_on_error[0x1];
	uint8_t		match_polarity[0x1];
	uint8_t		mask_mode[0x1];
	uint8_t		miss_rank[0x2];
};

/* TAG of lookup type ethernet L2 source */
struct dr_ste_struct_eth_l2_src_bits {
	uint8_t		smac_47_16[0x20];
	uint8_t		smac_15_0[0x10];
	uint8_t		l3_ethertype[0x10];
	uint8_t		qp_type[0x2];
	uint8_t		ethertype_filter[0x1];
	uint8_t		reserved_at_43[0x1];
	uint8_t		sx_sniffer[0x1];
	uint8_t		force_lb[0x1];
	uint8_t		functional_lb[0x1];
	uint8_t		port[0x1];
	uint8_t		reserved_at_48[0x4];
	uint8_t		first_priority[0x3];
	uint8_t		first_cfi[0x1];
	uint8_t		first_vlan_qualifier[0x2];
	uint8_t		reserved_at_52[0x2];
	uint8_t		first_vlan_id[0xc];
	uint8_t		ip_fragmented[0x1];
	uint8_t		tcp_syn[0x1];
	uint8_t		encp_type[0x2];
	uint8_t		l3_type[0x2];
	uint8_t		l4_type[0x2];
	uint8_t		reserved_at_68[0x4];
	uint8_t		second_priority[0x3];
	uint8_t		second_cfi[0x1];
	uint8_t		second_vlan_qualifier[0x2];
	uint8_t		reserved_at_72[0x2];
	uint8_t		second_vlan_id[0xc];
};

/* TAG of lookup type ethernet L2 destination */
struct dr_ste_struct_eth_l2_dst_bits {
	uint8_t		dmac_47_16[0x20];
	uint8_t		dmac_15_0[0x10];
	uint8_t		l3_ethertype[0x10];
	uint8_t		qp_type[0x2];
	uint8_t		ethertype_filter[0x1];
	uint8_t		reserved_at_43[0x1];
	uint8_t		sx_sniffer[0x1];
	uint8_t		force_lb[0x1];
	uint8_t		functional_lb[0x1];
	uint8_t		port[0x1];
	uint8_t		reserved_at_48[0x4];
	uint8_t		first_priority[0x3];
	uint8_t		first_cfi[0x1];
	uint8_t		first_vlan_qualifier[0x2];
	uint8_t		reserved_at_52[0x2];
	uint8_t		first_vlan_id[0xc];
	uint8_t		ip_fragmented[0x1];
	uint8_t		tcp_syn[0x1];
	uint8_t		encp_type[0x2];
	uint8_t		l3_type[0x2];
	uint8_t		l4_type[0x2];
	uint8_t		reserved_at_68[0x4];
	uint8_t		second_priority[0x3];
	uint8_t		second_cfi[0x1];
	uint8_t		second_vlan_qualifier[0x2];
	uint8_t		reserved_at_72[0x2];
	uint8_t		second_vlan_id[0xc];
};

/* TAG of lookup type ethernet L2 source destination */
struct dr_ste_struct_eth_l2_src_dst_bits {
	uint8_t		dmac_47_16[0x20];
	uint8_t		dmac_15_0[0x10];
	uint8_t		smac_47_32[0x10];
	uint8_t		smac_31_0[0x20];
	uint8_t		sx_sniffer[0x1];
	uint8_t		force_lb[0x1];
	uint8_t		functional_lb[0x1];
	uint8_t		port[0x1];
	uint8_t		l3_type[0x2];
	uint8_t		reserved_at_66[0x6];
	uint8_t		first_priority[0x3];
	uint8_t		first_cfi[0x1];
	uint8_t		first_vlan_qualifier[0x2];
	uint8_t		reserved_at_72[0x2];
	uint8_t		first_vlan_id[0xc];
};

/* TAG of lookup type ethernet L3 IPv4 5 tuple */
struct dr_ste_struct_eth_l3_ipv4_5_tuple_bits {
	uint8_t		destination_address[0x20];
	uint8_t		source_address[0x20];
	uint8_t		source_port[0x10];
	uint8_t		destination_port[0x10];
	uint8_t		fragmented[0x1];
	uint8_t		first_fragment[0x1];
	uint8_t		reserved_at_62[0x2];
	uint8_t		reserved_at_64[0x1];
	uint8_t		ecn[0x2];
	uint8_t		tcp_ns[0x1];
	uint8_t		tcp_cwr[0x1];
	uint8_t		tcp_ece[0x1];
	uint8_t		tcp_urg[0x1];
	uint8_t		tcp_ack[0x1];
	uint8_t		tcp_psh[0x1];
	uint8_t		tcp_rst[0x1];
	uint8_t		tcp_syn[0x1];
	uint8_t		tcp_fin[0x1];
	uint8_t		dscp[0x6];
	uint8_t		reserved_at_76[0x2];
	uint8_t		protocol[0x8];
};

/* TAG of lookup type ethernet L3 IPv6 dst */
struct dr_ste_struct_eth_l3_ipv6_dst_bits {
	uint8_t		dst_ip_127_96[0x20];
	uint8_t		dst_ip_95_64[0x20];
	uint8_t		dst_ip_63_32[0x20];
	uint8_t		dst_ip_31_0[0x20];
};

/* TAG of lookup type ethernet L2 tnl */
struct dr_ste_struct_eth_l2_tnl_bits {
	uint8_t		dmac_47_16[0x20];
	uint8_t		dmac_15_0[0x10];
	uint8_t		l3_ethertype[0x10];
	uint8_t		l2_tunneling_network_id[0x20];
	uint8_t		ip_fragmented[0x1];
	uint8_t		tcp_syn[0x1];
	uint8_t		encp_type[0x2];
	uint8_t		l3_type[0x2];
	uint8_t		l4_type[0x2];
	uint8_t		first_priority[0x3];
	uint8_t		first_cfi[0x1];
	uint8_t		reserved_at_6c[0x3];
	uint8_t		gre_key_flag[0x1];
	uint8_t		first_vlan_qualifier[0x2];
	uint8_t		reserved_at_72[0x2];
	uint8_t		first_vlan_id[0xc];
};

/* TAG of lookup type ethernet L3 IPv6 src */
struct dr_ste_struct_eth_l3_ipv6_src_bits {
	uint8_t		src_ip_127_96[0x20];
	uint8_t		src_ip_95_64[0x20];
	uint8_t		src_ip_63_32[0x20];
	uint8_t		src_ip_31_0[0x20];
};

/* TAG of lookup type ethernet L3 IPv4 misc */
struct dr_ste_struct_eth_l3_ipv4_misc_bits {
	uint8_t		version[0x4];
	uint8_t		ihl[0x4];
	uint8_t		reserved_at_8[0x8];
	uint8_t		total_length[0x10];
	uint8_t		identification[0x10];
	uint8_t		flags[0x3];
	uint8_t		fragment_offset[0xd];
	uint8_t		time_to_live[0x8];
	uint8_t		reserved_at_48[0x8];
	uint8_t		checksum[0x10];
	uint8_t		reserved_at_60[0x20];
};

/* TAG of lookup type ethernet l4 */
struct dr_ste_struct_eth_l4_bits {
	uint8_t		fragmented[0x1];
	uint8_t		first_fragment[0x1];
	uint8_t		reserved_at_2[0x6];
	uint8_t		protocol[0x8];
	uint8_t		dst_port[0x10];
	uint8_t		ipv6_version[0x4];
	uint8_t		reserved_at_24[0x1];
	uint8_t		ecn[0x2];
	uint8_t		tcp_ns[0x1];
	uint8_t		tcp_cwr[0x1];
	uint8_t		tcp_ece[0x1];
	uint8_t		tcp_urg[0x1];
	uint8_t		tcp_ack[0x1];
	uint8_t		tcp_psh[0x1];
	uint8_t		tcp_rst[0x1];
	uint8_t		tcp_syn[0x1];
	uint8_t		tcp_fin[0x1];
	uint8_t		src_port[0x10];
	uint8_t		ipv6_payload_length[0x10];
	uint8_t		ipv6_hop_limit[0x8];
	uint8_t		dscp[0x6];
	uint8_t		reserved_at_5e[0x2];
	uint8_t		tcp_data_offset[0x4];
	uint8_t		reserved_at_64[0x8];
	uint8_t		flow_label[0x14];
};

/* TAG of lookup type ethernet l4 misc */
struct dr_ste_struct_eth_l4_misc_bits {
	uint8_t		checksum[0x10];
	uint8_t		length[0x10];
	uint8_t		seq_num[0x20];
	uint8_t		ack_num[0x20];
	uint8_t		urgent_pointer[0x10];
	uint8_t		window_size[0x10];
};

/* TAG of lookup type mpls */
struct dr_ste_struct_mpls_bits {
	uint8_t		mpls0_label[0x14];
	uint8_t		mpls0_exp[0x3];
	uint8_t		mpls0_s_bos[0x1];
	uint8_t		mpls0_ttl[0x8] ;
	uint8_t		mpls1_label[0x20];
	uint8_t		mpls2_label[0x20];
	uint8_t		reserved_at_60[0x16];
	uint8_t		mpls4_s_bit[0x1];
	uint8_t		mpls4_qualifier[0x1];
	uint8_t		mpls3_s_bit[0x1];
	uint8_t		mpls3_qualifier[0x1];
	uint8_t		mpls2_s_bit[0x1];
	uint8_t		mpls2_qualifier[0x1];
	uint8_t		mpls1_s_bit[0x1];
	uint8_t		mpls1_qualifier[0x1];
	uint8_t		mpls0_s_bit[0x1];
	uint8_t		mpls0_qualifier[0x1];
};

/* TAG of lookup type_register_0 */
struct dr_ste_struct_register_0_bits {
	uint8_t		register_0_h[0x20];
	uint8_t		register_0_l[0x20];
	uint8_t		register_1_h[0x20];
	uint8_t		register_1_l[0x20];
};

/* TAG of lookup type_register_1 */
struct dr_ste_struct_register_1_bits {
	uint8_t		register_2_h[0x20];
	uint8_t		register_2_l[0x20];
	uint8_t		register_3_h[0x20];
	uint8_t		register_3_l[0x20];
};

/* TAG of lookup type gre */
struct dr_ste_struct_gre_bits {
	uint8_t		gre_c_present[0x1];
	uint8_t		reserved_at_30[0x1];
	uint8_t		gre_k_present[0x1];
	uint8_t		gre_s_present[0x1];
	uint8_t		strict_src_route[0x1];
	uint8_t		recur[0x3];
	uint8_t		flags[0x5];
	uint8_t		version[0x3];
	uint8_t		gre_protocol[0x10];
	uint8_t		checksum[0x10];
	uint8_t		offset[0x10];
	uint8_t		gre_key_h[0x18];
	uint8_t		gre_key_l[0x8];
	uint8_t		seq_num[0x20];
};
/* TAG of lookup type flex_parser_0 */
struct dr_ste_struct_flex_parser_0_bits {
	uint8_t		parser_3_label[0x14];
	uint8_t		parser_3_exp[0x3];
	uint8_t		parser_3_s_bos[0x1];
	uint8_t		parser_3_ttl[0x8] ;
	uint8_t		flex_parser_2[0x20];
	uint8_t		flex_parser_1[0x20];
	uint8_t		flex_parser_0[0x20];
};

/* TAG of lookup type flex_parser_1 */
struct dr_ste_struct_flex_parser_1_bits {
	uint8_t		flex_parser_7[0x20];
	uint8_t		flex_parser_6[0x20];
	uint8_t		flex_parser_5[0x20];
	uint8_t		flex_parser_4[0x20];
};

/* TAG of lookup type flex_parser_tnl */
struct dr_ste_struct_flex_parser_tnl_bits {
	uint8_t		flex_parser_tunneling_header_63_32[0x20];
	uint8_t		flex_parser_tunneling_header_31_0[0x20];
	uint8_t		reserved_at_40[0x40];
};

/* TAG of lookup type general_purpose field (metadata)*/
struct dr_ste_struct_general_purpose_bits {
	uint8_t		general_purpose_lookup_field[0x20];
	uint8_t		reserved_at_20[0x20];
	uint8_t		reserved_at_40[0x20];
	uint8_t		reserved_at_60[0x20];
};

/* TAG of lookup type source gvmi and qp*/
struct dr_ste_struct_src_gvmi_qp_bits {
	uint8_t		loopback_syndrome[0x8];
	uint8_t		reserved_at_8[0x8];
	uint8_t		source_gvmi[0x10];
	uint8_t		reserved_at_20[0x5];
	uint8_t		force_lb[0x1];
	uint8_t		functional_lb[0x1];
	uint8_t		source_is_requestor[0x1];
	uint8_t		source_qp[0x18];
	uint8_t		reserved_at_40[0x20];
	uint8_t		reserved_at_60[0x20];
};

struct dr_hw_ste_format {
	uint8_t ctrl[DR_STE_SIZE_CTRL];
	uint8_t tag[DR_STE_SIZE_TAG];
	uint8_t mask[DR_STE_SIZE_MASK];
};

static inline void dr_ste_calc_lookup_entry(uint32_t (*tbl)[256], uint8_t i,
					    uint8_t j)
{
	tbl[i][j] = (tbl[i-1][j] >> 8) ^ tbl[0][tbl[i-1][j] & 0xFF];
}

static void dr_ste_crc32_init_tab(void)
{
	uint32_t crc, i, j;

	for (i = 0; i < 256; i++) {
		crc = i;
		for (j = 0; j < 8; j++) {
			if (crc & 0x00000001L)
				crc = (crc >> 1) ^ DR_STE_CRC_POLY;
			else
				crc = crc >> 1;
		}
		dr_ste_crc_tab32[0][i] = crc;
	}

	/* init CRC lookup tables according to crc_slice_8 algorithm */
	for (i = 0; i < 256; i++) {
		dr_ste_calc_lookup_entry(dr_ste_crc_tab32, 1, i);
		dr_ste_calc_lookup_entry(dr_ste_crc_tab32, 2, i);
		dr_ste_calc_lookup_entry(dr_ste_crc_tab32, 3, i);
		dr_ste_calc_lookup_entry(dr_ste_crc_tab32, 4, i);
		dr_ste_calc_lookup_entry(dr_ste_crc_tab32, 5, i);
		dr_ste_calc_lookup_entry(dr_ste_crc_tab32, 6, i);
		dr_ste_calc_lookup_entry(dr_ste_crc_tab32, 7, i);
	}

	dr_ste_crc32_init = true;
}

/* compute CRC32 (Slicing-by-8 algorithm) */
static uint32_t dr_crc32_slice8_calc(const void *input_data, size_t length)
{
	const uint32_t* current = (const uint32_t*) input_data;
	const uint8_t* current_char;
	uint32_t crc = 0, one, two;

	if (!input_data)
		return 0;

	if (!dr_ste_crc32_init)
		dr_ste_crc32_init_tab();

	/* process eight bytes at once (Slicing-by-8) */
	while (length >= 8) {
		one = *current++ ^ crc;
		two = *current++;

		crc = dr_ste_crc_tab32[0][(two >> 24) & 0xFF]
		      ^ dr_ste_crc_tab32[1][(two >> 16)& 0xFF]
		      ^ dr_ste_crc_tab32[2][(two >> 8) & 0xFF]
		      ^ dr_ste_crc_tab32[3][two & 0xFF]
		      ^ dr_ste_crc_tab32[4][(one >> 24) & 0xFF]
		      ^ dr_ste_crc_tab32[5][(one >> 16) & 0xFF]
		      ^ dr_ste_crc_tab32[6][(one >> 8) & 0xFF]
		      ^ dr_ste_crc_tab32[7][one & 0xFF];

		length -= 8;
	}

	current_char = (const uint8_t*) current;
	/* remaining 1 to 7 bytes (standard algorithm) */
	while (length-- != 0)
		crc = (crc >> 8) ^ dr_ste_crc_tab32[0][(crc & 0xFF)
						       ^ *current_char++];

	return ((crc>>24) & 0xff) | ((crc<<8) & 0xff0000) |
		((crc>>8) & 0xff00) | ((crc<<24) & 0xff000000);
}

uint32_t dr_ste_calc_hash_index(uint8_t *hw_ste_p,
				struct dr_ste_htbl *htbl)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	uint8_t masked[DR_STE_SIZE_TAG] = {};
	uint32_t crc32, index;
	uint16_t bit;
	int i;

	/* Don't calculate CRC if the result is predicted */
	if (htbl->chunk->num_of_entries == 1 || htbl->byte_mask == 0)
		return 0;

	/* Mask tag using byte mask, bit per byte */
	bit = 1 << (DR_STE_SIZE_TAG - 1);
	for (i = 0; i < DR_STE_SIZE_TAG; i++) {
		if (htbl->byte_mask & bit)
			masked[i] = hw_ste->tag[i];

		bit = bit >> 1;
	}
	crc32 = dr_crc32_slice8_calc(masked, DR_STE_SIZE_TAG);
	index = crc32 % htbl->chunk->num_of_entries;

	return index;
}

static uint16_t dr_ste_conv_bit_to_byte_mask(uint8_t *bit_mask)
{
	uint16_t byte_mask = 0;
	int i;

	for (i = 0; i < DR_STE_SIZE_MASK; i++) {
		byte_mask = byte_mask << 1;
		if (bit_mask[i] == 0xff)
			byte_mask |= 1;
	}
	return byte_mask;
}

void dr_ste_set_bit_mask(uint8_t *hw_ste_p, uint8_t *bit_mask)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;

	memcpy(hw_ste->mask, bit_mask, DR_STE_SIZE_MASK);
}

void dr_ste_rx_set_flow_tag(uint8_t *hw_ste_p, uint32_t flow_tag)
{
	DR_STE_SET(rx_steering_mult, hw_ste_p, qp_list_pointer,
		   DR_STE_ENABLE_FLOW_TAG | flow_tag);
}

void dr_ste_set_counter_id(uint8_t *hw_ste_p, uint32_t ctr_id)
{
	/* This can be used for both rx_steering_mult and for sx_transmit */
	DR_STE_SET(rx_steering_mult, hw_ste_p, counter_trigger_15_0, ctr_id);
	DR_STE_SET(rx_steering_mult, hw_ste_p, counter_trigger_23_16, ctr_id >> 16);
}

void dr_ste_set_tx_push_vlan(uint8_t *hw_ste_p, uint32_t vlan_hdr)
{
	DR_STE_SET(sx_transmit, hw_ste_p, action_type,
		   DR_STE_ACTION_TYPE_PUSH_VLAN);
	DR_STE_SET(sx_transmit, hw_ste_p, encap_pointer_vlan_data, vlan_hdr);
}

void dr_ste_set_tx_encap(uint8_t *hw_ste_p, uint32_t reformat_id, int size, bool encap_l3)
{
	DR_STE_SET(sx_transmit, hw_ste_p, action_type,
		   encap_l3 ? DR_STE_ACTION_TYPE_ENCAP_L3 : DR_STE_ACTION_TYPE_ENCAP);
	/* the hardware expects here size in words (2 byte) */
	DR_STE_SET(sx_transmit, hw_ste_p, action_description, size / 2);
	DR_STE_SET(sx_transmit, hw_ste_p, encap_pointer_vlan_data, reformat_id);
}

void dr_ste_set_rx_decap(uint8_t *hw_ste_p)
{
	DR_STE_SET(rx_steering_mult, hw_ste_p, tunneling_action,
		   DR_STE_TUNL_ACTION_DECAP);
}

void dr_ste_set_rx_pop_vlan(uint8_t *hw_ste_p)
{
	DR_STE_SET(rx_steering_mult, hw_ste_p, tunneling_action,
		   DR_STE_TUNL_ACTION_POP_VLAN);
}

void dr_ste_set_rx_decap_l3(uint8_t *hw_ste_p, bool vlan)
{
	DR_STE_SET(rx_steering_mult, hw_ste_p, tunneling_action,
		   DR_STE_TUNL_ACTION_L3_DECAP);
	DR_STE_SET(modify_packet, hw_ste_p, action_description, vlan ? 1:0 );
}

void dr_ste_set_entry_type(uint8_t *hw_ste_p, uint8_t entry_type)
{
	DR_STE_SET(general, hw_ste_p, entry_type, entry_type);
}

uint8_t dr_ste_get_entry_type(uint8_t *hw_ste_p)
{
	return DR_STE_GET(general, hw_ste_p, entry_type);
}

void dr_ste_set_rewrite_actions(uint8_t *hw_ste_p, uint16_t num_of_actions,
				uint32_t re_write_index)
{
	DR_STE_SET(modify_packet, hw_ste_p, number_of_re_write_actions,
		   num_of_actions);
	DR_STE_SET(modify_packet, hw_ste_p, header_re_write_actions_pointer,
		   re_write_index);
}

void dr_ste_init(uint8_t *hw_ste_p, uint8_t lu_type, uint8_t entry_type,
		 uint16_t gvmi)
{
	DR_STE_SET(general, hw_ste_p, entry_type, entry_type);
	DR_STE_SET(general, hw_ste_p, entry_sub_type, lu_type);
	DR_STE_SET(general, hw_ste_p, next_lu_type, DR_STE_LU_TYPE_DONT_CARE);

	/* Set GVMI once, this is the same for RX/TX
	 * bits 63_48 of next table base / miss address encode the next GVMI */
	DR_STE_SET(rx_steering_mult, hw_ste_p, gvmi, gvmi);
	DR_STE_SET(rx_steering_mult, hw_ste_p, next_table_base_63_48, gvmi);
	DR_STE_SET(rx_steering_mult, hw_ste_p, miss_address_63_48, gvmi);
}

static void dr_ste_set_always_hit(struct dr_hw_ste_format *hw_ste)
{
	memset(&hw_ste->tag, 0, sizeof(hw_ste->tag));
	memset(&hw_ste->mask, 0, sizeof(hw_ste->mask));
}

static void dr_ste_set_always_miss(struct dr_hw_ste_format *hw_ste)
{
	hw_ste->tag[0] = 0xdc;
	hw_ste->mask[0] = 0;
}

uint64_t dr_ste_get_miss_addr(uint8_t *hw_ste)
{
	uint64_t index =
		(DR_STE_GET(rx_steering_mult, hw_ste, miss_address_31_6) |
		 DR_STE_GET(rx_steering_mult, hw_ste, miss_address_39_32) << 26);

	return index << 6;
}

void dr_ste_set_hit_addr(uint8_t *hw_ste, uint64_t icm_addr, uint32_t ht_size)
{
	uint64_t index = (icm_addr >> 5) | ht_size;

	DR_STE_SET(general, hw_ste, next_table_base_39_32_size, index >> 27);
	DR_STE_SET(general, hw_ste, next_table_base_31_5_size, index);
}

uint64_t dr_ste_get_icm_addr(struct dr_ste *ste)
{
	uint32_t index = ste - ste->htbl->ste_arr;

	return ste->htbl->chunk->icm_addr + DR_STE_SIZE * index;
}

uint64_t dr_ste_get_mr_addr(struct dr_ste *ste)
{
	uint32_t index = ste - ste->htbl->ste_arr;

	return ste->htbl->chunk->mr_addr + DR_STE_SIZE * index;
}

struct list_head *dr_ste_get_miss_list(struct dr_ste *ste)
{
	uint32_t index = ste - ste->htbl->ste_arr;

	return &ste->htbl->miss_list[index];
}

/* attention: the function assumes ste->hw_ste is 64B size */
void dr_ste_always_hit_htbl(struct dr_ste *ste, struct dr_ste_htbl *next_htbl)
{
	struct dr_icm_chunk *chunk = next_htbl->chunk;
	uint8_t *hw_ste = ste->hw_ste;

	DR_STE_SET(general, hw_ste, byte_mask, next_htbl->byte_mask);
	DR_STE_SET(general, hw_ste, next_lu_type, next_htbl->lu_type);
	dr_ste_set_hit_addr(hw_ste, chunk->icm_addr, chunk->num_of_entries);

	dr_ste_set_always_hit((struct dr_hw_ste_format *)ste->hw_ste);
}

bool dr_ste_is_last_in_rule(struct dr_matcher_rx_tx *nic_matcher,
			    uint8_t ste_location)
{
	return ste_location == nic_matcher->num_of_builders;
}

/*
 * replace relevant fields, except of:
 * htbl - keep the origin htbl
 * miss_list + list - already took the src from the list.
 * icm_addr/mr_addr - depeends on the hosting table.
 *
 * Before:
 * | a | -> | b | -> | c | ->
 *
 * After:
 * | a | -> | c | ->
 * While the data that was in b coppied to a.
 */
static void dr_ste_replace(struct dr_ste *dst, struct dr_ste *src)
{
	memcpy(dst->hw_ste, src->hw_ste, DR_STE_SIZE_REDUCED);
	dst->next_htbl = src->next_htbl;
	if (dst->next_htbl)
		dst->next_htbl->pointing_ste = dst;

	atomic_init(&dst->refcount, atomic_load(&src->refcount));

	list_head_init(&dst->rule_list);
	list_append_list(&dst->rule_list, &src->rule_list);
}

void dr_ste_free(struct dr_ste *ste,
		 struct mlx5dv_dr_matcher *matcher,
		 struct dr_matcher_rx_tx *nic_matcher)
{
	struct dr_ste_send_info *cur_ste_info, *tmp_ste_info;
	struct dr_ste_send_info ste_info_head, ste_info;
	struct dr_ste *next_ste, *prev_ste, *first_ste;
	struct mlx5dv_dr_ns *ns = matcher->tbl->ns;
	uint8_t tmp_data_ste[DR_STE_SIZE] = {0};
	LIST_HEAD(send_ste_list);
	uint64_t miss_addr;
	bool put_on_table = true;
	int ret;
	struct dr_ste_htbl *stats_tbl;

	/*
	 * 2 cases:
	 * 1. ste is head:
	 *	a. the only one in miss_list - free current
	 *	b. miss list exists - copy the first in the list to be instead
	 *	  of the origin ste.
	 * 2. part of miss_list:
	 *	update prev and current.
	 */
	first_ste = list_top(dr_ste_get_miss_list(ste), struct dr_ste, miss_list_node);
	stats_tbl = first_ste->htbl;

	if (first_ste == ste) { /* head */
		struct dr_ste_htbl *next_miss_htbl;

		next_ste = list_next(dr_ste_get_miss_list(ste), ste, miss_list_node);
		if (!next_ste) { /* one and only entry in the list */
			struct dr_ste tmp_ste = {};

			tmp_ste.hw_ste = tmp_data_ste;
			/*
			 * use temp ste because dr_ste_always_miss_addr
			 * touches bit_mask area which doesn't exist at
			 * ste->hw_ste.
			 * copy it back to ste->hw_ste after that.
			 */
			memcpy(tmp_ste.hw_ste, ste->hw_ste, DR_STE_SIZE_REDUCED);
			miss_addr = nic_matcher->e_anchor->chunk->icm_addr;
			dr_ste_always_miss_addr(&tmp_ste, miss_addr);
			memcpy(ste->hw_ste, tmp_ste.hw_ste, DR_STE_SIZE_REDUCED);

			list_del_init(&ste->miss_list_node);
			/*write full STE size in order to have "always_miss"*/
			dr_fill_and_append_ste_send_info(ste, DR_STE_SIZE,
							 0, tmp_data_ste,
							 &ste_info_head,
							 &send_ste_list,
							 false);

			stats_tbl->ctrl.num_of_valid_entries--;

			goto out_update;
		}
		/* |_ste_| --> |_next_ste_| -->|__| -->|__| --> */
		next_miss_htbl = next_ste->htbl;
		/* remove from the miss_list the next_ste before copy */
		list_del_init(&next_ste->miss_list_node);

		/* all rule-members that use next_ste should know about that */
		dr_rule_update_rule_member(next_ste, ste);
		/* move data from next into ste */
		dr_ste_replace(ste, next_ste);

		/*
		 * del the htbl that contains the ste that was "moved".
		 * no need to "put" the new table that the ste was moved to it.
		 */
		dr_htbl_put(next_miss_htbl);
		put_on_table = false;

		dr_fill_and_append_ste_send_info(ste, DR_STE_SIZE_REDUCED, 0,
						 ste->hw_ste,
						 &ste_info_head,
						 &send_ste_list, false);

		stats_tbl->ctrl.num_of_collisions--;
		stats_tbl->ctrl.num_of_valid_entries--;

		goto out_update;
	}

	prev_ste = list_prev(dr_ste_get_miss_list(ste), ste, miss_list_node);
	if (!prev_ste) {
		fprintf(stderr, "%s No previous list member\n", __func__);
		goto out_update;
	}

	/*|__| -->|_prev_ste_|->|_ste_|-->|_next_ste_|*/
	miss_addr = dr_ste_get_miss_addr(ste->hw_ste);
	dr_ste_set_miss_addr(prev_ste->hw_ste, miss_addr);

	dr_fill_and_append_ste_send_info(prev_ste, DR_STE_SIZE_REDUCED,
					 0, prev_ste->hw_ste,
					 &ste_info, &send_ste_list, false);

	list_del_init(&ste->miss_list_node);

	stats_tbl->ctrl.num_of_valid_entries--;
	stats_tbl->ctrl.num_of_collisions--;

out_update:
	/* update HW */
	list_for_each_safe(&send_ste_list, cur_ste_info, tmp_ste_info, send_list) {
		list_del(&cur_ste_info->send_list);
		ret = dr_postsend_ste(ns, cur_ste_info->ste,
				      cur_ste_info->data, cur_ste_info->size,
				      cur_ste_info->offset);
		if (ret)
			printf("%s %d Failed sending ste!!!\n",
			       __func__, __LINE__);
	}

	if (put_on_table)
		dr_htbl_put(ste->htbl);
}

bool dr_ste_equal_tag(void *src, void *dst)
{
	struct dr_hw_ste_format *s_hw_ste = (struct dr_hw_ste_format *)src;
	struct dr_hw_ste_format *d_hw_ste = (struct dr_hw_ste_format *)dst;

	return (!memcmp(s_hw_ste->tag, d_hw_ste->tag, DR_STE_SIZE_TAG));
}

void dr_ste_set_hit_addr_by_next_htbl(uint8_t *hw_ste,
				      struct dr_ste_htbl *next_htbl)
{
	struct dr_icm_chunk *chunk = next_htbl->chunk;

	dr_ste_set_hit_addr(hw_ste, chunk->icm_addr, chunk->num_of_entries);
}

void dr_ste_set_miss_addr(uint8_t *hw_ste_p, uint64_t miss_addr)
{
	uint64_t index = miss_addr >> 6;

	/* Miss address for TX and RX STEs located in the same offsets */
	DR_STE_SET(rx_steering_mult, hw_ste_p, miss_address_39_32, index >> 26);
	DR_STE_SET(rx_steering_mult, hw_ste_p, miss_address_31_6, index);
}

void dr_ste_always_miss_addr(struct dr_ste *ste, uint64_t miss_addr)
{
	uint8_t *hw_ste = ste->hw_ste;

	DR_STE_SET(rx_steering_mult, hw_ste, next_lu_type, DR_STE_LU_TYPE_DONT_CARE);
	dr_ste_set_miss_addr(hw_ste, miss_addr);
	dr_ste_set_always_miss((struct dr_hw_ste_format *)ste->hw_ste);
}

/*
 * The assumption here is that we don't update the ste->hw_ste if it is not
 * used ste, so it will be all zero, checking the next_lu_type.
 * TBD: will add something in reserved field instead ?
 */
bool dr_ste_is_not_valid_entry(uint8_t *p_hw_ste)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)p_hw_ste;

	if (DR_STE_GET(general, hw_ste, next_lu_type) ==
	    DR_STE_LU_TYPE_NOP)
		return true;

	return false;
}

bool dr_ste_not_used_ste(struct dr_ste *ste)
{
	return !atomic_load(&ste->refcount);
}

static inline uint16_t get_bits_per_mask(uint16_t byte_mask)
{
	uint16_t bits = 0;

	while (byte_mask) {
		byte_mask = byte_mask & (byte_mask - 1);
		bits++;
	}

	return bits;
}

void dr_ste_set_formated_ste(struct mlx5dv_dr_ns *ns,
			     struct dr_ns_rx_tx *nic_ns,
			     struct dr_ste_htbl *htbl,
			     uint8_t *formated_ste,
			     struct dr_htbl_connect_info *connect_info)
{
	struct dr_ste ste = {};

	/* init tmp ste for init data array, the ste is initialized and */
	dr_ste_init(formated_ste, htbl->lu_type, nic_ns->ste_type,
		    ns->info.caps.gvmi);
	ste.hw_ste = formated_ste;

	switch (connect_info->type) {
	case CONNECT_HIT:
		dr_ste_always_hit_htbl(&ste, connect_info->hit_next_htbl);
		break;
	case CONNECT_MISS:
		dr_ste_always_miss_addr(&ste, connect_info->miss_icm_addr);
		break;
	default:
		fprintf(stderr, "%s unknown value: %d\n",
			__func__, connect_info->type);
	}
}

int dr_ste_htbl_init_and_postsend(struct mlx5dv_dr_ns *ns,
				  struct dr_ns_rx_tx *nic_ns,
				  struct dr_ste_htbl *htbl,
				  struct dr_htbl_connect_info *connect_info,
				  bool update_hw_ste)
{
	uint8_t formated_ste[DR_STE_SIZE] = {0};

	dr_ste_set_formated_ste(ns, nic_ns, htbl, formated_ste, connect_info);

	return dr_postsend_formated_htbl(ns, htbl, formated_ste, update_hw_ste);
}

#define BITS_IN_BYTE 8

int dr_ste_create_next_htbl(struct mlx5dv_dr_matcher *matcher,
			    struct dr_matcher_rx_tx *nic_matcher,
			    struct dr_ste *ste,
			    uint8_t *cur_hw_ste,
			    uint32_t table_size)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)cur_hw_ste;
	struct dr_ns_rx_tx *nic_ns = nic_matcher->nic_tbl->nic_ns;
	struct mlx5dv_dr_ns *ns = matcher->tbl->ns;
	struct dr_htbl_connect_info info;
	struct dr_ste_htbl *next_htbl;
	uint16_t byte_mask;
	uint8_t next_lu_type;

	next_lu_type = DR_STE_GET(general, hw_ste, next_lu_type);
	byte_mask = DR_STE_GET(general, hw_ste, byte_mask);

	if (!dr_ste_is_last_in_rule(nic_matcher, ste->ste_chain_location)) {
		uint32_t bits_in_mask;

		/* don't allocate table more than required */
		bits_in_mask = get_bits_per_mask(byte_mask) * BITS_IN_BYTE;
		table_size = min(table_size, bits_in_mask);

		next_htbl = dr_ste_htbl_alloc(ns->ste_icm_pool,
					      table_size,
					      next_lu_type,
					      byte_mask);
		if (!next_htbl) {
			printf("%s failed allocating table\n", __func__);
			return ENOMEM;
		}

		/* write new table to HW */
		info.type = CONNECT_MISS;
		info.miss_icm_addr = nic_matcher->e_anchor->chunk->icm_addr;
		if (dr_ste_htbl_init_and_postsend(ns, nic_ns, next_htbl,
						  &info, false)) {
			printf("%s failed writing table to HW\n", __func__);
			goto free_table;
		}
		/*
		 * TBD: performance improvment:
		 * The current code, write twice on each table, one for the table and one for the ste,
		 * we need to do that once.
		 * 2 options:
		 * 1. write htbl instead of ste's
		 * 2. better: send one write requests that contains all the relevant writes for this rule.
		*/
		dr_ste_set_hit_addr_by_next_htbl(cur_hw_ste, next_htbl);
		ste->next_htbl = next_htbl;
		next_htbl->pointing_ste = ste;
	}

	return 0;

free_table:
	dr_ste_htbl_free(next_htbl);
	return ENOENT;
}

static void dr_ste_set_ctrl(struct dr_ste_htbl *htbl)
{
	struct dr_ste_htbl_ctrl *ctrl = &htbl->ctrl;

	htbl->ctrl.may_grow = true;
	htbl->ctrl.may_shrink = true;
	if (htbl->chunk_size == DR_CHUNK_SIZE_MAX - 1) {
		htbl->ctrl.may_grow = false;
		return;
	}

	if (htbl->chunk_size == DR_CHUNK_SIZE_MIN) {
		htbl->ctrl.may_shrink = false;
		return;
	}

	switch (htbl->chunk_size) {
	case DR_CHUNK_SIZE_1:
		ctrl->increase_threashold = 1;
		break;
	case DR_CHUNK_SIZE_2:
		ctrl->increase_threashold = 1;
		ctrl->decrease_threashold = 0;
		break;
	case DR_CHUNK_SIZE_4:
		ctrl->increase_threashold = 2;
		ctrl->decrease_threashold = 0;
		break;
	case DR_CHUNK_SIZE_8:
		ctrl->increase_threashold = 4;
		ctrl->decrease_threashold = 1;
		break;
	case DR_CHUNK_SIZE_16:
		ctrl->increase_threashold = 8;
		ctrl->decrease_threashold = 1;
		break;
	case DR_CHUNK_SIZE_32:
		ctrl->increase_threashold = 1 << 4;
		ctrl->decrease_threashold = 1 << 1;
		break;
	case DR_CHUNK_SIZE_64:
		ctrl->increase_threashold = 1 << 5;
		ctrl->decrease_threashold = 1 << 2;
		break;
	case DR_CHUNK_SIZE_128:
		ctrl->increase_threashold = 1 << 6;
		ctrl->decrease_threashold = 1 << 3;
		break;
	case DR_CHUNK_SIZE_256:
		ctrl->increase_threashold = 1 << 7;
		ctrl->decrease_threashold = 1 << 4;
		break;
	case DR_CHUNK_SIZE_512:
		ctrl->increase_threashold = 1 << 8;
		ctrl->decrease_threashold = 1 << 5;
		break;
	case DR_CHUNK_SIZE_1K:
		ctrl->increase_threashold = 1 << 9;
		ctrl->decrease_threashold = 1 << 6;
		break;
	case DR_CHUNK_SIZE_2K:
		ctrl->increase_threashold = 1 << 10;
		ctrl->decrease_threashold = 1 << 7;
		break;
	case DR_CHUNK_SIZE_4K:
		ctrl->increase_threashold = 1 << 11;
		ctrl->decrease_threashold = 1 << 8;
		break;
	case DR_CHUNK_SIZE_8K:
		ctrl->increase_threashold = 1 << 12;
		ctrl->decrease_threashold = 1 << 9;
		break;
	case DR_CHUNK_SIZE_16K:
		ctrl->increase_threashold = 1 << 13;
		ctrl->decrease_threashold = 1 << 10;
		break;
	case DR_CHUNK_SIZE_32K:
		ctrl->increase_threashold = 1 << 14;
		ctrl->decrease_threashold = 1 << 11;
		break;
	case DR_CHUNK_SIZE_64K:
		ctrl->increase_threashold = 1 << 15;
		ctrl->decrease_threashold = 1 << 12;
		break;
	case DR_CHUNK_SIZE_128K:
		ctrl->increase_threashold = 1 << 16;
		ctrl->decrease_threashold = 1 << 13;
		break;
	case DR_CHUNK_SIZE_256K:
		ctrl->increase_threashold = 1 << 17;
		ctrl->decrease_threashold = 1 << 14;
		break;
	case DR_CHUNK_SIZE_512K:
		ctrl->increase_threashold = 1 << 18;
		ctrl->decrease_threashold = 1 << 15;
		break;
	case DR_CHUNK_SIZE_1024K:
		ctrl->increase_threashold = 1 << 19;
		ctrl->decrease_threashold = 1 << 16;
		break;
	case DR_CHUNK_SIZE_2048K:
		ctrl->increase_threashold = 1 << 20;
		ctrl->decrease_threashold = 1 << 17;
		break;
	default:
		assert(true);
	}
}

struct dr_ste_htbl *dr_ste_htbl_alloc(struct dr_icm_pool *pool,
				      enum dr_icm_chunk_size chunk_size,
				      uint8_t lu_type, uint16_t byte_mask)
{
	struct dr_icm_chunk *chunk;
	struct dr_ste_htbl *htbl;
	int i;

	htbl = calloc(1, sizeof(struct dr_ste_htbl));
	if (!htbl)
		return NULL;

	chunk = dr_icm_alloc_mem(pool, chunk_size);
	if (!chunk)
		goto out_free_htbl;

	htbl->chunk = chunk;
	htbl->lu_type = lu_type;
	htbl->byte_mask = byte_mask;
	htbl->ste_arr = chunk->ste_arr;
	htbl->hw_ste_arr = chunk->hw_ste_arr;
	htbl->miss_list = chunk->miss_list;
	atomic_init(&htbl->refcount, 0);

	for (i = 0; i < chunk->num_of_entries; i++) {
		struct dr_ste *ste = &htbl->ste_arr[i];

		ste->hw_ste = htbl->hw_ste_arr + i * DR_STE_SIZE_REDUCED;
		ste->htbl = htbl;
		atomic_init(&ste->refcount, 0);
		list_node_init(&ste->miss_list_node);
		list_head_init(&htbl->miss_list[i]);
		list_head_init(&ste->rule_list);
	}

	htbl->chunk_size = chunk_size;
	dr_ste_set_ctrl(htbl);
	return htbl;

out_free_htbl:
	free(htbl);
	return NULL;
}

int dr_ste_htbl_free(struct dr_ste_htbl *htbl)
{
	if (atomic_load(&htbl->refcount))
		return EBUSY;

	dr_icm_free_mem(htbl->chunk);
	free(htbl);
	return 0;
}

static int dr_ste_build_pre_check_spec(struct dr_match_spec *m_spec,
				       struct dr_match_spec *v_spec)
{
	if (m_spec->ip_version) {
		if (m_spec->ip_version != 4 && m_spec->ip_version != 6) {
			fprintf(stderr, "IP version must be specified v4 or v6\n");
			return EINVAL;
		}

		if (v_spec && (v_spec->ip_version != m_spec->ip_version)) {
			fprintf(stderr, "mask and value IP version must be equal\n");
			return EINVAL;
		}
	}
	return 0;
}

int dr_ste_build_pre_check(uint8_t match_criteria,
			   struct dr_match_param *mask,
			   struct dr_match_param *value)
{
	if (match_criteria & DR_MATCHER_CRITERIA_OUTER) {
		if (dr_ste_build_pre_check_spec(&mask->outer,
						value ? &value->outer : NULL))
			return EINVAL;
	}

	if (match_criteria & DR_MATCHER_CRITERIA_INNER) {
		if (dr_ste_build_pre_check_spec(&mask->inner,
						value ? &value->inner : NULL))
			return EINVAL;
	}
	return 0;
}

int dr_ste_build_ste_arr(struct mlx5dv_dr_matcher *matcher,
			 struct dr_matcher_rx_tx *nic_matcher,
			 struct dr_match_param *value,
			 uint8_t *ste_arr)
{
	struct dr_ns_rx_tx *nic_ns = nic_matcher->nic_tbl->nic_ns;
	struct mlx5dv_dr_ns *ns = matcher->tbl->ns;
	int ret, i;

	ret = dr_ste_build_pre_check(matcher->match_criteria,
				     &matcher->mask,
				     value);
	if (ret)
		return ret;

	for (i = 0; i < nic_matcher->num_of_builders; i++) {
		struct dr_ste_build *sb = &nic_matcher->ste_builder[i];

		dr_ste_init(ste_arr,
			    sb->lu_type,
			    nic_ns->ste_type,
			    ns->info.caps.gvmi);
		dr_ste_set_bit_mask(ste_arr, sb->bit_mask);

		ret = sb->ste_build_tag_func(value, sb, ste_arr);
		if (ret)
			return ret;

		/* Connect the STEs */
		if (i < (nic_matcher->num_of_builders - 1)) {
			sb++;
			DR_STE_SET(general, ste_arr, next_lu_type, sb->lu_type);
			DR_STE_SET(general, ste_arr, byte_mask, sb->byte_mask);
		}
		ste_arr += DR_STE_SIZE;
	}
	return 0;
}

static int dr_ste_build_eth_l2_src_des_bit_mask(struct dr_match_param *value,
						bool inner, uint8_t *bit_mask)
{
	struct dr_match_spec *mask = inner ? &value->inner : &value->outer;

	DR_STE_SET_MASK_V(eth_l2_src_dst, bit_mask, dmac_47_16, mask, dmac_47_16);
	DR_STE_SET_MASK_V(eth_l2_src_dst, bit_mask, dmac_15_0, mask, dmac_15_0);

	if (mask->smac_47_16 || mask->smac_15_0) {
		DR_STE_SET(eth_l2_src_dst, bit_mask, smac_47_32,
			   mask->smac_47_16 >> 16);
		DR_STE_SET(eth_l2_src_dst, bit_mask, smac_31_0,
			   mask->smac_47_16 << 16 | mask->smac_15_0);
		mask->smac_47_16 = 0;
		mask->smac_15_0 = 0;
	}

	DR_STE_SET_MASK_V(eth_l2_src_dst, bit_mask, first_vlan_id, mask, first_vid);
	DR_STE_SET_MASK_V(eth_l2_src_dst, bit_mask, first_cfi, mask, first_cfi);
	DR_STE_SET_MASK_V(eth_l2_src_dst, bit_mask, first_priority, mask, first_prio);
	DR_STE_SET_MASK(eth_l2_src_dst, bit_mask, l3_type, mask, ip_version);

	if (mask->cvlan_tag) {
		DR_STE_SET(eth_l2_src_dst, bit_mask, first_vlan_qualifier, -1);
		mask->cvlan_tag = 0;
	} else if (mask->svlan_tag) {
		DR_STE_SET(eth_l2_src_dst, bit_mask, first_vlan_qualifier, -1);
		mask->svlan_tag = 0;
	}

	if (mask->cvlan_tag || mask->svlan_tag) {
		fprintf(stderr, "Invalid c/svlan mask configuration\n");
		return EINVAL;
	}

	return 0;
}

static int dr_ste_build_eth_l2_src_des_tag(struct dr_match_param *value,
					   struct dr_ste_build *sb,
					   uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_spec *spec = sb->inner ? &value->inner : &value->outer;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(eth_l2_src_dst, tag, dmac_47_16, spec, dmac_47_16);
	DR_STE_SET_TAG(eth_l2_src_dst, tag, dmac_15_0, spec, dmac_15_0);

	if (spec->smac_47_16 || spec->smac_15_0) {
		DR_STE_SET(eth_l2_src_dst, tag, smac_47_32,
			   spec->smac_47_16 >> 16);
		DR_STE_SET(eth_l2_src_dst, tag, smac_31_0,
			   spec->smac_47_16 << 16 | spec->smac_15_0);
		spec->smac_47_16 = 0;
		spec->smac_15_0 = 0;
	}

	if (spec->ip_version) {
		if (spec->ip_version == IP_VERSION_IPV4) {
			DR_STE_SET(eth_l2_src_dst, tag, l3_type, STE_IPV4);
			spec->ip_version = 0;
		} else if (spec->ip_version == IP_VERSION_IPV6) {
			DR_STE_SET(eth_l2_src_dst, tag, l3_type, STE_IPV6);
			spec->ip_version = 0;
		} else {
			fprintf(stderr, "Unsupported ip_version value\n");
			return EINVAL;
		}
	}

	DR_STE_SET_TAG(eth_l2_src_dst, tag, first_vlan_id, spec, first_vid);
	DR_STE_SET_TAG(eth_l2_src_dst, tag, first_cfi, spec, first_cfi);
	DR_STE_SET_TAG(eth_l2_src_dst, tag, first_priority, spec, first_prio);

	if (spec->cvlan_tag) {
		DR_STE_SET(eth_l2_src_dst, tag, first_vlan_qualifier, STE_CVLAN);
		spec->cvlan_tag = 0;
	} else if (spec->svlan_tag) {
		DR_STE_SET(eth_l2_src_dst, tag, first_vlan_qualifier, STE_SVLAN);
		spec->svlan_tag = 0;
	}
	return 0;
}

int dr_ste_build_eth_l2_src_des(struct dr_ste_build *sb,
				struct dr_match_param *mask,
				bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_eth_l2_src_des_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_CALC_LU_TYPE(ETHL2_SRC_DST, rx, inner);
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_eth_l2_src_des_tag;

	return 0;
}

static int dr_ste_build_eth_l3_ipv6_dst_bit_mask(struct dr_match_param *value,
						 bool inner, uint8_t *bit_mask)
{
	struct dr_match_spec *mask = inner ? &value->inner : &value->outer;

	DR_STE_SET_MASK_V(eth_l3_ipv6_dst, bit_mask, dst_ip_127_96, mask, dst_ip_127_96);
	DR_STE_SET_MASK_V(eth_l3_ipv6_dst, bit_mask, dst_ip_95_64, mask, dst_ip_95_64);
	DR_STE_SET_MASK_V(eth_l3_ipv6_dst, bit_mask, dst_ip_63_32, mask, dst_ip_63_32);
	DR_STE_SET_MASK_V(eth_l3_ipv6_dst, bit_mask, dst_ip_31_0, mask, dst_ip_31_0);

	return 0;
}

static int dr_ste_build_eth_l3_ipv6_dst_tag(struct dr_match_param *value,
					    struct dr_ste_build *sb,
					    uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_spec *spec = sb->inner ? &value->inner : &value->outer;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(eth_l3_ipv6_dst, tag, dst_ip_127_96, spec, dst_ip_127_96);
	DR_STE_SET_TAG(eth_l3_ipv6_dst, tag, dst_ip_95_64, spec, dst_ip_95_64);
	DR_STE_SET_TAG(eth_l3_ipv6_dst, tag, dst_ip_63_32, spec, dst_ip_63_32);
	DR_STE_SET_TAG(eth_l3_ipv6_dst, tag, dst_ip_31_0, spec, dst_ip_31_0);

	return 0;
}

int dr_ste_build_eth_l3_ipv6_dst(struct dr_ste_build *sb,
				struct dr_match_param *mask,
				bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_eth_l3_ipv6_dst_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_CALC_LU_TYPE(ETHL3_IPV6_DST, rx, inner);
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_eth_l3_ipv6_dst_tag;

	return 0;
}

static int dr_ste_build_eth_l3_ipv6_src_bit_mask(struct dr_match_param *value,
						 bool inner, uint8_t *bit_mask)
{
	struct dr_match_spec *mask = inner ? &value->inner : &value->outer;

	DR_STE_SET_MASK_V(eth_l3_ipv6_src, bit_mask, src_ip_127_96, mask, src_ip_127_96);
	DR_STE_SET_MASK_V(eth_l3_ipv6_src, bit_mask, src_ip_95_64, mask, src_ip_95_64);
	DR_STE_SET_MASK_V(eth_l3_ipv6_src, bit_mask, src_ip_63_32, mask, src_ip_63_32);
	DR_STE_SET_MASK_V(eth_l3_ipv6_src, bit_mask, src_ip_31_0, mask, src_ip_31_0);

	return 0;
}

static int dr_ste_build_eth_l3_ipv6_src_tag(struct dr_match_param *value,
					    struct dr_ste_build *sb,
					    uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_spec *spec = sb->inner ? &value->inner : &value->outer;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(eth_l3_ipv6_src, tag, src_ip_127_96, spec, src_ip_127_96);
	DR_STE_SET_TAG(eth_l3_ipv6_src, tag, src_ip_95_64, spec, src_ip_95_64);
	DR_STE_SET_TAG(eth_l3_ipv6_src, tag, src_ip_63_32, spec, src_ip_63_32);
	DR_STE_SET_TAG(eth_l3_ipv6_src, tag, src_ip_31_0, spec, src_ip_31_0);

	return 0;
}

int dr_ste_build_eth_l3_ipv6_src(struct dr_ste_build *sb,
				struct dr_match_param *mask,
				bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_eth_l3_ipv6_src_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_CALC_LU_TYPE(ETHL3_IPV6_SRC, rx, inner);
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_eth_l3_ipv6_src_tag;

	return 0;
}

static int dr_ste_build_eth_l3_ipv4_5_tuple_bit_mask(struct dr_match_param *value,
						     bool inner,
						     uint8_t *bit_mask)
{
	struct dr_match_spec *mask = inner ? &value->inner : &value->outer;

	DR_STE_SET_MASK_V(eth_l3_ipv4_5_tuple, bit_mask, destination_address, mask, dst_ip_31_0);
	DR_STE_SET_MASK_V(eth_l3_ipv4_5_tuple, bit_mask, source_address, mask, src_ip_31_0);
	DR_STE_SET_MASK_V(eth_l3_ipv4_5_tuple, bit_mask, destination_port, mask, tcp_dport);
	DR_STE_SET_MASK_V(eth_l3_ipv4_5_tuple, bit_mask, destination_port, mask, udp_dport);
	DR_STE_SET_MASK_V(eth_l3_ipv4_5_tuple, bit_mask, source_port, mask, tcp_sport);
	DR_STE_SET_MASK_V(eth_l3_ipv4_5_tuple, bit_mask, source_port, mask, udp_sport);
	DR_STE_SET_MASK_V(eth_l3_ipv4_5_tuple, bit_mask, protocol, mask, ip_protocol);
	DR_STE_SET_MASK_V(eth_l3_ipv4_5_tuple, bit_mask, fragmented, mask, frag);
	DR_STE_SET_MASK_V(eth_l3_ipv4_5_tuple, bit_mask, dscp, mask, ip_dscp);
	DR_STE_SET_MASK_V(eth_l3_ipv4_5_tuple, bit_mask, ecn, mask, ip_ecn);

	if (mask->tcp_flags) {
		DR_STE_SET_TCP_FLAGS(eth_l3_ipv4_5_tuple, bit_mask, mask);
		mask->tcp_flags = 0;
	}

	return 0;
}

static int dr_ste_build_eth_l3_ipv4_5_tuple_tag(struct dr_match_param *value,
						struct dr_ste_build *sb,
						uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_spec *spec = sb->inner ? &value->inner : &value->outer;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(eth_l3_ipv4_5_tuple, tag, destination_address, spec, dst_ip_31_0);
	DR_STE_SET_TAG(eth_l3_ipv4_5_tuple, tag, source_address, spec, src_ip_31_0);
	DR_STE_SET_TAG(eth_l3_ipv4_5_tuple, tag, destination_port, spec, tcp_dport);
	DR_STE_SET_TAG(eth_l3_ipv4_5_tuple, tag, destination_port, spec, udp_dport);
	DR_STE_SET_TAG(eth_l3_ipv4_5_tuple, tag, source_port, spec, tcp_sport);
	DR_STE_SET_TAG(eth_l3_ipv4_5_tuple, tag, source_port, spec, udp_sport);
	DR_STE_SET_TAG(eth_l3_ipv4_5_tuple, tag, protocol, spec, ip_protocol);
	DR_STE_SET_TAG(eth_l3_ipv4_5_tuple, tag, fragmented, spec, frag);
	DR_STE_SET_TAG(eth_l3_ipv4_5_tuple, tag, dscp, spec, ip_dscp);
	DR_STE_SET_TAG(eth_l3_ipv4_5_tuple, tag, ecn, spec, ip_ecn);

	if (spec->tcp_flags) {
		DR_STE_SET_TCP_FLAGS(eth_l3_ipv4_5_tuple, tag, spec);
		spec->tcp_flags = 0;
	}

	return 0;
}

int dr_ste_build_eth_l3_ipv4_5_tuple(struct dr_ste_build *sb,
				     struct dr_match_param *mask,
				     bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_eth_l3_ipv4_5_tuple_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_CALC_LU_TYPE(ETHL3_IPV4_5_TUPLE, rx, inner);
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_eth_l3_ipv4_5_tuple_tag;

	return 0;
}

static int
dr_ste_build_eth_l2_src_or_dst_bit_mask(struct dr_match_param *value,
					bool inner, uint8_t *bit_mask)
{
	struct dr_match_spec *mask = inner ? &value->inner : &value->outer;
	struct dr_match_misc *misc_mask = &value->misc;

	DR_STE_SET_MASK_V(eth_l2_src, bit_mask, first_vlan_id, mask, first_vid);
	DR_STE_SET_MASK_V(eth_l2_src, bit_mask, first_cfi, mask, first_cfi);
	DR_STE_SET_MASK_V(eth_l2_src, bit_mask, first_priority, mask, first_prio);
	DR_STE_SET_MASK_V(eth_l2_src, bit_mask, ip_fragmented, mask, frag);
	DR_STE_SET_MASK_V(eth_l2_src, bit_mask, l3_ethertype, mask, ethertype);
	DR_STE_SET_MASK(eth_l2_src, bit_mask, l3_type, mask, ip_version);

	if (mask->svlan_tag || mask->cvlan_tag) {
		DR_STE_SET(eth_l2_src, bit_mask, first_vlan_qualifier, -1);
		mask->cvlan_tag = 0;
		mask->svlan_tag = 0;
	}

	if (inner) {
		if (misc_mask->inner_second_cvlan_tag ||
		    misc_mask->inner_second_svlan_tag) {
			DR_STE_SET(eth_l2_src, bit_mask, second_vlan_qualifier, -1);
			misc_mask->inner_second_cvlan_tag = 0;
			misc_mask->inner_second_svlan_tag = 0;
		}

		DR_STE_SET_MASK_V(eth_l2_src, bit_mask, second_vlan_id, misc_mask, inner_second_vid);
		DR_STE_SET_MASK_V(eth_l2_src, bit_mask, second_cfi, misc_mask, inner_second_cfi);
		DR_STE_SET_MASK_V(eth_l2_src, bit_mask, second_priority, misc_mask, inner_second_prio);
	} else {
		if (misc_mask->outer_second_cvlan_tag ||
		    misc_mask->outer_second_svlan_tag) {
			DR_STE_SET(eth_l2_src, bit_mask, second_vlan_qualifier, -1);
			misc_mask->outer_second_cvlan_tag = 0;
			misc_mask->outer_second_svlan_tag = 0;
		}

		DR_STE_SET_MASK_V(eth_l2_src, bit_mask, second_vlan_id, misc_mask, outer_second_vid);
		DR_STE_SET_MASK_V(eth_l2_src, bit_mask, second_cfi, misc_mask, outer_second_cfi);
		DR_STE_SET_MASK_V(eth_l2_src, bit_mask, second_priority, misc_mask, outer_second_prio);
	}

	return 0;
}

static int dr_ste_build_eth_l2_src_or_dst_tag(struct dr_match_param *value,
					      bool inner, uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_spec *spec = inner ? &value->inner : &value->outer;
	struct dr_match_misc *misc_spec = &value->misc;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(eth_l2_src, tag, first_vlan_id, spec, first_vid);
	DR_STE_SET_TAG(eth_l2_src, tag, first_cfi, spec, first_cfi);
	DR_STE_SET_TAG(eth_l2_src, tag, first_priority, spec, first_prio);
	DR_STE_SET_TAG(eth_l2_src, tag, ip_fragmented, spec, frag);
	DR_STE_SET_TAG(eth_l2_src, tag, l3_ethertype, spec, ethertype);

	if (spec->ip_version) {
		if (spec->ip_version == IP_VERSION_IPV4) {
			DR_STE_SET(eth_l2_src, tag, l3_type, STE_IPV4);
			spec->ip_version = 0;
		} else if (spec->ip_version == IP_VERSION_IPV6) {
			DR_STE_SET(eth_l2_src, tag, l3_type, STE_IPV6);
			spec->ip_version = 0;
		} else {
			fprintf(stderr, "Unsupported ip_version value\n");
			return EINVAL;
		}
	}

	if (spec->cvlan_tag) {
		DR_STE_SET(eth_l2_src, tag, first_vlan_qualifier, STE_CVLAN);
		spec->cvlan_tag = 0;
	} else if (spec->svlan_tag) {
		DR_STE_SET(eth_l2_src, tag, first_vlan_qualifier, STE_SVLAN);
		spec->svlan_tag = 0;
	}

	if (inner) {
		if (misc_spec->inner_second_cvlan_tag) {
			DR_STE_SET(eth_l2_src, tag, second_vlan_qualifier, STE_CVLAN);
			misc_spec->inner_second_cvlan_tag = 0;
		} else if (misc_spec->inner_second_svlan_tag) {
			DR_STE_SET(eth_l2_src, tag, second_vlan_qualifier, STE_SVLAN);
			misc_spec->inner_second_svlan_tag = 0;
		}

		DR_STE_SET_TAG(eth_l2_src, tag, second_vlan_id, misc_spec, inner_second_vid);
		DR_STE_SET_TAG(eth_l2_src, tag, second_cfi, misc_spec, inner_second_cfi);
		DR_STE_SET_TAG(eth_l2_src, tag, second_priority, misc_spec, inner_second_prio);
	} else {
		if (misc_spec->outer_second_cvlan_tag) {
			DR_STE_SET(eth_l2_src, tag, second_vlan_qualifier, STE_CVLAN);
			misc_spec->outer_second_cvlan_tag = 0;
		} else if (misc_spec->outer_second_svlan_tag) {
			DR_STE_SET(eth_l2_src, tag, second_vlan_qualifier, STE_SVLAN);
			misc_spec->outer_second_svlan_tag = 0;
		}
		DR_STE_SET_TAG(eth_l2_src, tag, second_vlan_id, misc_spec, outer_second_vid);
		DR_STE_SET_TAG(eth_l2_src, tag, second_cfi, misc_spec, outer_second_cfi);
		DR_STE_SET_TAG(eth_l2_src, tag, second_priority, misc_spec, outer_second_prio);
	}

	return 0;
}

static int dr_ste_build_eth_l2_src_bit_mask(struct dr_match_param *value,
					    bool inner, uint8_t *bit_mask)
{
	struct dr_match_spec *mask = inner ? &value->inner : &value->outer;

	DR_STE_SET_MASK_V(eth_l2_src, bit_mask, smac_47_16, mask, smac_47_16);
	DR_STE_SET_MASK_V(eth_l2_src, bit_mask, smac_15_0, mask, smac_15_0);

	return dr_ste_build_eth_l2_src_or_dst_bit_mask(value, inner, bit_mask);
}

static int dr_ste_build_eth_l2_src_tag(struct dr_match_param *value,
				       struct dr_ste_build *sb,
				       uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_spec *spec = sb->inner ? &value->inner : &value->outer;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(eth_l2_src, tag, smac_47_16, spec, smac_47_16);
	DR_STE_SET_TAG(eth_l2_src, tag, smac_15_0, spec, smac_15_0);

	return dr_ste_build_eth_l2_src_or_dst_tag(value, sb->inner, hw_ste_p);
}

int dr_ste_build_eth_l2_src(struct dr_ste_build *sb,
				struct dr_match_param *mask,
				bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_eth_l2_src_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_CALC_LU_TYPE(ETHL2_SRC, rx, inner);
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_eth_l2_src_tag;

	return 0;
}

static int dr_ste_build_eth_l2_dst_bit_mask(struct dr_match_param *value,
					    bool inner, uint8_t *bit_mask)
{
	struct dr_match_spec *mask = inner ? &value->inner : &value->outer;

	DR_STE_SET_MASK_V(eth_l2_dst, bit_mask, dmac_47_16, mask, dmac_47_16);
	DR_STE_SET_MASK_V(eth_l2_dst, bit_mask, dmac_15_0, mask, dmac_15_0);

	return dr_ste_build_eth_l2_src_or_dst_bit_mask(value, inner, bit_mask);
}

static int dr_ste_build_eth_l2_dst_tag(struct dr_match_param *value,
				       struct dr_ste_build *sb,
				       uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_spec *spec = sb->inner ? &value->inner : &value->outer;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(eth_l2_dst, tag, dmac_47_16, spec, dmac_47_16);
	DR_STE_SET_TAG(eth_l2_dst, tag, dmac_15_0, spec, dmac_15_0);

	return dr_ste_build_eth_l2_src_or_dst_tag(value, sb->inner, hw_ste_p);
}

int dr_ste_build_eth_l2_dst(struct dr_ste_build *sb,
			    struct dr_match_param *mask,
			    bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_eth_l2_dst_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_CALC_LU_TYPE(ETHL2_DST, rx, inner);
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_eth_l2_dst_tag;

	return 0;
}

static int dr_ste_build_eth_l2_tnl_bit_mask(struct dr_match_param *value,
					    bool inner, uint8_t *bit_mask)
{
	struct dr_match_spec *mask = inner ? &value->inner : &value->outer;
	struct dr_match_misc *misc = &value->misc;

	DR_STE_SET_MASK_V(eth_l2_tnl, bit_mask, dmac_47_16, mask, dmac_47_16);
	DR_STE_SET_MASK_V(eth_l2_tnl, bit_mask, dmac_15_0, mask, dmac_15_0);
	DR_STE_SET_MASK_V(eth_l2_tnl, bit_mask, first_vlan_id, mask, first_vid);
	DR_STE_SET_MASK_V(eth_l2_tnl, bit_mask, first_cfi, mask, first_cfi);
	DR_STE_SET_MASK_V(eth_l2_tnl, bit_mask, first_priority, mask, first_prio);
	DR_STE_SET_MASK_V(eth_l2_tnl, bit_mask, ip_fragmented, mask, frag);
	DR_STE_SET_MASK_V(eth_l2_tnl, bit_mask, l3_ethertype, mask, ethertype);
	DR_STE_SET_MASK(eth_l2_tnl, bit_mask, l3_type, mask, ip_version);

	if (misc->vxlan_vni) {
		DR_STE_SET(eth_l2_tnl, bit_mask, l2_tunneling_network_id, (misc->vxlan_vni << 8));
		misc->vxlan_vni = 0;
	}

	if (mask->svlan_tag || mask->cvlan_tag) {
		DR_STE_SET(eth_l2_tnl, bit_mask, first_vlan_qualifier, -1);
		mask->cvlan_tag = 0;
		mask->svlan_tag = 0;
	}

	return 0;
}

static int dr_ste_build_eth_l2_tnl_tag(struct dr_match_param *value,
				       struct dr_ste_build *sb,
				       uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_spec *spec = sb->inner ? &value->inner : &value->outer;
	struct dr_match_misc *misc = &value->misc;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(eth_l2_tnl, tag, dmac_47_16, spec, dmac_47_16);
	DR_STE_SET_TAG(eth_l2_tnl, tag, dmac_15_0, spec, dmac_15_0);
	DR_STE_SET_TAG(eth_l2_tnl, tag, first_vlan_id, spec, first_vid);
	DR_STE_SET_TAG(eth_l2_tnl, tag, first_cfi, spec, first_cfi);
	DR_STE_SET_TAG(eth_l2_tnl, tag, ip_fragmented, spec, frag);
	DR_STE_SET_TAG(eth_l2_tnl, tag, first_priority, spec, first_prio);
	DR_STE_SET_TAG(eth_l2_tnl, tag, l3_ethertype, spec, ethertype);

	if (misc->vxlan_vni) {
		DR_STE_SET(eth_l2_tnl, tag, l2_tunneling_network_id,
			   (misc->vxlan_vni << 8));
		misc->vxlan_vni = 0;
	}

	if (spec->cvlan_tag) {
		DR_STE_SET(eth_l2_tnl, tag, first_vlan_qualifier, STE_CVLAN);
		spec->cvlan_tag = 0;
	} else if (spec->svlan_tag) {
		DR_STE_SET(eth_l2_tnl, tag, first_vlan_qualifier, STE_SVLAN);
		spec->svlan_tag = 0;
	}

	if (spec->ip_version) {
		if (spec->ip_version == IP_VERSION_IPV4) {
			DR_STE_SET(eth_l2_tnl, tag, l3_type, STE_IPV4);
			spec->ip_version = 0;
		} else if (spec->ip_version == IP_VERSION_IPV6) {
			DR_STE_SET(eth_l2_tnl, tag, l3_type, STE_IPV6);
			spec->ip_version = 0;
		} else {
			fprintf(stderr, "Unsupported ip_version value\n");
			return EINVAL;
		}
	}

	return 0;
}

int dr_ste_build_eth_l2_tnl(struct dr_ste_build *sb,
			    struct dr_match_param *mask, bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_eth_l2_tnl_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_LU_TYPE_ETHL2_TUNNELING_I;
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_eth_l2_tnl_tag;

	return 0;
}

static int dr_ste_build_eth_l3_ipv4_misc_bit_mask(struct dr_match_param *value,
						  bool inner, uint8_t *bit_mask)
{
	struct dr_match_spec *mask = inner ? &value->inner : &value->outer;

	DR_STE_SET_MASK_V(eth_l3_ipv4_misc, bit_mask, time_to_live, mask, ip_ttl_hoplimit);

	return 0;
}

static int dr_ste_build_eth_l3_ipv4_misc_tag(struct dr_match_param *value,
					     struct dr_ste_build *sb,
					     uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_spec *spec = sb->inner ? &value->inner : &value->outer;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(eth_l3_ipv4_misc, tag, time_to_live, spec, ip_ttl_hoplimit);

	return 0;
}

int dr_ste_build_eth_l3_ipv4_misc(struct dr_ste_build *sb,
				  struct dr_match_param *mask,
				  bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_eth_l3_ipv4_misc_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_CALC_LU_TYPE(ETHL3_IPV4_MISC, rx, inner);
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_eth_l3_ipv4_misc_tag;

	return 0;
}

static int dr_ste_build_ipv6_l3_l4_bit_mask(struct dr_match_param *value,
					    bool inner, uint8_t *bit_mask)
{
	struct dr_match_spec *mask = inner ? &value->inner : &value->outer;

	DR_STE_SET_MASK_V(eth_l4, bit_mask, dst_port, mask, tcp_dport);
	DR_STE_SET_MASK_V(eth_l4, bit_mask, src_port, mask, tcp_sport);
	DR_STE_SET_MASK_V(eth_l4, bit_mask, dst_port, mask, udp_dport);
	DR_STE_SET_MASK_V(eth_l4, bit_mask, src_port, mask, udp_sport);
	DR_STE_SET_MASK_V(eth_l4, bit_mask, protocol, mask, ip_protocol);
	DR_STE_SET_MASK_V(eth_l4, bit_mask, fragmented, mask, frag);
	DR_STE_SET_MASK_V(eth_l4, bit_mask, dscp, mask, ip_dscp);
	DR_STE_SET_MASK_V(eth_l4, bit_mask, ecn, mask, ip_ecn);
	DR_STE_SET_MASK_V(eth_l4, bit_mask, ipv6_hop_limit, mask, ip_ttl_hoplimit);

	// TODO flow_label is missing

	if (mask->tcp_flags) {
		DR_STE_SET_TCP_FLAGS(eth_l4, bit_mask, mask);
		mask->tcp_flags = 0;
	}

	return 0;
}

static int dr_ste_build_ipv6_l3_l4_tag(struct dr_match_param *value,
				       struct dr_ste_build *sb,
				       uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_spec *spec = sb->inner ? &value->inner : &value->outer;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(eth_l4, tag, dst_port, spec, tcp_dport);
	DR_STE_SET_TAG(eth_l4, tag, src_port, spec, tcp_sport);
	DR_STE_SET_TAG(eth_l4, tag, dst_port, spec, udp_dport);
	DR_STE_SET_TAG(eth_l4, tag, src_port, spec, udp_sport);
	DR_STE_SET_TAG(eth_l4, tag, protocol, spec, ip_protocol);
	DR_STE_SET_TAG(eth_l4, tag, fragmented, spec, frag);
	DR_STE_SET_TAG(eth_l4, tag, dscp, spec, ip_dscp);
	DR_STE_SET_TAG(eth_l4, tag, ecn, spec, ip_ecn);
	DR_STE_SET_TAG(eth_l4, tag, ipv6_hop_limit, spec, ip_ttl_hoplimit);

	// TODO flow_label is missing

	if (spec->tcp_flags) {
		DR_STE_SET_TCP_FLAGS(eth_l4, tag, spec);
		spec->tcp_flags = 0;
	}

	return 0;
}

int dr_ste_build_ipv6_l3_l4(struct dr_ste_build *sb,
			    struct dr_match_param *mask,
			    bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_ipv6_l3_l4_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_CALC_LU_TYPE(ETHL4, rx, inner);
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_ipv6_l3_l4_tag;

	return 0;
}

static int dr_ste_build_empty_always_hit_tag(struct dr_match_param *value,
					     struct dr_ste_build *sb,
					     uint8_t *hw_ste_p)
{
	return 0;
}

int dr_ste_build_empty_always_hit(struct dr_ste_build *sb, bool rx)
{
	sb->rx = rx;
	sb->lu_type = DR_STE_LU_TYPE_DONT_CARE;
	sb->byte_mask = 0;
	sb->ste_build_tag_func = &dr_ste_build_empty_always_hit_tag;

	return 0;
}

static int dr_ste_build_mpls_bit_mask(struct dr_match_param *value,
				      bool inner, uint8_t *bit_mask)
{
	struct dr_match_misc2 *misc2_mask = &value->misc2;

	if (inner)
		DR_STE_SET_MPLS_MASK(mpls, misc2_mask, inner, bit_mask);
	else
		DR_STE_SET_MPLS_MASK(mpls, misc2_mask, outer, bit_mask);

	return 0;
}

static int dr_ste_build_mpls_tag(struct dr_match_param *value,
				 struct dr_ste_build *sb,
				 uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_misc2 *misc2_mask = &value->misc2;
	uint8_t *tag = hw_ste->tag;

	if (sb->inner)
		DR_STE_SET_MPLS_TAG(mpls, misc2_mask, inner, tag);
	else
		DR_STE_SET_MPLS_TAG(mpls, misc2_mask, outer, tag);

	return 0;
}

int dr_ste_build_mpls(struct dr_ste_build *sb, struct dr_match_param *mask,
		      bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_mpls_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_CALC_LU_TYPE(MPLS_FIRST, rx, inner);
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_mpls_tag;

	return 0;
}

static int dr_ste_build_gre_bit_mask(struct dr_match_param *value,
				     bool inner, uint8_t *bit_mask)
{
	struct dr_match_misc *misc_mask = &value->misc;

	DR_STE_SET_MASK_V(gre, bit_mask, gre_protocol, misc_mask, gre_protocol);
	DR_STE_SET_MASK_V(gre, bit_mask, gre_k_present, misc_mask, gre_k_present);
	DR_STE_SET_MASK_V(gre, bit_mask, gre_key_h, misc_mask, gre_key_h);
	DR_STE_SET_MASK_V(gre, bit_mask, gre_key_l, misc_mask, gre_key_l);

	DR_STE_SET_MASK_V(gre, bit_mask, gre_c_present, misc_mask, gre_c_present);
	DR_STE_SET_MASK_V(gre, bit_mask, gre_s_present, misc_mask, gre_s_present);

	return 0;
}

static int dr_ste_build_gre_tag(struct dr_match_param *value,
				struct dr_ste_build *sb,
				uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct  dr_match_misc *misc = &value->misc;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(gre, tag, gre_protocol, misc, gre_protocol);

	DR_STE_SET_TAG(gre, tag, gre_k_present, misc, gre_k_present);
	DR_STE_SET_TAG(gre, tag, gre_key_h, misc, gre_key_h);
	DR_STE_SET_TAG(gre, tag, gre_key_l, misc, gre_key_l);

	DR_STE_SET_TAG(gre, tag, gre_c_present, misc, gre_c_present);

	DR_STE_SET_TAG(gre, tag, gre_s_present, misc, gre_s_present);

	return 0;
}

int dr_ste_build_gre(struct dr_ste_build *sb,struct dr_match_param *mask,
		     bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_gre_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_LU_TYPE_GRE;
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_gre_tag;

	return 0;
}

static int dr_ste_build_flex_parser_0_bit_mask(struct dr_match_param *value,
					       bool inner, uint8_t *bit_mask)
{
	struct dr_match_misc2 *misc_2_mask = &value->misc2;

	if (DR_STE_IS_OUTER_MPLS_OVER_GRE_SET(misc_2_mask)) {
		DR_STE_SET_MASK_V(flex_parser_0, bit_mask, parser_3_label,
				  misc_2_mask, outer_first_mpls_over_gre_label);

		DR_STE_SET_MASK_V(flex_parser_0, bit_mask, parser_3_exp,
				  misc_2_mask, outer_first_mpls_over_gre_exp);

		DR_STE_SET_MASK_V(flex_parser_0, bit_mask, parser_3_s_bos,
				  misc_2_mask, outer_first_mpls_over_gre_s_bos);

		DR_STE_SET_MASK_V(flex_parser_0, bit_mask, parser_3_ttl,
				  misc_2_mask, outer_first_mpls_over_gre_ttl);
	} else {
		DR_STE_SET_MASK_V(flex_parser_0, bit_mask, parser_3_label,
				  misc_2_mask, outer_first_mpls_over_udp_label);

		DR_STE_SET_MASK_V(flex_parser_0, bit_mask, parser_3_exp,
				  misc_2_mask, outer_first_mpls_over_udp_exp);

		DR_STE_SET_MASK_V(flex_parser_0, bit_mask, parser_3_s_bos,
				  misc_2_mask, outer_first_mpls_over_udp_s_bos);

		DR_STE_SET_MASK_V(flex_parser_0, bit_mask, parser_3_ttl,
				  misc_2_mask, outer_first_mpls_over_udp_ttl);
	}

	return 0;
}

static int dr_ste_build_flex_parser_0_tag(struct dr_match_param *value,
					  struct dr_ste_build *sb,
					  uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_misc2 *misc_2_mask = &value->misc2;
	uint8_t *tag = hw_ste->tag;

	if (DR_STE_IS_OUTER_MPLS_OVER_GRE_SET(misc_2_mask)) {
		DR_STE_SET_TAG(flex_parser_0, tag, parser_3_label,
			       misc_2_mask, outer_first_mpls_over_gre_label);

		DR_STE_SET_TAG(flex_parser_0, tag, parser_3_exp,
			       misc_2_mask, outer_first_mpls_over_gre_exp);

		DR_STE_SET_TAG(flex_parser_0, tag, parser_3_s_bos,
			       misc_2_mask, outer_first_mpls_over_gre_s_bos);

		DR_STE_SET_TAG(flex_parser_0, tag, parser_3_ttl,
			       misc_2_mask, outer_first_mpls_over_gre_ttl);
	} else {
		DR_STE_SET_TAG(flex_parser_0, tag, parser_3_label,
			       misc_2_mask, outer_first_mpls_over_udp_label);

		DR_STE_SET_TAG(flex_parser_0, tag, parser_3_exp,
			       misc_2_mask, outer_first_mpls_over_udp_exp);

		DR_STE_SET_TAG(flex_parser_0, tag, parser_3_s_bos,
			       misc_2_mask, outer_first_mpls_over_udp_s_bos);

		DR_STE_SET_TAG(flex_parser_0, tag, parser_3_ttl,
			       misc_2_mask, outer_first_mpls_over_udp_ttl);
	}
	return 0;
}

int dr_ste_build_flex_parser_0(struct dr_ste_build *sb,
			       struct dr_match_param *mask,
			       bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_flex_parser_0_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_LU_TYPE_FLEX_PARSER_0;
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_flex_parser_0_tag;

	return 0;
}

#define ICMP_TYPE_OFFSET_FIRST_DW		(24)
#define ICMP_CODE_OFFSET_FIRST_DW		(16)
#define ICMP_HEADER_DATA_OFFSET_SECOND_DW	(0)

static int dr_ste_build_flex_parser_1_bit_mask(struct dr_match_param *mask,
					       struct dr_devx_caps *caps,
					       uint8_t *bit_mask)
{
	struct dr_match_misc3 *misc_3_mask = &mask->misc3;
	bool is_ipv4_mask = DR_MASK_IS_FLEX_PARSER_ICMPV4_SET(misc_3_mask);
	uint32_t icmp_header_data_mask;
	uint32_t icmp_type_mask;
	uint32_t icmp_code_mask;
	int dw0_location;
	int dw1_location;

	if (is_ipv4_mask) {
		icmp_header_data_mask	= misc_3_mask->icmpv4_header_data;
		icmp_type_mask		= misc_3_mask->icmpv4_type;
		icmp_code_mask		= misc_3_mask->icmpv4_code;
		dw0_location		= caps->flex_parser_id_icmp_dw0;
		dw1_location		= caps->flex_parser_id_icmp_dw1;
	} else {
		icmp_header_data_mask	= misc_3_mask->icmpv6_header_data;
		icmp_type_mask		= misc_3_mask->icmpv6_type;
		icmp_code_mask		= misc_3_mask->icmpv6_code;
		dw0_location		= caps->flex_parser_id_icmpv6_dw0;
		dw1_location		= caps->flex_parser_id_icmpv6_dw1;
	}

	/* TBD: add where it can be 5,6,7 and not only 4 */
	switch (dw0_location) {
	case 4:
		if (icmp_type_mask) {
			DR_STE_SET(flex_parser_1, bit_mask, flex_parser_4,
				   (icmp_type_mask << ICMP_TYPE_OFFSET_FIRST_DW));
			if (is_ipv4_mask)
				misc_3_mask->icmpv4_type = 0;
			else
				misc_3_mask->icmpv6_type = 0;
		}
		if (icmp_code_mask) {
			uint32_t cur_val = DR_STE_GET(flex_parser_1, bit_mask,
						      flex_parser_4);
			DR_STE_SET(flex_parser_1, bit_mask, flex_parser_4,
				   cur_val | (icmp_code_mask << ICMP_CODE_OFFSET_FIRST_DW));
			if (is_ipv4_mask)
				misc_3_mask->icmpv4_code = 0;
			else
				misc_3_mask->icmpv6_code = 0;
		}
		break;
	default:
		fprintf(stderr, "%s New position for dw0 (%d), not supported (expecting 4)\n",
			__func__, dw0_location);
		return ENOENT;
	}

	switch (dw1_location) {
	case 5:
		if (icmp_header_data_mask) {
			DR_STE_SET(flex_parser_1, bit_mask, flex_parser_5,
				   (icmp_header_data_mask << ICMP_HEADER_DATA_OFFSET_SECOND_DW));
			if (is_ipv4_mask)
				misc_3_mask->icmpv4_header_data = 0;
			else
				misc_3_mask->icmpv6_header_data = 0;
		}
		break;
	default:
		fprintf(stderr, "%s New position for dw1 (%d), not supported (expecting 5)\n",
			__func__, dw1_location);
		return ENOENT;
	}

	return 0;
}

static int dr_ste_build_flex_parser_1_tag(struct dr_match_param *value,
					  struct dr_ste_build *sb,
					  uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_misc3 *misc_3 = &value->misc3;
	bool is_ipv4 = DR_MASK_IS_FLEX_PARSER_ICMPV4_SET(misc_3);
	uint8_t *tag = hw_ste->tag;
	uint32_t icmp_header_data;
	uint32_t icmp_type;
	uint32_t icmp_code;
	int dw0_location;
	int dw1_location;

	if (is_ipv4) {
		icmp_header_data	= misc_3->icmpv4_header_data;
		icmp_type		= misc_3->icmpv4_type;
		icmp_code		= misc_3->icmpv4_code;
		dw0_location		= sb->caps->flex_parser_id_icmp_dw0;
		dw1_location		= sb->caps->flex_parser_id_icmp_dw1;
	} else {
		icmp_header_data	= misc_3->icmpv6_header_data;
		icmp_type		= misc_3->icmpv6_type;
		icmp_code		= misc_3->icmpv6_code;
		dw0_location		= sb->caps->flex_parser_id_icmpv6_dw0;
		dw1_location		= sb->caps->flex_parser_id_icmpv6_dw1;
	}

	switch (dw0_location) {
	case 4:
		if (icmp_type) {
			DR_STE_SET(flex_parser_1, tag, flex_parser_4,
				   (icmp_type << ICMP_TYPE_OFFSET_FIRST_DW));
			if (is_ipv4)
				misc_3->icmpv4_type = 0;
			else
				misc_3->icmpv6_type = 0;
		}

		if (icmp_code) {
			uint32_t cur_val = DR_STE_GET(flex_parser_1, tag,
						      flex_parser_4);
			DR_STE_SET(flex_parser_1, tag, flex_parser_4,
				   cur_val | (icmp_code << ICMP_CODE_OFFSET_FIRST_DW));
			if (is_ipv4)
				misc_3->icmpv4_code = 0;
			else
				misc_3->icmpv6_code = 0;
		}
		break;
	default:
		fprintf(stderr, "%s New position for dw0 (%d), not supported (expecting 4)\n",
			__func__, dw0_location);
		return ENOENT;
	}
	switch (dw1_location) {
	case 5:
		if (icmp_header_data) {
			DR_STE_SET(flex_parser_1, tag, flex_parser_5,
				   (icmp_header_data << ICMP_HEADER_DATA_OFFSET_SECOND_DW));
			if (is_ipv4)
				misc_3->icmpv4_header_data = 0;
			else
				misc_3->icmpv6_header_data = 0;
		}
		break;
	default:
		fprintf(stderr, "%s New position for dw1 (%d), not supported (expecting 5)\n",
			__func__, dw1_location);
		return ENOENT;
	}

	return 0;
}

int dr_ste_build_flex_parser_1(struct dr_ste_build *sb,
			       struct dr_match_param *mask,
			       struct dr_devx_caps *caps,
			       bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_flex_parser_1_bit_mask(mask,
						  caps,
						  sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->caps = caps;
	sb->lu_type = DR_STE_LU_TYPE_FLEX_PARSER_1;
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_flex_parser_1_tag;

	return 0;
}

static int dr_ste_build_general_purpose_bit_mask(struct dr_match_param *value,
						 bool inner, uint8_t *bit_mask)
{
	struct dr_match_misc2 *misc_2_mask = &value->misc2;

	DR_STE_SET_MASK_V(general_purpose, bit_mask,
			  general_purpose_lookup_field, misc_2_mask,
			  metadata_reg_a);

	return 0;
}

static int dr_ste_build_general_purpose_tag(struct dr_match_param *value,
					    struct dr_ste_build *sb,
					    uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_misc2 *misc_2_mask = &value->misc2;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(general_purpose, tag, general_purpose_lookup_field,
		       misc_2_mask, metadata_reg_a);

	return 0;
}

int dr_ste_build_general_purpose(struct dr_ste_build *sb,
				 struct dr_match_param *mask,
				 bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_general_purpose_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_LU_TYPE_GENERAL_PURPOSE;
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_general_purpose_tag;

	return 0;
}

static int dr_ste_build_eth_l4_misc_bit_mask(struct dr_match_param *value,
						 bool inner, uint8_t *bit_mask)
{
	struct dr_match_misc3 *misc_3_mask = &value->misc3;

	if (inner) {
		DR_STE_SET_MASK_V(eth_l4_misc, bit_mask, seq_num, misc_3_mask,
				  inner_tcp_seq_num);
		DR_STE_SET_MASK_V(eth_l4_misc, bit_mask, ack_num, misc_3_mask,
				  inner_tcp_ack_num);
	} else {
		DR_STE_SET_MASK_V(eth_l4_misc, bit_mask, seq_num, misc_3_mask,
				  outer_tcp_seq_num);
		DR_STE_SET_MASK_V(eth_l4_misc, bit_mask, ack_num, misc_3_mask,
				  outer_tcp_ack_num);
	}

	return 0;
}

static int dr_ste_build_eth_l4_misc_tag(struct dr_match_param *value,
					struct dr_ste_build *sb,
					uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_misc3 *misc3 = &value->misc3;
	uint8_t *tag = hw_ste->tag;

	if (sb->inner) {
		DR_STE_SET_TAG(eth_l4_misc, tag, seq_num, misc3, inner_tcp_seq_num);
		DR_STE_SET_TAG(eth_l4_misc, tag, ack_num, misc3, inner_tcp_ack_num);
	} else {
		DR_STE_SET_TAG(eth_l4_misc, tag, seq_num, misc3, outer_tcp_seq_num);
		DR_STE_SET_TAG(eth_l4_misc, tag, ack_num, misc3, outer_tcp_ack_num);
	}

	return 0;
}

int dr_ste_build_eth_l4_misc(struct dr_ste_build *sb,
				 struct dr_match_param *mask,
				 bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_eth_l4_misc_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_CALC_LU_TYPE(ETHL4_MISC, rx, inner);
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_eth_l4_misc_tag;

	return 0;
}

static int dr_ste_build_flex_parser_tnl_bit_mask(struct dr_match_param *value,
						 bool inner, uint8_t *bit_mask)
{
	struct dr_match_misc3 *misc_3_mask = &value->misc3;

	if (misc_3_mask->outer_vxlan_gpe_flags ||
	    misc_3_mask->outer_vxlan_gpe_next_protocol) {
		DR_STE_SET(flex_parser_tnl, bit_mask,
			   flex_parser_tunneling_header_63_32,
			   (misc_3_mask->outer_vxlan_gpe_flags << 24) |
			   (misc_3_mask->outer_vxlan_gpe_next_protocol));
		misc_3_mask->outer_vxlan_gpe_flags = 0;
		misc_3_mask->outer_vxlan_gpe_next_protocol = 0;
	}

	if (misc_3_mask->outer_vxlan_gpe_vni) {
		DR_STE_SET(flex_parser_tnl, bit_mask,
			   flex_parser_tunneling_header_31_0,
			   misc_3_mask->outer_vxlan_gpe_vni << 8);
		misc_3_mask->outer_vxlan_gpe_vni = 0;
	}

	return 0;
}

static int dr_ste_build_flex_parser_tnl_tag(struct dr_match_param *value,
					    struct dr_ste_build *sb,
					    uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_misc3 *misc3 = &value->misc3;
	uint8_t *tag = hw_ste->tag;

	if (misc3->outer_vxlan_gpe_flags ||
	    misc3->outer_vxlan_gpe_next_protocol) {
		DR_STE_SET(flex_parser_tnl, tag,
			   flex_parser_tunneling_header_63_32,
			   (misc3->outer_vxlan_gpe_flags << 24) |
			   (misc3->outer_vxlan_gpe_next_protocol));
		misc3->outer_vxlan_gpe_flags = 0;
		misc3->outer_vxlan_gpe_next_protocol = 0;
	}

	if (misc3->outer_vxlan_gpe_vni) {
		DR_STE_SET(flex_parser_tnl, tag,
			   flex_parser_tunneling_header_31_0,
			   misc3->outer_vxlan_gpe_vni << 8);
		misc3->outer_vxlan_gpe_vni = 0;
	}

	return 0;
}

int dr_ste_build_flex_parser_tnl(struct dr_ste_build *sb,
				 struct dr_match_param *mask,
				 bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_flex_parser_tnl_bit_mask(mask, inner, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_LU_TYPE_FLEX_PARSER_TNL_HEADER;
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_flex_parser_tnl_tag;

	return 0;
}

static int dr_ste_build_register_0_bit_mask(struct dr_match_param *value,
					    uint8_t *bit_mask)
{
	struct dr_match_misc2 *misc_2_mask = &value->misc2;

	DR_STE_SET_MASK_V(register_0, bit_mask, register_0_h,
			  misc_2_mask, metadata_reg_c_0);
	DR_STE_SET_MASK_V(register_0, bit_mask, register_0_l,
			  misc_2_mask, metadata_reg_c_1);
	DR_STE_SET_MASK_V(register_0, bit_mask, register_1_h,
			  misc_2_mask, metadata_reg_c_2);
	DR_STE_SET_MASK_V(register_0, bit_mask, register_1_l,
			  misc_2_mask, metadata_reg_c_3);

	return 0;
}

static int dr_ste_build_register_0_tag(struct dr_match_param *value,
				       struct dr_ste_build *sb,
				       uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_misc2 *misc2 = &value->misc2;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(register_0, tag, register_0_h, misc2, metadata_reg_c_0);
	DR_STE_SET_TAG(register_0, tag, register_0_l, misc2, metadata_reg_c_1);
	DR_STE_SET_TAG(register_0, tag, register_1_h, misc2, metadata_reg_c_2);
	DR_STE_SET_TAG(register_0, tag, register_1_l, misc2, metadata_reg_c_3);

	return 0;
}

int dr_ste_build_register_0(struct dr_ste_build *sb,
			    struct dr_match_param *mask,
			    bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_register_0_bit_mask(mask, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_LU_TYPE_STEERING_REGISTERS_0;
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_register_0_tag;

	return 0;
}

static int dr_ste_build_register_1_bit_mask(struct dr_match_param *value,
					    uint8_t *bit_mask)
{
	struct dr_match_misc2 *misc_2_mask = &value->misc2;

	DR_STE_SET_MASK_V(register_1, bit_mask, register_2_h,
			  misc_2_mask, metadata_reg_c_4);
	DR_STE_SET_MASK_V(register_1, bit_mask, register_2_l,
			  misc_2_mask, metadata_reg_c_5);
	DR_STE_SET_MASK_V(register_1, bit_mask, register_3_h,
			  misc_2_mask, metadata_reg_c_6);
	DR_STE_SET_MASK_V(register_1, bit_mask, register_3_l,
			  misc_2_mask, metadata_reg_c_7);

	return 0;
}

static int dr_ste_build_register_1_tag(struct dr_match_param *value,
				       struct dr_ste_build *sb,
				       uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_misc2 *misc2 = &value->misc2;
	uint8_t *tag = hw_ste->tag;

	DR_STE_SET_TAG(register_1, tag, register_2_h, misc2, metadata_reg_c_4);
	DR_STE_SET_TAG(register_1, tag, register_2_l, misc2, metadata_reg_c_5);
	DR_STE_SET_TAG(register_1, tag, register_3_h, misc2, metadata_reg_c_6);
	DR_STE_SET_TAG(register_1, tag, register_3_l, misc2, metadata_reg_c_7);

	return 0;
}

int dr_ste_build_register_1(struct dr_ste_build *sb,
			    struct dr_match_param *mask,
			    bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_register_1_bit_mask(mask, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->inner = inner;
	sb->lu_type = DR_STE_LU_TYPE_STEERING_REGISTERS_1;
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_register_1_tag;

	return 0;
}

static int dr_ste_build_src_gvmi_qpn_bit_mask(struct dr_match_param *value,
					      uint8_t *bit_mask)
{
	struct dr_match_misc *misc_mask = &value->misc;

	if (misc_mask->source_port != 0xffff) {
		fprintf(stderr, "Partial misc source_post is not supported\n");
		return EINVAL;
	}
	DR_STE_SET_MASK(src_gvmi_qp, bit_mask, source_gvmi, misc_mask, source_port);
	DR_STE_SET_MASK(src_gvmi_qp, bit_mask, source_qp, misc_mask, source_sqn);

	return 0;
}

static int dr_ste_build_src_gvmi_qpn_tag(struct dr_match_param *value,
					 struct dr_ste_build *sb,
					 uint8_t *hw_ste_p)
{
	struct dr_hw_ste_format *hw_ste = (struct dr_hw_ste_format *)hw_ste_p;
	struct dr_match_misc *misc = &value->misc;
	struct dr_devx_vport_cap *vport_cap;
	uint8_t *tag = hw_ste->tag;

	vport_cap = dr_get_vport_cap(sb->caps, misc->source_port);
	if (!vport_cap)
		return EINVAL;

	DR_STE_SET_TAG(src_gvmi_qp, tag, source_qp, misc, source_sqn);

	if (vport_cap->gvmi)
		DR_STE_SET(src_gvmi_qp, tag, source_gvmi, vport_cap->gvmi);

	misc->source_port = 0;

	return 0;
}

int dr_ste_build_src_gvmi_qpn(struct dr_ste_build *sb,
			      struct dr_match_param *mask,
			      struct dr_devx_caps *caps,
			      bool inner, bool rx)
{
	int ret;

	ret = dr_ste_build_src_gvmi_qpn_bit_mask(mask, sb->bit_mask);
	if (ret)
		return ret;

	sb->rx = rx;
	sb->caps = caps;
	sb->inner = inner;
	sb->lu_type = DR_STE_LU_TYPE_SRC_GVMI_AND_QP;
	sb->byte_mask = dr_ste_conv_bit_to_byte_mask(sb->bit_mask);
	sb->ste_build_tag_func = &dr_ste_build_src_gvmi_qpn_tag;

	return 0;
}
