/*
 * Copyright (c) 2018 Mellanox Technologies, Inc.  All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING in the main directory of this source tree, or the
 * OpenIB.org BSD license below:
 *
 *     Redistribution and use in source and binary forms, with or
 *     without modification, are permitted provided that the following
 *     conditions are met:
 *
 *      - Redistributions of source code must retain the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer.
 *
 *      - Redistributions in binary form must reproduce the above
 *        copyright notice, this list of conditions and the following
 *        disclaimer in the documentation and/or other materials
 *        provided with the distribution.
 *
 * THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND,
 * EXPRESS OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF
 * MERCHANTABILITY, FITNESS FOR A PARTICULAR PURPOSE AND
 * NONINFRINGEMENT. IN NO EVENT SHALL THE AUTHORS OR COPYRIGHT HOLDERS
 * BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER IN AN
 * ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
 * CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
 * SOFTWARE.
 */

#include <stdlib.h>
#include <stdio.h>
#include <malloc.h>
#include <unistd.h>
#include <util/mmio.h>
#include "mlx5dv.h"
#include "mlx5dv_dr.h"
#include "mlx5.h"

enum {
	CQ_OK = 0,
	CQ_EMPTY = -1,
	CQ_POLL_ERR = -2
};

struct dr_cq {
	uint8_t				*buf;
	uint32_t			cons_index;
	int				ncqe;
	struct dr_qp			*qp; /* Assume cq per QP */
	__be32				*db;
	struct mlx5dv_devx_obj		*obj;
	struct mlx5dv_devx_uar		*uar;
	struct mlx5dv_devx_umem		*buf_umem;
	struct mlx5dv_devx_umem		*db_umem;
};

struct dr_sg_copy_ptr {
	int	index;
	int	offset;
};

struct dr_wqe_inline_seg {
	__be32	byte_count;
};

static inline void *dr_cq_get_cqe(struct dr_cq *dr_cq, int n)
{
	return dr_cq->buf + n * sizeof(struct mlx5_cqe64);
}

static inline void *dr_cq_get_sw_cqe(struct dr_cq *dr_cq, int n)
{
	struct mlx5_cqe64 *cqe64 = dr_cq_get_cqe(dr_cq, n & (dr_cq->ncqe - 1));

	if (likely(mlx5dv_get_cqe_opcode(cqe64) != MLX5_CQE_INVALID) &&
	    !((cqe64->op_own & MLX5_CQE_OWNER_MASK) ^
	      !!(n & dr_cq->ncqe)))
		return cqe64;
	else
		return NULL;
}

struct dr_cq *dr_create_cq(struct ibv_context *ctx,
			   struct mlx5dv_devx_uar *uar,
			   int cqen)
{
	struct mlx5dv_devx_umem *umem;
	struct mlx5dv_devx_obj *obj;
	struct mlx5_cqe64 *cqe;
	int cqe_sz = sizeof(*cqe);
	struct dr_cq *dr_cq;
	uint8_t *buf;
	uint32_t eqn;
	int ncqe;
	int err;
	int i;

	dr_cq = calloc(1, sizeof(*dr_cq));
	if (!dr_cq)
		return NULL;

	err = mlx5dv_devx_query_eqn(ctx, 0, &eqn);
	if (err)
		goto free_dr_cq;

	ncqe = mlx5_round_up_power_of_two(cqen);
	if (ncqe < 0)
		goto free_dr_cq;

	buf = memalign(sysconf(_SC_PAGESIZE), cqe_sz * ncqe);
	if (!buf)
		goto free_dr_cq;

	for (i = 0; i < ncqe; i++) {
		cqe = (struct mlx5_cqe64 *)(buf + i * sizeof(*cqe));
		cqe->op_own = MLX5_CQE_INVALID << 4;
	}
	dr_cq->buf = buf;

	umem = mlx5dv_devx_umem_reg(ctx, buf, cqe_sz * ncqe,
				    IBV_ACCESS_LOCAL_WRITE |
				    IBV_ACCESS_REMOTE_WRITE |
				    IBV_ACCESS_REMOTE_READ);
	if (!umem)
		goto err_buf_umem;

	dr_cq->buf_umem = umem;
	dr_cq->ncqe = ncqe;

	dr_cq->db = memalign(8, 8);
	if (!dr_cq->db)
		goto err_db_alloc;

	dr_cq->db[MLX5_CQ_SET_CI] = 0;
	dr_cq->db[MLX5_CQ_ARM_DB] = 0;
	umem = mlx5dv_devx_umem_reg(ctx, dr_cq->db,
				    8,
				    IBV_ACCESS_LOCAL_WRITE |
				    IBV_ACCESS_REMOTE_WRITE |
				    IBV_ACCESS_REMOTE_READ);
	if (!umem)
		goto err_db_umem;

	dr_cq->db_umem = umem;

	obj = dr_devx_create_cq(ctx, uar->page_id,
				dr_cq->buf_umem->umem_id,
				dr_cq->db_umem->umem_id,
				eqn, ncqe, cqen);
	if (!obj)
		goto err_cq_obj_create;

	dr_cq->obj = obj;
	dr_cq->uar = uar;

	return dr_cq;

err_cq_obj_create:
	mlx5dv_devx_umem_dereg(dr_cq->db_umem);
err_db_umem:
	free(dr_cq->db);
err_db_alloc:
	mlx5dv_devx_umem_dereg(dr_cq->buf_umem);
err_buf_umem:
	free(dr_cq->buf);
free_dr_cq:
	free(dr_cq);
	return NULL;
}

int dr_destroy_cq(struct dr_cq *dr_cq)
{
	int ret;

	ret = mlx5dv_devx_obj_destroy(dr_cq->obj);
	if (ret)
		return ret;

	ret = mlx5dv_devx_umem_dereg(dr_cq->db_umem);
	if (ret)
		return ret;

	ret = mlx5dv_devx_umem_dereg(dr_cq->buf_umem);
	if (ret)
		return ret;

	free(dr_cq->db);
	free(dr_cq->buf);
	free(dr_cq);

	return 0;
}

static inline int dr_get_next_cqe(struct dr_cq *dr_cq,
				  struct mlx5_cqe64 **pcqe64)
{
	struct mlx5_cqe64 *cqe64;

	cqe64 = dr_cq_get_sw_cqe(dr_cq, dr_cq->cons_index);
	if (!cqe64)
		return CQ_EMPTY;

	++dr_cq->cons_index;
	/*
	 * Make sure we read CQ entry contents after we've checked the
	 * ownership bit.
	 */
	udma_from_device_barrier();

	*pcqe64 = cqe64;

	return CQ_OK;
}

static inline int dr_parse_cqe(struct dr_cq *dr_cq, struct mlx5_cqe64 *cqe64)
{
	uint16_t wqe_ctr;
	uint8_t opcode;
	int idx;

	wqe_ctr = be16toh(cqe64->wqe_counter);
	opcode = mlx5dv_get_cqe_opcode(cqe64);
	if (opcode == MLX5_CQE_REQ_ERR) {
		idx = wqe_ctr & (dr_cq->qp->sq.wqe_cnt - 1);
		dr_cq->qp->sq.tail = dr_cq->qp->sq.wqe_head[idx] + 1;
	} else if (opcode == MLX5_CQE_RESP_ERR) {
		++dr_cq->qp->sq.tail;
	} else {
		idx = wqe_ctr & (dr_cq->qp->sq.wqe_cnt - 1);
		dr_cq->qp->sq.tail = dr_cq->qp->sq.wqe_head[idx] + 1;

		return CQ_OK;
	}

	return CQ_POLL_ERR;
}

static inline int dr_cq_poll_one(struct dr_cq *dr_cq)
{
	struct mlx5_cqe64 *cqe64;
	int err;

	err = dr_get_next_cqe(dr_cq, &cqe64);
	if (err)
		return CQ_EMPTY;

	return dr_parse_cqe(dr_cq, cqe64);
}

int dr_poll_cq(struct dr_cq *dr_cq, int ne)
{
	int npolled;
	int err = 0;

	for (npolled = 0; npolled < ne; ++npolled) {
		err = dr_cq_poll_one(dr_cq);
		if (err != CQ_OK)
			break;
	}
	dr_cq->db[MLX5_CQ_SET_CI] = htobe32(dr_cq->cons_index &
					       0xffffff);
	return err == CQ_POLL_ERR ? err : npolled;
}

static int dr_qp_calc_send_wqe(struct dr_qp_init_attr *attr)
{
	int size;
	int inl_size = 0;
	int tot_size;

	size = sizeof(struct mlx5_wqe_ctrl_seg) +
		sizeof(struct mlx5_wqe_raddr_seg);
	if (attr->cap.max_inline_data) {
		inl_size = size + align(sizeof(struct mlx5_wqe_inl_data_seg) +
					attr->cap.max_inline_data, 16);
	}

	size += attr->cap.max_send_sge * sizeof(struct mlx5_wqe_data_seg);
	tot_size = max_int(size, inl_size);

	return align(tot_size, MLX5_SEND_WQE_BB);
}

static int dr_calc_sq_size(struct dr_qp *dr_qp,
			   struct dr_qp_init_attr *attr)
{
	int wqe_size;
	int wq_size;

	wqe_size = dr_qp_calc_send_wqe(attr);
	if (wqe_size < 0)
		return EINVAL;

	dr_qp->max_inline_data = wqe_size -
		(sizeof(struct  mlx5_wqe_ctrl_seg) +
		 sizeof(struct mlx5_wqe_raddr_seg)) -
		sizeof(struct mlx5_wqe_inl_data_seg);

	wq_size = mlx5_round_up_power_of_two(attr->cap.max_send_wr * wqe_size);
	dr_qp->sq.wqe_cnt = wq_size / MLX5_SEND_WQE_BB;
	dr_qp->sq.wqe_shift = mlx5_ilog2(MLX5_SEND_WQE_BB);
	dr_qp->sq.max_gs = attr->cap.max_send_sge;
	dr_qp->sq.max_post = wq_size / wqe_size;

	return wq_size;
}

static int dr_qp_calc_recv_wqe(struct dr_qp_init_attr *attr)
{
	uint32_t size;
	int num_scatter;

	num_scatter = max_t(uint32_t, attr->cap.max_recv_sge, 1);
	size = sizeof(struct mlx5_wqe_data_seg) * num_scatter;

	size = mlx5_round_up_power_of_two(size);

	return size;
}

static int dr_calc_rq_size(struct dr_qp *dr_qp,
			   struct dr_qp_init_attr *attr)
{
	int wqe_size;
	int wq_size;

	wqe_size = dr_qp_calc_recv_wqe(attr);
	if (wqe_size < 0)
		return EINVAL;

	wq_size = mlx5_round_up_power_of_two(attr->cap.max_recv_wr) * wqe_size;
	wq_size = max(wq_size, MLX5_SEND_WQE_BB);
	dr_qp->rq.wqe_cnt = wq_size / wqe_size;
	dr_qp->rq.wqe_shift = mlx5_ilog2(wqe_size);
	dr_qp->rq.max_post = 1 << mlx5_ilog2(wq_size / wqe_size);
	dr_qp->rq.max_gs = wqe_size / sizeof(struct mlx5_wqe_data_seg);

	return wq_size;
}

static int dr_calc_wq_size(struct dr_qp *dr_qp, struct dr_qp_init_attr *attr)
{
	int result;
	int ret;

	ret = dr_calc_sq_size(dr_qp, attr);
	if (ret < 0)
		return ret;

	result = ret;
	ret = dr_calc_rq_size(dr_qp, attr);
	if (ret < 0)
		return ret;

	result += ret;
	dr_qp->sq.offset = ret;
	dr_qp->rq.offset = 0;

	return result;
}

static int dr_qp_alloc_buf(struct dr_qp *dr_qp, int size)
{
	int al_size;
	int ret;

	dr_qp->sq.wqe_head = malloc(dr_qp->sq.wqe_cnt *
				    sizeof(*dr_qp->sq.wqe_head));
	if (!dr_qp->sq.wqe_head)
		return ENOMEM;

	al_size = align(size, sysconf(_SC_PAGESIZE));
	ret = posix_memalign(&dr_qp->buf.buf, sysconf(_SC_PAGESIZE), al_size);
	if (ret)
		return ret;

	dr_qp->buf.length = al_size;
	dr_qp->buf.type = MLX5_ALLOC_TYPE_ANON;
	memset(dr_qp->buf.buf, 0, dr_qp->buf.length);

	return 0;
}

struct dr_qp *dr_create_qp(struct ibv_context *ctx,
			   struct dr_qp_init_attr *attr)
{
	struct mlx5dv_devx_obj *obj;
	struct dr_qp *dr_qp;
	int ret;

	if (attr->qp_type != IBV_QPT_RC || attr->cq->qp)
		return NULL;

	dr_qp = calloc(1, sizeof(*dr_qp));
	if (!dr_qp)
		return NULL;

	ret = dr_calc_wq_size(dr_qp, attr);
	if (ret < 0)
		return NULL;

	if (dr_qp_alloc_buf(dr_qp, ret))
		goto err_alloc_bufs;

	dr_qp->sq_start = dr_qp->buf.buf + dr_qp->sq.offset;
	dr_qp->sq.qend = dr_qp->buf.buf + dr_qp->sq.offset +
		(dr_qp->sq.wqe_cnt << dr_qp->sq.wqe_shift);
	dr_qp->rq.head = 0;
	dr_qp->rq.tail = 0;
	dr_qp->sq.cur_post = 0;

	dr_qp->db = memalign(8, 8);
	if (!dr_qp->db)
		goto err_db_alloc;

	dr_qp->db[MLX5_RCV_DBR] = 0;
	dr_qp->db[MLX5_SND_DBR] = 0;
	dr_qp->db_umem = mlx5dv_devx_umem_reg(ctx, dr_qp->db, 8,
					      IBV_ACCESS_LOCAL_WRITE |
					      IBV_ACCESS_REMOTE_WRITE |
					      IBV_ACCESS_REMOTE_READ);
	if (!dr_qp->db_umem)
		goto err_db_umem;

	dr_qp->buf_umem = mlx5dv_devx_umem_reg(ctx, dr_qp->buf.buf,
					       dr_qp->buf.length,
					       IBV_ACCESS_LOCAL_WRITE |
					       IBV_ACCESS_REMOTE_WRITE |
					       IBV_ACCESS_REMOTE_READ);
	if (!dr_qp->buf_umem)
		goto err_buf_umem;

	obj = dr_devx_create_qp(ctx,
				attr->uar->page_id,
				attr->pdn,
				attr->cq->obj->object_id,
				MLX5_QPC_PM_STATE_MIGRATED,
				MLX5_QPC_ST_RC,
				dr_qp->buf_umem->umem_id,
				dr_qp->db_umem->umem_id,
				dr_qp->sq.wqe_cnt,
				dr_qp->rq.wqe_cnt,
				dr_qp->rq.wqe_shift);
	if (!obj)
		goto err_qp_create;

	attr->cq->qp = dr_qp;
	dr_qp->uar = attr->uar;
	dr_qp->obj = obj;

	return dr_qp;

err_qp_create:
	mlx5dv_devx_umem_dereg(dr_qp->buf_umem);
err_buf_umem:
	mlx5dv_devx_umem_dereg(dr_qp->db_umem);
err_db_umem:
	free(dr_qp->db);
err_db_alloc:
err_alloc_bufs:
	if (dr_qp->sq.wqe_head)
		free(dr_qp->sq.wqe_head);

	if (dr_qp->buf.buf)
		free(dr_qp->buf.buf);

	free(dr_qp);

	return NULL;
}

int dr_destroy_qp(struct dr_qp *dr_qp)
{
	int ret;

	ret = mlx5dv_devx_obj_destroy(dr_qp->obj);
	if (ret)
		return ret;

	ret = mlx5dv_devx_umem_dereg(dr_qp->buf_umem);
	if (ret)
		return ret;

	ret = mlx5dv_devx_umem_dereg(dr_qp->db_umem);
	if (ret)
		return ret;

	free(dr_qp->db);
	free(dr_qp->sq.wqe_head);
	free(dr_qp->buf.buf);
	free(dr_qp);

	return 0;
}

static int dr_wq_overflow(struct dr_wq *wq, int nreq)
{
	unsigned cur;

	cur = wq->head - wq->tail;
	if (cur + nreq < wq->max_post)
		return 0;

	cur = wq->head - wq->tail;

	return cur + nreq >= wq->max_post;
}

static inline void set_raddr_seg(struct mlx5_wqe_raddr_seg *rseg,
				 uint64_t remote_addr, uint32_t rkey)
{
	rseg->raddr    = htobe64(remote_addr);
	rseg->rkey     = htobe32(rkey);
	rseg->reserved = 0;
}

static void set_data_ptr_seg(struct mlx5_wqe_data_seg *dseg, struct ibv_sge *sg,
			     int offset)
{
	dseg->byte_count = htobe32(sg->length - offset);
	dseg->lkey       = htobe32(sg->lkey);
	dseg->addr       = htobe64(sg->addr + offset);
}

static const uint32_t dr_ib_opcode[] = {
	[IBV_WR_RDMA_WRITE]		= MLX5_OPCODE_RDMA_WRITE,
	[IBV_WR_RDMA_READ]		= MLX5_OPCODE_RDMA_READ,
};

static inline void dr_post_send_db(struct dr_qp *dr_qp, int nreq, int inl,
				   int size, void *ctrl)
{
	if (unlikely(!nreq))
		return;

	dr_qp->sq.head += nreq;

	/*
	 * Make sure that descriptors are written before
	 * updating doorbell record and ringing the doorbell
	 */
	udma_to_device_barrier();
	dr_qp->db[MLX5_SND_DBR] = htobe32(dr_qp->sq.cur_post & 0xffff);

	/* Make sure that the doorbell write happens before the memcpy
	 * to WC memory below
	 */
	mmio_wc_start();

	if (inl && (nreq == 1)) {
		uint64_t *dst = dr_qp->uar->reg_addr;
		const uint64_t *src = ctrl;
		unsigned bytecnt = align(size * 16, 64);

		do {
			mmio_memcpy_x64(dst, src, 64);
			bytecnt -= 64;
			dst += 8;
			src += 8;
			if (unlikely(src == dr_qp->sq.qend))
				src = dr_qp->sq_start;
		} while (bytecnt > 0);
	} else {
		mmio_write64_be((uint8_t *)dr_qp->uar->reg_addr,
				*(__be64 *)ctrl);
	}

	mmio_flush_writes();
}

static inline int set_data_inl_seg(struct dr_qp *dr_qp,
				   struct ibv_send_wr *wr,
				   void *wqe, int *sz,
				   struct dr_sg_copy_ptr *sg_copy_ptr)
{
	int offset = sg_copy_ptr->offset;
	struct dr_wqe_inline_seg *seg;
	void *qend = dr_qp->sq.qend;
	int inl = 0;
	void *addr;
	int copy;
	int len;
	int i;

	seg = wqe;
	wqe += sizeof(*seg);
	for (i = sg_copy_ptr->index; i < wr->num_sge; ++i) {
		addr = (void *) (unsigned long)(wr->sg_list[i].addr + offset);
		len  = wr->sg_list[i].length - offset;
		inl += len;
		offset = 0;

		if (unlikely(inl > dr_qp->max_inline_data))
			return ENOMEM;

		if (unlikely(wqe + len > qend)) {
			copy = qend - wqe;
			memcpy(wqe, addr, copy);
			addr += copy;
			len -= copy;
			wqe = dr_qp->sq_start;
		}
		memcpy(wqe, addr, len);
		wqe += len;
	}

	if (likely(inl)) {
		seg->byte_count = htobe32(inl | MLX5_INLINE_SEG);
		*sz = align(inl + sizeof(seg->byte_count), 16) / 16;
	} else {
		*sz = 0;
	}

	return 0;
}

/* Assume post_send and poll can't be called at the same time, dr code */
int dr_post_send(struct dr_qp *dr_qp, struct ibv_send_wr *wr)
{
	struct dr_sg_copy_ptr sg_copy_ptr = {.index = 0, .offset = 0};
	struct mlx5_wqe_ctrl_seg *ctrl = NULL;
	struct mlx5_wqe_data_seg *dpseg;
	void *qend = dr_qp->sq.qend;
	uint32_t dr_opcode;
	unsigned idx;
	int size = 0;
	int inl = 0;
	int err = 0;
	void *seg;
	int nreq;
	int i;

	for (nreq = 0; wr; ++nreq, wr = wr->next) {
		if (unlikely(dr_wq_overflow(&dr_qp->sq, nreq))) {
			err = ENOMEM;
			goto out;
		}

		if (unlikely(wr->num_sge > dr_qp->sq.max_gs)) {
			err = ENOMEM;
			goto out;
		}

		idx = dr_qp->sq.cur_post & (dr_qp->sq.wqe_cnt - 1);
		ctrl = seg = dr_qp->sq_start + (idx << MLX5_SEND_WQE_SHIFT);
		*(uint32_t *)(seg + 8) = 0;
		ctrl->imm = 0;
		ctrl->fm_ce_se = wr->send_flags & IBV_SEND_SIGNALED ?
			 MLX5_WQE_CTRL_CQ_UPDATE : 0;

		seg += sizeof(*ctrl);
		size = sizeof(*ctrl) / 16;

		/* Only RDMA_WRITE/RDMA_READ */
		set_raddr_seg(seg, wr->wr.rdma.remote_addr,
			      wr->wr.rdma.rkey);
		seg  += sizeof(struct mlx5_wqe_raddr_seg);
		size += sizeof(struct mlx5_wqe_raddr_seg) / 16;

		if (wr->send_flags & IBV_SEND_INLINE && wr->num_sge) {
			int uninitialized_var(sz);

			err = set_data_inl_seg(dr_qp, wr, seg, &sz,
					       &sg_copy_ptr);
			if (unlikely(err))
				goto out;
			inl = 1;
			size += sz;
		} else {
			dpseg = seg;
			for (i = sg_copy_ptr.index; i < wr->num_sge; ++i) {
				if (unlikely(dpseg == qend)) {
					seg = dr_qp->sq_start;
					dpseg = seg;
				}
				if (likely(wr->sg_list[i].length)) {
					set_data_ptr_seg(dpseg, wr->sg_list + i,
							 sg_copy_ptr.offset);
					sg_copy_ptr.offset = 0;
					++dpseg;
					size += sizeof(struct mlx5_wqe_data_seg) / 16;
				}
			}
		}
		dr_opcode = dr_ib_opcode[wr->opcode];
		ctrl->opmod_idx_opcode =
			htobe32(((dr_qp->sq.cur_post & 0xffff) << 8) |
				dr_opcode);
		ctrl->qpn_ds = htobe32(size | (dr_qp->obj->object_id << 8));
		dr_qp->sq.wqe_head[idx] = dr_qp->sq.head + nreq;
		dr_qp->sq.cur_post += DIV_ROUND_UP(size * 16, MLX5_SEND_WQE_BB);
	}
out:
	dr_post_send_db(dr_qp, nreq, inl, size, ctrl);
	return err;
}
